﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Bentley.Interop.MicroStationDGN;
using System.Windows.Forms;

namespace ISY.CAD.Noise
{
    class AddBuildingMgr : ILocateCommandEvents
    {
        private static AddBuildingForm form;
        static List<Point3d> OutLines;
        static List<Point3d> InLines;


        private static int numberOfTopLines = 0;
        private static string typeOfOutline = string.Empty;


        public AddBuildingMgr()
        {
            form = new AddBuildingForm();
        }

        public static void StartCommand()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            AddBuildingMgr command = new AddBuildingMgr();
            InLines = null;
            OutLines = null;
            App.CommandState.StartLocate(command);
            App.CommandState.StartDynamics();
        }

        private void AddElement(Element Element)
        {
            if ((Element.Type == MsdElementType.LineString) || (Element.Type == MsdElementType.Line))
            {
                LineElement el = Element.AsLineElement();
                if (OutLines == null)
                    OutLines = new List<Point3d>();
                foreach (Point3d p in el.GetVertices())
                    if (!OutLines.Contains(p))
                        OutLines.Add(p);
                scanForTopLines(el.Range);
                typeOfOutline = "Line";
            }
            else if (Element.Type == MsdElementType.CellHeader)
            {
                CellElement el = Element.AsCellElement();
                ElementEnumerator ee = el.GetSubElements();
                while (ee.MoveNext())
                {
                    ComplexShapeElement shape;
                    shape = ee.Current.AsComplexShapeElement();
                    ElementEnumerator ee2 = shape.GetSubElements();
                    while (ee2.MoveNext())
                    {
                        LineElement e;
                        e = ee2.Current.AsLineElement();
                        if (OutLines == null)
                            OutLines = new List<Point3d>();
                        foreach (Point3d p in e.GetVertices())
                            OutLines.Add(p);
                        scanForTopLines(e.Range);
                        typeOfOutline = "Cell";
                    }
                }
            }
            else if (Element.Type == MsdElementType.Shape)
            {
                ShapeElement el = Element.AsShapeElement();
                if (OutLines == null)
                    OutLines = new List<Point3d>();
                foreach (Point3d p in el.GetVertices())
                    OutLines.Add(p);
                scanForTopLines(el.Range);
                typeOfOutline = "Shape";
            }
            CreateAndReset();
        }

        private void scanForTopLines(Range3d range3d)
        {
            numberOfTopLines = 0;

            Bentley.Interop.MicroStationDGN.Application App = mainAddin.Instance.GetComApp;
            Level lev = App.ActiveDesignFile.Levels.Find("Mønelinje");

            List<LineElement> Lines = new List<LineElement>();

            ElementScanCriteria rangeCriteria = new ElementScanCriteriaClass();
            ElementScanCriteria levelCriteria = new ElementScanCriteriaClass();

            range3d.High.Z += 100;
            rangeCriteria.IncludeOnlyWithinRange(range3d);

            if (lev == null)
                return;
            levelCriteria.IncludeLevel(lev);

            ElementEnumerator ee = App.ActiveModelReference.Scan(rangeCriteria);

            while (ee.MoveNext())
            {
                LineElement el;
                if (ee.Current.Level != null && ee.Current.Level.Name == "Mønelinje")
                {
                    el = ee.Current.AsLineElement();
                    el.EndPoint.ToString();
                    if (InLines == null)
                        InLines = new List<Point3d>();
                    foreach (Point3d p in el.GetVertices())
                        if (!InLines.Contains(p))
                            InLines.Add(p);
                    numberOfTopLines += 1;
                }
            }
        }

        public static void CreateAndReset()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;
            CreateBuilding();
            InLines = null;
            OutLines = null;
            App.ShowCommand("Add Outlines");
        }

        //public static void AddInlinesButton_Click()
        //{
        //    Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;
        //    if (OutLines == null)
        //            MessageBox.Show("Error: No outlines have been added!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            
        //    else if (inLines)
        //    {
        //        if (InLines == null)
        //        {
        //            MessageBox.Show("Warning: No inlines have been added!", "Info", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
        //        }
        //            inLines = false;
        //            CreateBuilding();
        //            MessageBox.Show("Building added.");
        //            InLines = null;
        //            OutLines = null;
        //            App.ShowCommand("Add Outlines");
        //    }
        //}

        private static void CreateBuilding()
        {
            if (InLines == null)
                InLines = OutLines;
            double[,] outL = RunCalc.convertPoints(OutLines.ToArray());
            double[,] inL = RunCalc.convertPoints(InLines.ToArray());
            bool computeFacadePoints = form.checkFacadePoints();

            // A building have to have clockwise defined points
            
            if (!isClockWise(outL))
            {
                outL = convertToClockwise(outL);
                if (outL.GetLength(0) == inL.GetLength(0))
                    inL = convertToClockwise(inL);
            }
            NoiseComputation.NoiseBuilding newBuilding;
            if (computeFacadePoints)
                newBuilding = new NoiseComputation.NoiseBuilding(outL, inL, computeFacadePoints);
            else
                newBuilding = new NoiseComputation.NoiseBuilding(outL, inL);
            MessageBox.Show("A building has been added:\n Outline is a " + typeOfOutline +", and there are "+numberOfTopLines+" top line(s).");
            RunCalc.AddBuilding(newBuilding);
        }

        private static bool isClockWise(double[,] lines)
        {
            int index = 0;
            double minX = double.PositiveInfinity;
            for (int i = 0; i< lines.GetLength(0); i++)
                if (lines[i, 0] < minX)
                {
                    minX = lines[i, 0];
                    index = i;
                }
            int nextIndex = index == lines.GetLength(0)-1 ? 1 : index + 1;
            int prevIndex = index == 0 ? lines.GetLength(0) - 2 : index - 1;
            if (lines[prevIndex, 1] < lines[nextIndex, 1])
                return true;
            else return false;
        }

        private static double[,] convertToClockwise(double[,] lines)
        {
            double[,] newOutL = new double[lines.GetLength(0), 3];
            int length = lines.GetLength(0);
            for (int i = 0; i < length; i++)
            {
                newOutL[length - 1 - i, 0] = lines[i, 0];
                newOutL[length - 1 - i, 1] = lines[i, 1];
                newOutL[length - 1 - i, 2] = lines[i, 2];
            }
            return newOutL;
        }

        #region ILocateCommandEvents Members

        public void Accept(Element Element, ref Point3d Point, Bentley.Interop.MicroStationDGN.View View)
        {
            AddElement(Element);
        }

        public void Cleanup()
        {
            form.DetachFromMicroStation();
        }

        public void Dynamics(ref Point3d Point, Bentley.Interop.MicroStationDGN.View View, MsdDrawingMode DrawMode)
        {
            
        }

        public void LocateFailed()
        {
            
        }

        public void LocateFilter(Element Element, ref Point3d Point, ref bool Accepted)
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            if (Element.Type == MsdElementType.LineString)
            {
                App.ShowPrompt("Add line");
                Accepted = true;
            }
            else if (Element.Type == MsdElementType.CellHeader)
            {
                App.ShowPrompt("Add Cell");
                Accepted = true;
            }
            else if (Element.Type == MsdElementType.Shape)
            {
                App.ShowPrompt("Add Shape");
                Accepted = true;
            }
            else
                Accepted = false;

            App.CommandState.StartDynamics();
        }

        public void LocateReset()
        {
            
        }

        public void Start()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            LocateCriteria lc;

            lc = App.CommandState.CreateLocateCriteria(true);
            App.CommandState.SetLocateCriteria(lc);

            App.CommandState.SetLocateCursor();

            App.CommandState.EnableAccuSnap();

            App.ShowPrompt("Select Line");
            App.ShowCommand("Add Outlines");

            form.AttachToToolSettings(mainAddin.Instance);
        }

        #endregion

    }
}
