﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Bentley.Interop.MicroStationDGN;
using System.Windows.Forms;

namespace ISY.CAD.Noise
{
    class CalcPointsMgr : ILocateCommandEvents
    {
        
        static List<Point3d> calcPoints;
   
        public static Point3d[] StartVertices()
        {
            Bentley.Interop.MicroStationDGN.Application App = mainAddin.Instance.GetComApp;

            List<LineElement> BuildingLines = scanLines(App, "StøyBygning");
            if (BuildingLines == null)
                return null;
            List<ShapeElement> BuildingShapes = new List<ShapeElement>();
            if (BuildingShapes == null)
                return null;
            if (BuildingLines.Count == 0)
                BuildingShapes = scanShapes(App, "StøyBygning");
            List<Point3d> tempPoints = new List<Point3d>();
            if (BuildingLines.Count != 0)
            {
                foreach (LineElement el in BuildingLines)
                {
                    foreach (Point3d p in el.GetVertices())
                    {
                        if (!tempPoints.Contains(p))
                            tempPoints.Add(p);
                    }
                }
            }
            else if (BuildingShapes.Count != 0)
            {
                foreach (LineElement el in BuildingShapes)
                {
                    foreach (Point3d p in el.GetVertices())
                    {
                        if (!tempPoints.Contains(p))
                            tempPoints.Add(p);
                    }
                }
            }

            Point3d[] BuildingPoints = new Point3d[tempPoints.Count];
            int i = 0;
            foreach (Point3d p in tempPoints)
            {
                BuildingPoints[i++] = p;
            }
            return BuildingPoints;
        }

        private static List<LineElement> scanLines(Bentley.Interop.MicroStationDGN.Application App, string level)
        {
            List<LineElement> Lines = new List<LineElement>();

            ElementScanCriteria sc = new ElementScanCriteriaClass();
            sc.ExcludeAllTypes();
            sc.IncludeType(MsdElementType.LineString);
            sc.IncludeType(MsdElementType.Line);
            sc.ExcludeAllLevels();

            Level lev = App.ActiveDesignFile.Levels.Find(level);
            if (lev == null)
                return null;
            sc.IncludeLevel(lev);

            ElementEnumerator ee = App.ActiveModelReference.Scan(sc);

            while (ee.MoveNext())
            {
                LineElement el;
                el = ee.Current.AsLineElement();

                Lines.Add(el);
            }
            return Lines;
        }

        private static List<ShapeElement> scanShapes(Bentley.Interop.MicroStationDGN.Application App, string level)
        {
            List<ShapeElement> Shapes = new List<ShapeElement>();

            ElementScanCriteria sc = new ElementScanCriteriaClass();
            sc.ExcludeAllTypes();
            sc.IncludeType(MsdElementType.Shape);
            sc.IncludeType(MsdElementType.ComplexShape);
            sc.ExcludeAllLevels();

            Level lev = App.ActiveDesignFile.Levels.Find(level);
            if (lev == null)
                return null;
            sc.IncludeLevel(lev);

            ElementEnumerator ee = App.ActiveModelReference.Scan(sc);

            while (ee.MoveNext())
            {
                ShapeElement el;
                el = ee.Current.AsShapeElement();

                Shapes.Add(el);
            }
            return Shapes;
        }

        internal static void StartCommand()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            CalcPointsMgr command = new CalcPointsMgr();
            App.CommandState.StartLocate(command);
            
            App.CommandState.StartDynamics();
            calcPoints = new List<Point3d>();
        }

        private void AddPointsFromElement(Element Element)
        {
            
            if ((Element.Type == MsdElementType.LineString) || (Element.Type == MsdElementType.Line))
            {
                LineElement el = Element.AsLineElement();
                foreach (Point3d p in el.GetVertices())
                    calcPoints.Add(p);
            }
            else if (Element.Type == MsdElementType.Shape)
            {
                ShapeElement el = Element.AsShapeElement();
                foreach (Point3d p in el.GetVertices())
                    calcPoints.Add(p);
            }
            else if (Element.Type == MsdElementType.Solid)
            {
                SmartSolidElement el = Element.AsSmartSolidElement;

                foreach (Point3d p in el.GetVertices())
                    calcPoints.Add(p);
            }
            else if (Element.Type == MsdElementType.CellHeader)
            {
                CellElement el = Element.AsCellElement();
                ElementEnumerator ee = el.GetSubElements();
                while (ee.MoveNext())
                {
                    ComplexShapeElement shape;
                    shape = ee.Current.AsComplexShapeElement();
                    ElementEnumerator ee2 = shape.GetSubElements();
                    while (ee2.MoveNext())
                    {
                        LineElement e;
                        e = ee2.Current.AsLineElement();
                        foreach (Point3d p in e.GetVertices())
                            calcPoints.Add(p);
                    }
                }
            }
        }

        #region ILocateCommandEvents Members

        public void Accept(Element Element, ref Point3d Point, Bentley.Interop.MicroStationDGN.View View)
        {
            AddPointsFromElement(Element);
            MessageBox.Show("Element added");
        }

        public void Cleanup()
        {
            RunCalc.AddBuildingCalculationPoints(calcPoints.ToArray());
        }

        public void Dynamics(ref Point3d Point, Bentley.Interop.MicroStationDGN.View View, MsdDrawingMode DrawMode)
        {
            
        }

        public void LocateFailed()
        {
            
        }

        public void LocateFilter(Element Element, ref Point3d Point, ref bool Accepted)
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            if ((Element.Type == MsdElementType.LineString) ||
                (Element.Type == MsdElementType.Line) ||
                (Element.Type == MsdElementType.Shape) ||
                (Element.Type == MsdElementType.Solid))
            {
                App.ShowPrompt("Add Points");
                Accepted = true;
            }
            else if (Element.Type == MsdElementType.CellHeader)
            {
                App.ShowPrompt("Add Points");
                Accepted = true;
            }
            else
            {
                App.ShowPrompt("Wrong element type");
                Accepted = false;
            }

            App.CommandState.StartDynamics();
        }

        public void LocateReset()
        {
            
        }

        public void Start()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            LocateCriteria lc;

            lc = App.CommandState.CreateLocateCriteria(true);
            App.CommandState.SetLocateCriteria(lc);

            App.CommandState.SetLocateCursor();

            App.CommandState.EnableAccuSnap();

            App.ShowPrompt("Select Point, Line, Shape or Solid");
            App.ShowCommand("Add calculation points");

        }
        #endregion
    }
}
