﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Bentley.Interop.MicroStationDGN;
using System.Runtime.InteropServices;

namespace ISY.CAD.Noise
{
    public class Grid
    {
        public Point3d[] Points;
        public float[,] zGrid;
        public double XStart;
        public double XEnd;
        public double YStart;
        public double YEnd;

        List<string> noIntersec = new List<string>();

        public Grid()
        {
            findGrid();
        }

        private void findGrid() {
            Bentley.Interop.MicroStationDGN.Application App = mainAddin.Instance.GetComApp;

            // search active model for linestring
            List<LineElement> EastLines = scan(App, "Secondary Grid Lines");
            List<LineElement> NorthLines = scan(App, "Primary Grid Lines");

            int xlinecount = NorthLines.Count;
            int ylinecount = EastLines.Count;

            Point3d[] points = new Point3d[xlinecount * ylinecount];
            
            zGrid = new float[xlinecount, ylinecount];

            Matrix3d mat = App.Matrix3dIdentity();
            int i = 0;
            double xstart = double.PositiveInfinity;
            double xend = double.NegativeInfinity;
            double ystart = double.PositiveInfinity;
            double yend = double.NegativeInfinity;

            Point3d intersectionPoint;
            int countIntPoints = 0;
            int x=0;
            int y=0;

            foreach (LineElement n_el in NorthLines)
            {
                y = 0;
                IntersectableElement n_inter = n_el.AsIntersectableElement();
                foreach (LineElement e_el in EastLines)
                {
                    Point3d[] p = n_inter.GetIntersectionPoints(e_el, ref mat);
                    if (p.Length == 0)
                    {
                        countIntPoints++;
                        intersectionPoint.X = intersectionPoint.Y = intersectionPoint.Z = 0;
                    }
                    else
                        intersectionPoint = p[0];
                    if (intersectionPoint.X != 0)
                    {
                        points[i++] = intersectionPoint;
                        zGrid[x, y] = (float)intersectionPoint.Z;
                        if (intersectionPoint.X < xstart)
                            xstart = intersectionPoint.X;
                        if (intersectionPoint.X > xend)
                            xend = intersectionPoint.X;
                        if (intersectionPoint.Y < ystart)
                            ystart = intersectionPoint.Y;
                        if (intersectionPoint.Y > yend)
                            yend = intersectionPoint.Y;
                    }
                    y++;
                }
                x++;
            }
            Points = points;
            XStart = xstart;
            XEnd = xend;
            YStart = ystart;
            YEnd = yend;
            if (countIntPoints > 0)
                MessageBox.Show(countIntPoints + " of " + x*y + " points does not have an intersection. \n More information may be found in IntersectionPoints.txt");
        }

        private static List<LineElement> scan(Bentley.Interop.MicroStationDGN.Application App, string level) 
        {
            List<LineElement> Lines = new List<LineElement>();

            ElementScanCriteria sc = new ElementScanCriteriaClass();
            sc.ExcludeAllTypes();
            sc.IncludeType(MsdElementType.LineString);
            sc.ExcludeAllLevels();

            Level lev = App.ActiveDesignFile.Levels.Find(level);
            sc.IncludeLevel(lev);

            ElementEnumerator ee = App.ActiveModelReference.Scan(sc);

            while (ee.MoveNext())
            {
                LineElement el;
                el = ee.Current.AsLineElement();
                Lines.Add(el);
            }
            return Lines;
        }
    }
}

