﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using Bentley.Interop.MicroStationDGN;

namespace ISY.CAD.Noise
{
    // Dersom man skal kunne kjøre på PowerDraft må Password="dlm kode" være med
    [Bentley.MicroStation.AddInAttribute(KeyinTree = "ISY.CAD.Noise.commands.xml",
                     MdlTaskID = "Noise")]

    public sealed class mainAddin : Bentley.MicroStation.AddIn
    {
        static mainAddin _app = null;

        /// <summary>Private constructor required for all AddIn classes derived from
        /// Bentley.MicroStation.AddIn.</summary>
        private mainAddin(System.IntPtr mdlDesc)
            : base(mdlDesc)
        {
            _app = this;
        }

        public Bentley.Interop.MicroStationDGN.Application GetComApp
        {
            get
            {
                return Bentley.MicroStation.InteropServices.Utilities.ComApp;
            }
        }

        internal static mainAddin Instance
        {
            get { return _app; }
        }
        
        protected override int Run(System.String[] commandLine)
        {

            //  Register reload and unload events
            this.ReloadEvent += new ReloadEventHandler(mainAddin_ReloadEvent);
            //this.UnloadingEvent += new UnloadingEventHandler(mainAddin_UnloadEvent);

            return 0;
        }

        /// <summary>Handles MDL LOAD requests after the application has been loaded.
        /// </summary>
        private void mainAddin_ReloadEvent(Bentley.MicroStation.AddIn sender,
                        ReloadEventArgs eventArgs)
        {
            // Add reload methods
            MessageBox.Show("Reload");
        }

    }  // End of mainAddin class

    /// <summary>Class used for running key-ins.  The key-ins
    /// XML file provides the name of the class and the methods.
    /// </summary>
    public class KeyinCommands
    {
        /// <summary>
        /// This is the main method that checks for noise sources, adds
        /// measuring points and run the calculation
        /// </summary>
        public static void RunCalculation(System.String unparsed)
        {
            RunCalc.StartCommand();
        }

        /// <summary>
        /// This method should tag a linestring as a noise source road 
        /// and if needed, add attributes to replace the default values
        /// </summary>
        public static void AddNoiseSourceRoad(System.String unparsed)
        {
            RoadMgr.StartCommand();
        }

        /// <summary>
        /// This method should tag points, linestrings, shapes or solids
        /// as specific measuring points to be calculated 
        /// </summary>
        public static void AddMeasuringPoints(System.String unparsed)
        {
            CalcPointsMgr.StartCommand();
        }

        /// <summary>
        /// Convert mesh to grid
        /// </summary>
        public static void Mesh2Grid(System.String unparsed)
        {
            Mesh2GridMgr.StartCommand();
        }

        public static void AddBuilding(System.String unparsed)
        {
            AddBuildingMgr.StartCommand();
        }

        public static void AddMound(System.String unparsed)
        {
            MoundMgr.StartCommand();
        }

        public static void TestProgram(System.String unparsed)
        {
            NoiseComputation.Program.Main(new string[0]);

            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;


            if (NoiseComputation.Program.contours != null && NoiseComputation.Program.contours.Count > 0)
            {
                NoiseComputation.NoiseContour[][] newC = NoiseComputation.Program.contours.ToArray();
                RunCalc.drawContours(newC[0], App, "Lden", 418208.5783, 6981471.4255, 418543.5783, 6981746.4255);
                RunCalc.drawContours(newC[1], App, "LdenDetail", 418208.5783, 6981471.4255, 418543.5783, 6981746.4255);
                //MessageBox.Show("Contours created!");
            }
        }

        public static void AddNoiseSourceIndustry(System.String unparsed)
        {
            IndustryNoiseMgr.StartCommand();
        }
    }  // End of KeyinCommands
}  // End of the namespace

