﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Bentley.Interop.MicroStationDGN;
using System.Windows.Forms;


namespace ISY.CAD.Noise
{

    class Mesh2GridMgr : ILocateCommandEvents
    {
        public void Mesh2Grid(Element elMesh)
        {
            int gridSize = 5;

            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            // sletter "ISY CAD Noise Calculation" om den finnes fra før
            deleteModel(App, "ISY CAD Noise Calculation");

            // lager mesh i ny modell dersom ønskelig:

            // creating new model, "ISY CAD Noise Calculation" with the mesh
            ModelReference model = null;
            try
            {
                ModelReference modRef = App.ActiveDesignFile.DefaultModelReference;
                model = App.ActiveDesignFile.Models.Add(modRef, "ISY CAD Noise Calculation", "Used for Noise calculation", MsdModelType.Default, true);
            }
            catch
            {
                return;
            }

            model.AddElement(elMesh);
            model.Activate();

            App.ActiveModelReference.UnselectAllElements();

            // Find start and end points
            Point3d startPnt = findStartPnt(elMesh);
            Point3d endPnt = findEndPnt(elMesh, true);

            double yLength = Math.Abs(endPnt.Y - startPnt.Y);

            Level lev = createAndSetLevel(App,"Primary Grid Lines");
            addLines(App, elMesh, true, gridSize, startPnt, endPnt, lev); 

            startPnt = findStartPnt(elMesh);
            endPnt = findEndPnt(elMesh, false);

            double xLength = Math.Abs(endPnt.X - startPnt.X);

            lev = createAndSetLevel(App, "Secondary Grid Lines");
            addLines(App, elMesh, false, gridSize, startPnt, endPnt, lev);

            // sender stencil kommando, bør også tvinge riktig toolsettings her
            App.CadInputQueue.SendCommand("MDL KEYIN CIVVIS civvis imprint", true);
            // setter ut punkt to ganger for at det skal fungere, bruker min punktet på range til mesh
            Point3d pnt;
            pnt = elMesh.Range.Low;
            App.CadInputQueue.SendDataPoint(ref pnt, 1, 0);
            App.CadInputQueue.SendDataPoint(ref pnt, 1, 0);

            // sletter mesh dersom ønskelig
            App.ActiveModelReference.RemoveElement(elMesh);
            //return;
            removeBottomGridLines(App);

            // Group together lines
            List<LineElement> Lines = scanLines(App);

            // Find start and end points
            double xstart = double.PositiveInfinity;
            double xend = double.NegativeInfinity;
            double ystart = double.PositiveInfinity;
            double yend = double.NegativeInfinity;
            foreach (LineElement el in Lines)
            {
                if (el.StartPoint.X < xstart)
                    xstart = el.StartPoint.X;
                if (el.StartPoint.Y < ystart)
                    ystart = el.StartPoint.Y;
                if (el.EndPoint.X > xend)
                    xend = el.EndPoint.X;
                if (el.EndPoint.Y > yend)
                    yend = el.EndPoint.Y;
            }

            // Merge lines by using Point3d
            List<LineElement> TooShortLines = mergeLines(App, "Primary Grid Lines", true, Lines);

            TooShortLines = mergeLines(App, "Secondary Grid Lines", false, TooShortLines);

            List<LineElement> FinishedLines = scanLines(App);

            FinishedLines = elongateLines(FinishedLines, xstart, ystart, xend, yend);


            App.ActiveDesignFile.Models.Delete(model);
            App.ActiveModelReference.Activate();

            foreach (LineElement el in FinishedLines)
                App.ActiveModelReference.AddElement(el);
        }

        private static void deleteModel(Bentley.Interop.MicroStationDGN.Application App, string model)
        {
            ModelReference modelToDelete = null;
            foreach (ModelReference m in App.ActiveDesignFile.Models)
            {
                if (m.Name == model)
                    modelToDelete = m;
            }
            if (modelToDelete != null)
                App.ActiveDesignFile.Models.Delete(modelToDelete);
        }

        private Point3d findStartPnt(Element elMesh)
        {
            Point3d startPnt = elMesh.Range.Low;
            startPnt.Z = 0;
            return startPnt;
        }

        private Point3d findEndPnt(Element elMesh, bool isX)
        {
            Point3d endPnt = elMesh.Range.High;
            if (isX)
                endPnt.X = elMesh.Range.Low.X;
            else
                endPnt.Y = elMesh.Range.Low.Y;
            endPnt.Z = 0;
            return endPnt;
        }

        private Level createAndSetLevel(Bentley.Interop.MicroStationDGN.Application App, string p)
        {
            Level lev;
            if ((lev = App.ActiveDesignFile.Levels.Find(p)) == null)
            {
                App.ActiveDesignFile.AddNewLevel(p);
                App.ActiveDesignFile.RewriteLevels();
                lev = App.ActiveDesignFile.Levels.Find(p);
            }
            return lev;
        }

        private void addLines(Bentley.Interop.MicroStationDGN.Application App, Element elMesh, bool isX, int gridSize, Point3d startPnt, Point3d endPnt, Level lev)
        {
            int numLines;
            if (isX)
            {
                numLines = (int)((elMesh.Range.High.X - elMesh.Range.Low.X) / gridSize) + 1;
                for (int i = 0; i < numLines; i++)
                {
                    // Adding Primary Grid Lines
                    startPnt.X += gridSize;
                    endPnt.X += gridSize;
                    LineElement lEl = App.CreateLineElement2(null, ref startPnt, ref endPnt);
                    lEl.let_Level(lev);
                    App.ActiveModelReference.AddElement(lEl);
                    App.ActiveModelReference.SelectElement(lEl, true);
                }
            }
            else
            {
                numLines = (int)((elMesh.Range.High.Y - elMesh.Range.Low.Y) / gridSize) + 1;
                for (int i = 0; i < numLines; i++)
                {
                    // Adding Secondary Grid Lines
                    startPnt.Y += gridSize;
                    endPnt.Y += gridSize;
                    LineElement lEl = App.CreateLineElement2(null, ref startPnt, ref endPnt);
                    lEl.let_Level(lev);
                    App.ActiveModelReference.AddElement(lEl);
                    App.ActiveModelReference.SelectElement(lEl, true);
                }
            }

        }

        private void removeBottomGridLines(Bentley.Interop.MicroStationDGN.Application App)
        {
            List<LineElement> Lines = scanForLines(App);
            foreach (LineElement l in Lines)
                App.ActiveModelReference.RemoveElement(l);
        }

        private static List<LineElement> scanForLines(Bentley.Interop.MicroStationDGN.Application App)
        {
            List<LineElement> Lines = new List<LineElement>();

            ElementScanCriteria sc = new ElementScanCriteriaClass();
            sc.ExcludeAllTypes();
            sc.IncludeType(MsdElementType.Line);

            ElementEnumerator ee = App.ActiveModelReference.Scan(sc);

            while (ee.MoveNext())
            {
                LineElement el;
                el = ee.Current.AsLineElement();
                Lines.Add(el);
            }
            return Lines;
        }

        private static List<LineElement> scanLines(Bentley.Interop.MicroStationDGN.Application App)
        {
            List<LineElement> Lines = new List<LineElement>();

            ElementScanCriteria sc = new ElementScanCriteriaClass();
            sc.ExcludeAllTypes();
            sc.IncludeType(MsdElementType.LineString);
            sc.ExcludeAllLevels();
            
            sc.IncludeLevel(App.ActiveDesignFile.Levels.Find("Primary Grid Lines"));
            sc.IncludeLevel(App.ActiveDesignFile.Levels.Find("Secondary Grid Lines"));

            ElementEnumerator ee = App.ActiveModelReference.Scan(sc);

            while (ee.MoveNext())
            {
                LineElement el;
                el = ee.Current.AsLineElement();
                Lines.Add(el);
            }
            return Lines;
        }

        private List<LineElement> findShortLines(List<LineElement> Lines, double xLength, double yLength)
        {
            List<LineElement> TooShortLines = new List<LineElement>();
            foreach (LineElement el in Lines)
            {
                if ((el.Level.Name == "Primary Grid Lines" && el.Length <= yLength) ||
                    (el.Level.Name == "Secondary Grid Lines" && el.Length <= xLength))
               
                    TooShortLines.Add(el);
            }
            return TooShortLines;
        }

        private List<LineElement> mergeLines(Bentley.Interop.MicroStationDGN.Application App, string levelName, bool isX, List<LineElement> TooShortLines)
        {
            bool first = true;
            List<Point3d> linepoints = new List<Point3d>();
            LineElement[] ArrayTooShortLines = TooShortLines.ToArray();

            if (isX)
                SortLinesByX(ArrayTooShortLines);
            else
                SortLinesByY(ArrayTooShortLines);

            Point3d[] lineArray;
            Point3d startPoint;
            startPoint.X = startPoint.Y = 0;

            double elStartCoord;
            double startCoord;
            
            foreach (LineElement el in ArrayTooShortLines)
            {
                if (el.Level.Name == levelName)
                {
                    if (first)
                    {
                        startPoint = el.StartPoint;

                        first = false;
                    }

                    if (isX)
                    {
                        elStartCoord = el.StartPoint.X;
                        startCoord = startPoint.X;
                    }
                    else
                    {
                        elStartCoord = el.StartPoint.Y;
                        startCoord = startPoint.Y;
                    }

                    // If the element is on the current line
                    if ((elStartCoord >= Math.Floor(startCoord)) && (elStartCoord <= Math.Ceiling(startCoord))) // add check for end of lines
                        addPoints(linepoints, el);

                    else 
                    {
                        lineArray = linepoints.ToArray();
                        if (isX)
                            SortPointsByY(lineArray);
                        else
                            SortPointsByX(lineArray);
                        
                        LineElement newLine = App.CreateLineElement1(null, lineArray);

                        newLine.Level = App.ActiveDesignFile.Levels.Find(levelName);

                        App.ActiveModelReference.AddElement(newLine);
                        App.ActiveModelReference.SelectElement(newLine, true);

                        linepoints.Clear();
                        startPoint = el.StartPoint;
                        addPoints(linepoints, el);

                    }
                    TooShortLines.Remove(el);
                    App.ActiveModelReference.RemoveElement(el);
                }
            }
            return TooShortLines;
        }

        private static void SortLinesByX(LineElement[] ArrayTooShortLines)
        {

            Array.Sort(ArrayTooShortLines, delegate(LineElement user1, LineElement user2)
                {
                    int result = user1.StartPoint.X.CompareTo(user2.StartPoint.X);
                    return result != 0 ? result : user1.StartPoint.Y.CompareTo(user2.StartPoint.Y);
                });
        }

        private static void SortLinesByY(LineElement[] ArrayTooShortLines)
        {
            Array.Sort(ArrayTooShortLines, delegate(LineElement user1, LineElement user2)
            {
                int result = user1.StartPoint.Y.CompareTo(user2.StartPoint.Y);
                return result != 0 ? result : user1.StartPoint.X.CompareTo(user2.StartPoint.X);
            });
        }

        private static void addPoints(List<Point3d> linepoints, LineElement el)
        {
            foreach (Point3d point in el.GetVertices())
                linepoints.Add(point);
        }

        private static void SortPointsByX(Point3d[] lineArray)
        {
            Array.Sort(lineArray, delegate(Point3d user1, Point3d user2)
            {
                return user1.X.CompareTo(user2.X);
            });
        }

        private static void SortPointsByY(Point3d[] lineArray)
        {
            Array.Sort(lineArray, delegate(Point3d user1, Point3d user2)
            {
                return user1.Y.CompareTo(user2.Y);
            });
        }

        private List<LineElement> elongateLines(List<LineElement> Lines, double xstart, double ystart, double xend, double yend)
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            LineElement[] temp = Lines.ToArray();
            foreach (LineElement el in temp)
            {
                Point3d startPoint = new Point3d();
                Point3d endPoint = new Point3d();
                if ((el.StartPoint.X <= el.EndPoint.X + 1) && (el.StartPoint.X >= el.EndPoint.X - 1))
                {
                    startPoint.X = endPoint.X = el.StartPoint.X;
                    startPoint.Y = ystart;
                    //endPoint.X = el.EndPoint.X; 
                    endPoint.Y = yend;
                }
                else if ((el.StartPoint.Y <= el.EndPoint.Y + 1) && (el.StartPoint.Y >= el.EndPoint.Y - 1))
                {
                    startPoint.X = xstart;
                    startPoint.Y = el.StartPoint.Y;
                    endPoint.X = xend;
                    endPoint.Y = el.EndPoint.Y;
                }
                startPoint.Z = el.StartPoint.Z;
                endPoint.Z = el.EndPoint.Z;
                List<Point3d> points = new List<Point3d>();
                points.Add(startPoint);
                foreach (Point3d p in el.GetVertices())
                    points.Add(p);
                points.Add(endPoint);

                LineElement newLine = App.CreateLineElement1(null, points.ToArray());
                newLine.Level = App.ActiveDesignFile.Levels.Find(el.Level.Name);

                Lines.Remove(el);
                Lines.Add(newLine); ;
            }
            return Lines;

        }

        #region ILocateCommandEvents Members

        public void Accept(Element Element, ref Point3d Point, Bentley.Interop.MicroStationDGN.View View)
        {
            Mesh2Grid(Element);
        }

        public void Cleanup()
        {

        }

        public void Dynamics(ref Point3d Point, Bentley.Interop.MicroStationDGN.View View, MsdDrawingMode DrawMode)
        {
        }

        public void LocateFailed()
        {

        }

        public void LocateFilter(Element Element, ref Point3d Point, ref bool Accepted)
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            if (Element.Type == MsdElementType.MeshHeader)
            {
                App.ShowPrompt("Data Point to create grid");
                Accepted = true;
            }
            else
                Accepted = false;

            //App.CommandState.StartDynamics();
            
        }

        public void LocateReset()
        {

        }

        public void Start()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            LocateCriteria lc;

            lc = App.CommandState.CreateLocateCriteria(true);
            App.CommandState.SetLocateCriteria(lc);

            App.CommandState.SetLocateCursor();

            App.CommandState.EnableAccuSnap();

            App.ShowPrompt("Select Mesh");
            App.ShowCommand("Create grid from mesh");
        }

        #endregion

        public static void StartCommand()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            Mesh2GridMgr command = new Mesh2GridMgr();
            App.CommandState.StartLocate(command);
            App.CommandState.StartDynamics();
        }

        
    }
}
