﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
//using SoundKernel;
using Bentley.Interop.MicroStationDGN;
using System.Windows.Forms;

namespace ISY.CAD.Noise
{
    class MoundMgr : ILocateCommandEvents
    {
        private MoundForm form;

        public MoundMgr()
        {
            form = new MoundForm();
        }

        public void CreateMound(Element e)
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            LineElement el = e.AsLineElement();
            Level lev = createAndSetLevel(App, "Voll");
            el.let_Level(lev);
            el.Rewrite();
            
            Point3d[] MoundPoints = el.GetVertices();
            MessageBox.Show("Number of vertices found: " + MoundPoints.Length);
       
            double[,] points = RunCalc.convertPoints(MoundPoints);
            float height = form.GetHeight();
            float rightHeight = 0;
            float topWidth = 0;
            float bottomWidth = 0;
            if (form.GetType() == "wall")
            {
                height = form.GetLeftHeight();
                rightHeight = form.GetRightHeight();
                topWidth = form.GetTopWidth();
                bottomWidth = form.GetTopWidth() + 0.2f;
            }
            else if (form.GetType() == "mound")
            {
                height = form.GetLeftHeight();
                rightHeight = form.GetRightHeight();
                topWidth = form.GetTopWidth();
                bottomWidth = form.GetBottomWidth();
            }

            NoiseComputation.NoiseMound mound = new NoiseComputation.NoiseMound(points, height, topWidth, bottomWidth, rightHeight);
     
            for (int i = 0; i < MoundPoints.Length; i++ )
                MoundPoints[i].Z += height;

            RunCalc.AddMound(mound);
            drawMound(MoundPoints, App);
            MessageBox.Show("Mound created");

        }

        private void drawMound(Point3d[] moundPoints, Bentley.Interop.MicroStationDGN.Application App)
        {
            Level lev = createAndSetLevel(App, "VollHøyde");

            LineElement newLine = App.CreateLineElement1(null, moundPoints);
            newLine.let_Level(lev);
            App.ActiveModelReference.AddElement(newLine);
            App.ActiveModelReference.SelectElement(newLine, true);
        }

        private Level createAndSetLevel(Bentley.Interop.MicroStationDGN.Application App, string p)
        {
            Level lev;
            if ((lev = App.ActiveDesignFile.Levels.Find(p)) == null)
            {
                App.ActiveDesignFile.AddNewLevel(p);
                App.ActiveDesignFile.RewriteLevels();
                lev = App.ActiveDesignFile.Levels.Find(p);
            }
            return lev;
        }


        #region ILocateCommandEvents Members

        public void Accept(Element Element, ref Point3d Point, Bentley.Interop.MicroStationDGN.View View)
        {
            CreateMound(Element);
        }

        public void Cleanup()
        {
            form.DetachFromMicroStation();
        }

        public void Dynamics(ref Point3d Point, Bentley.Interop.MicroStationDGN.View View, MsdDrawingMode DrawMode)
        {
        }

        public void LocateFailed()
        {
            
        }

        public void LocateFilter(Element Element, ref Point3d Point, ref bool Accepted)
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            if (Element.Type == MsdElementType.LineString || Element.Type == MsdElementType.Line)
            {
                App.ShowPrompt("Create base line for mound");
                Accepted = true;
            }
            else
                Accepted = false;

            App.CommandState.StartDynamics();
            
        }

        public void LocateReset()
        {

        }

        public void Start()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            

            LocateCriteria lc;

            lc = App.CommandState.CreateLocateCriteria(true);
            App.CommandState.SetLocateCriteria(lc);

            App.CommandState.SetLocateCursor();

            App.CommandState.EnableAccuSnap();

            App.ShowPrompt("Select Line");
            App.ShowCommand("Create mound from line");

            form.AttachToToolSettings(mainAddin.Instance);

        }

        #endregion

        public static void StartCommand()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            MoundMgr command = new MoundMgr();
            App.CommandState.StartLocate(command);
            App.CommandState.StartDynamics();

 
        }
    }
}
