﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Globalization;

namespace ISY.CAD.Noise
{
    public partial class RoadForm : Bentley.MicroStation.WinForms.Adapter
    {
        public RoadForm()
        {
            InitializeComponent();
            buttonDefineDistribution.Enabled = false;
            comboBoxTrafficType.SelectedIndex = 0;
        }

        public float GetWidth()
        {
            try
            {
                return (float)double.Parse(textBoxWidth.Text, CultureInfo.InvariantCulture);
            }
            catch (Exception ee)
            {
                MessageBox.Show("Error!" + ee.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return -1;
            }
        }

        public float GetVelocity()
        {
            float kmPerH;
            try
            {
                kmPerH = (float)double.Parse(textBoxVelocity.Text, CultureInfo.InvariantCulture);
            }
            catch (Exception ee)
            {
                MessageBox.Show("Error!" + ee.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return -1;
            }
            // Convert from km/h to m/s
            return (float)kmPerH*1000/3600;
        }

        public string GetTraffic()
        {
            return comboBoxTrafficType.SelectedItem.ToString();
        }

        public float GetYDT()
        {
            string s = textBoxYDT.Text;
            if (s.Contains(","))
            {
                MessageBox.Show("Error! Use point as decimal separator.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return -1;
            }

            try
            {
                return (float)double.Parse(s, CultureInfo.InvariantCulture);
            }
            catch (Exception ee)
            {
                MessageBox.Show("Error!" + ee.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return -1;
            }
        }

        public bool IsTunnel()
        {
            return checkBoxTunnel.Checked;
        }

        public bool IsBridge()
        {
            return checkBoxBridge.Checked;
        }

        public string getTrafficDistributionType()
        {
            if (radioButtonGeneral.Checked)
                return "General";
            else if (radioButtonWork.Checked)
                return "Work";
            else return "UserDefined";
        }


        private void buttonDefine_Click(object sender, EventArgs e)
        {
            Form specForm = new NoiseSourceSpecForm();
            specForm.Show();
        }

        private void comboBoxTrafficType_SelectedIndexChanged(object sender, EventArgs e)
        {
            if ((string)comboBoxTrafficType.SelectedItem == "User Defined")
                buttonDefine.Enabled = true;
            else
            {
                buttonDefine.Enabled = false;
                NoiseSourceSpecForm.freqSpectra = null;
            }
        }

        private void radioButtonUserDefined_CheckedChanged(object sender, EventArgs e)
        {
            if (radioButtonUserDefined.Checked)
            {
                NoiseTimeDistributionForm.distribution = NoiseTimeDistributionForm.defaultAadtDistribution;
                buttonDefineDistribution.Enabled = true;
                this.textBoxYDT.Text = NoiseTimeDistributionForm.defaultAadtDistribution[9].ToString();
            }
            else
                buttonDefineDistribution.Enabled = false;
        }

        private void radioButtonWork_CheckedChanged(object sender, EventArgs e)
        {
            if (radioButtonUserDefined.Checked)
                buttonDefineDistribution.Enabled = true;
            else
                buttonDefineDistribution.Enabled = false;
        }

        private void radioButtonGeneral_CheckedChanged(object sender, EventArgs e)
        {
            if (radioButtonUserDefined.Checked)
                buttonDefineDistribution.Enabled = true;
            else
                buttonDefineDistribution.Enabled = false;
        }

        private void buttonDefineDistribution_Click(object sender, EventArgs e)
        {
            NoiseTimeDistributionForm.amountunit = "aadt";
            NoiseTimeDistributionForm distributionForm = new NoiseTimeDistributionForm();
            distributionForm.Show();

        }

        //public static void SetAadt(double newAdt)
        //{
        //   textBoxYDT.Text = newAdt.ToString();
        //}
    }
}
