﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Collections;
using System.ComponentModel;

using Bentley.Interop.MicroStationDGN;
using NC = NoiseComputation;
using System.Drawing;


namespace ISY.CAD.Noise
{

    class RunCalc : ILocateCommandEvents
    {
        private static CalcForm form;

        static int numCalcPoints = 0;

        public static Grid baseGrid;

        public static NC.NoiseMound[] mounds;
        public static NC.NoiseRoad[] roads;
        static NC.NoiseSource[] sources;
        static Point3d[] calculationPoints;
        static NC.NoiseBuilding[] buildings;

        static List<string> lines = new List<string>();

        private static BackgroundWorker bw;
        private static ProgressBarForm pBarForm;

        public static void StartCalculation()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;
            
            if (baseGrid == null)
                baseGrid = new Grid();
            
            CreateGrid(baseGrid);

            // Find calculation points:
            findAndAddCalculationPoints(App);

            // Add computation units:
            addComputationUnits();


            NC.ComputationMgr.noiseMounds = mounds;
            NC.ComputationMgr.noiseRoads = roads;
            NC.ComputationMgr.noiseSources = sources;
            NC.ComputationMgr.noiseBuildings = buildings;

            bw = new BackgroundWorker();
            bw.WorkerSupportsCancellation = true;
            bw.DoWork += new DoWorkEventHandler(bw_DoWork);
            bw.RunWorkerCompleted += new RunWorkerCompletedEventHandler(bw_RunWorkerCompleted);

            if (bw.IsBusy != true)
            {
                bw.RunWorkerAsync();
            }

            pBarForm = new ProgressBarForm();
            pBarForm.Show();

        }

        private static void addComputationUnits()
        {
            NC.ComputationMgr.computationUnits = form.checkComputationUnits();
        }

        private static void CreateGrid(Grid baseGrid)
        {
            NC.ComputationMgr.x1 = baseGrid.XStart;
            NC.ComputationMgr.y1 = baseGrid.YStart;
            NC.ComputationMgr.x2 = baseGrid.XEnd;
            NC.ComputationMgr.y2 = baseGrid.YEnd;

            NC.ComputationMgr.zgrid = baseGrid.zGrid;
        }

        private static void findAndAddCalculationPoints(Bentley.Interop.MicroStationDGN.Application App)
        {
            //calculateAllPoints = form.runAllPoints();

            Point3d[] cp = calculationPoints;
            //if (calculateAllPoints)
            //    cp = baseGrid.Points;

            //if (form.checkForBuildings())
            //    cp = findAndAddBuildingPoints(cp);

            //if (calculationPoints != null)
            //cp = calculationPoints;

            if ((cp != null) && (cp.Length >= 0))
                numCalcPoints = cp.Length;

            //NC.ComputationMgr.numUnits = 5;

            if (numCalcPoints > 0)
                addCalculationPoints(cp);
            App.MessageCenter.StatusMessage = numCalcPoints + " calculation points";
        }

        private static Point3d[] findAndAddBuildingPoints(Point3d[] cp)
        {
            Point3d[] BuildingPoints = CalcPointsMgr.StartVertices();
            if (BuildingPoints == null)
            {
                MessageBox.Show("No existing calculation points found");
                return null;
            }
            if (cp != null)
                cp = mergeArrays(cp, BuildingPoints);
            else
                cp = BuildingPoints;
            return cp;
        }

        private static Point3d[] mergeArrays(Point3d[] cp, Point3d[] calculationPoints)
        {
            if (cp == null)
                return calculationPoints;
            if (calculationPoints == null)
                return cp;
            Point3d[] newPoints = new Point3d[cp.Length + calculationPoints.Length];

            string[] pointLines = new string[newPoints.Length];

            for (int i = 0; i < newPoints.Length; i++)
            {
                if (i < cp.Length)
                    newPoints[i] = cp[i];
                else
                    newPoints[i] = calculationPoints[i - cp.Length];
                pointLines[i] = "Point " + i + "\t X: \t" + newPoints[i].X.ToString() +
                    "\t Y: \t" + newPoints[i].Y.ToString() + "\t Z: \t" + newPoints[i].Z.ToString();
            }
            System.IO.File.WriteAllLines(@"C:\TEMP\WritePoints.txt", pointLines);
            return newPoints;
        }

        private static void addCalculationPoints(Point3d[] calcP)
        {
            double[,] points = new double[calcP.Length, 3];
            for (int i = 0; i < numCalcPoints; i++)
            {
                points[i, 0] = calcP[i].X;
                points[i, 1] = calcP[i].Y;
                points[i, 2] = calcP[i].Z;
            }
            NC.ComputationMgr.calcPoints = points;
            
        }

        //private static NC.NoiseRoad findRoadNoiseSource()
        //{
        //    Point3d[] RoadPoints = RoadMgr.FindRoadPoints();
        //    //bool isAlreadyInRoads = false;
        //    //List<Point3d> existingRoadPoints = new List<Point3d>();
        //    //if (RunCalc.roads != null)
        //    //    foreach (Point3d p in RoadPoints)
        //    //        foreach (NC.NoiseRoad r in RunCalc.roads)
        //    //for (int i = 0; i < r.numberOfPoints; i++)
        //    //    if ((p.X == r.pGon[i, 0]) && (p.Y == r.pGon[i, 1]) && (p.Z == r.hs[i]))
        //    //    {
        //    //        isAlreadyInRoads = true;
        //    //        existingRoadPoints.Add(p);
        //    //    }
        //    //if (isAlreadyInRoads && (existingRoadPoints.Count == RoadPoints.Length))
        //    //{
        //    //    MessageBox.Show("The road already exist");
        //    //    return null;
        //    //}

        //    if (RoadPoints == null)
        //        return null;
        //    // This method should probably be deleted over time
        //    return new NC.NoiseRoad(convertPoints(RoadPoints), 10F, "Normal", 50, 2000, false, false, true);
        //}

        public static double[,] convertPoints(Point3d[] RoadPoints)
        {
            double[,] points = new double[RoadPoints.Length, 3];
            for (int i = 0; i < RoadPoints.Length; i++)
            {
                points[i, 0] = RoadPoints[i].X;
                points[i, 1] = RoadPoints[i].Y;
                points[i, 2] = RoadPoints[i].Z;
            }
            return points;
        }

        public static void AddRoad(NC.NoiseRoad r)
        {
            if (roads == null)
            {
                roads = new NC.NoiseRoad[1];
                roads[0] = r;
            }
            else
            {
                NC.NoiseRoad[] newRoads = new NC.NoiseRoad[roads.Length + 1];
                for (int i = 0; i < newRoads.Length - 1; i++)
                {
                    newRoads[i] = roads[i];
                }
                newRoads[newRoads.Length - 1] = r;
                roads = newRoads;
            }
        }

        private static Level createAndSetLevel(Bentley.Interop.MicroStationDGN.Application App, string p)
        {
            Level lev;
            if ((lev = App.ActiveDesignFile.Levels.Find(p)) == null)
            {
                App.ActiveDesignFile.AddNewLevel(p);
                App.ActiveDesignFile.RewriteLevels();
                lev = App.ActiveDesignFile.Levels.Find(p);
            }
            return lev;
        }


        //private static double findMeanZ(Grid baseGrid)
        //{
        //    float sum = 0;
        //    for (int i = 0; i < baseGrid.XLineCount; i++)
        //        for (int j = 0; j<baseGrid.YLineCount; j++)
        //            sum += baseGrid.zGrid[i,j];
        //    return (double)sum / baseGrid.zGrid.Length;
        //}
     
        public static void AddNoiseSource(NC.NoiseSource s)
        {
            if (sources == null)
            {
                sources = new NC.NoiseSource[1];
                sources[0] = s;
            }
            else
            {
                NC.NoiseSource[] newSources = new NC.NoiseSource[sources.Length + 1];
                for (int i = 0; i < newSources.Length - 1; i++)
                    newSources[i] = sources[i];
                newSources[newSources.Length - 1] = s;
                sources = newSources;
            }
        }

        public static void AddMound(NC.NoiseMound m)
        {
            if (mounds == null)
            {
                mounds = new NC.NoiseMound[1];
                mounds[0] = m;
            }
            else
            {
                NC.NoiseMound[] temp = mounds;
                NC.NoiseMound[] newMounds = new NC.NoiseMound[temp.Length + 1];
                for (int i = 0; i < newMounds.Length - 1; i++)
                    newMounds[i] = mounds[i];
                newMounds[newMounds.Length - 1] = m;
                mounds = newMounds;
            }
        }

        public static void AddBuilding(NC.NoiseBuilding newBuilding)
        {
            if (buildings == null)
            {
                buildings = new NC.NoiseBuilding[1];
                buildings[0] = newBuilding;
            }
            else
            {
                NC.NoiseBuilding[] temp = buildings;
                NC.NoiseBuilding[] newBuildings = new NC.NoiseBuilding[temp.Length + 1];
                for (int i = 0; i < newBuildings.Length - 1; i++)
                    newBuildings[i] = buildings[i];
                newBuildings[newBuildings.Length - 1] = newBuilding;
                buildings = newBuildings;
            }
        }

        public static void AddBuildingCalculationPoints(Point3d[] p)
        {
            if (calculationPoints == null)
                calculationPoints = p;
            else
            {
                Point3d[] temp = calculationPoints;
                Point3d[] newPoints = new Point3d[calculationPoints.Length + p.Length];
                for (int i = 0; i < newPoints.Length; i++)
                {
                    if (i < calculationPoints.Length)
                        newPoints[i] = calculationPoints[i];
                    else
                        newPoints[i] = p[i - calculationPoints.Length];
                }
                calculationPoints = newPoints;
            }
        }

        internal static void StartCommand()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            form = new CalcForm();
            RunCalc command = new RunCalc();
            App.CommandState.StartLocate(command);
        }

        internal static void CreateContours()
        {
            if (!NC.ComputationMgr.successfullyRun)
            {
                MessageBox.Show("The computation need to be run before contours can be created.");
                return;
            }
            else if (!NC.ComputationMgr.computationUnits[0])
            {
                MessageBox.Show("The computation needs to be run for Lden before contours can be created.");
                return;
            }
            
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;
            string contour = form.checkContours();
            NC.ComputationMgr.FindAndAddContours(contour);



            if (NC.ComputationMgr.contours != null && NC.ComputationMgr.contours.Length > 0)
            {
                drawContours(NC.ComputationMgr.contours, App, contour);
                //MessageBox.Show("Contours created!");
            }
            else MessageBox.Show("No contours have been found");
        }

        private static void drawContours(NC.NoiseContour[] contours, Bentley.Interop.MicroStationDGN.Application App, string contour)
        {
            //Create background:
            Point3d[] backgroundPoints = new Point3d[5];
            backgroundPoints[0].X = baseGrid.XStart; backgroundPoints[0].Y = baseGrid.YStart; backgroundPoints[0].Z = 0;
            backgroundPoints[1].X = baseGrid.XStart; backgroundPoints[1].Y = baseGrid.YEnd; backgroundPoints[1].Z = 0;
            backgroundPoints[2].X = baseGrid.XEnd; backgroundPoints[2].Y = baseGrid.YEnd; backgroundPoints[2].Z = 0;
            backgroundPoints[3].X = baseGrid.XEnd; backgroundPoints[3].Y = baseGrid.YStart; backgroundPoints[3].Z = 0;
            backgroundPoints[4].X = baseGrid.XStart; backgroundPoints[4].Y = baseGrid.YStart; backgroundPoints[0].Z = 0;
            ShapeElement g = App.CreateShapeElement1(null, backgroundPoints, MsdFillMode.Filled);

            if (contour == "Lden")
            {
                g.FillColor = 2;
                g.let_Level(createAndSetLevel(App, "Kote0dB"));
                App.ActiveModelReference.AddElement(g);
                for (int i = 0; i < contours.Length; i++)               // contour
                {
                    Level lev = createAndSetLevel(App, "Kote" + contours[i].level + "dB");
                    for (int j = 0; j < contours[i].Points.Count; j++)      // line
                    {
                        Point3d[] contourPoints = new Point3d[contours[i].Points[j].GetLength(0)];
                        for (int k = 0; k < contours[i].Points[j].GetLength(0); k++)      // point
                        {
                            contourPoints[k].X = contours[i].Points[j][k, 0];
                            contourPoints[k].Y = contours[i].Points[j][k, 1];
                            if (contours[i].level == 55)
                                contourPoints[k].Z = 1; // Yellow
                            else if (contours[i].level == 65)
                                contourPoints[k].Z = 2; // Red

                        }
                        ShapeElement s = App.CreateShapeElement1(null, contourPoints, MsdFillMode.Filled);
                        if (contours[i].level == 55)
                            s.FillColor = 4; // Yellow
                        else if (contours[i].level == 65)
                            s.FillColor = 3; // Red

                        s.let_Level(lev);
                        App.ActiveModelReference.AddElement(s);
                    }
                }
            }
            else if (contour == "LdenDetail")
            {
                //Color c = Color.FromArgb(112, 25, 25);
                g.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(112, 25, 25).ToArgb());
                g.let_Level(createAndSetLevel(App, "Kote0dB"));
                App.ActiveModelReference.AddElement(g);

                for (int i = 0; i < contours.Length; i++) // contour
                {
                    Level lev = createAndSetLevel(App, "Kote" + contours[i].level + "dB");
                    for (int j = 0; j < contours[i].Points.Count; j++)      // line
                    {
                        Point3d[] contourPoints = new Point3d[contours[i].Points[j].GetLength(0)];
                        for (int k = 0; k < contours[i].Points[j].GetLength(0); k++)      // point
                        {
                            contourPoints[k].X = contours[i].Points[j][k, 0];
                            contourPoints[k].Y = contours[i].Points[j][k, 1];

                            AddingZValueToPoint(contours, i, contourPoints, k);
                        }
                        ShapeElement s = App.CreateShapeElement1(null, contourPoints, MsdFillMode.Filled);

                        AddFillColorToShape(contours, App, i, s);

                        s.let_Level(lev);
                        App.ActiveModelReference.AddElement(s);
                    }
                }
            }
        }

        public static void drawContours(NC.NoiseContour[] contours, Bentley.Interop.MicroStationDGN.Application App, string contour, double x1, double y1, double x2, double y2)
        {
            //Create background:
            Point3d[] backgroundPoints = new Point3d[5];
            backgroundPoints[0].X = x1; backgroundPoints[0].Y = y1; backgroundPoints[0].Z = 0;
            backgroundPoints[1].X = x1; backgroundPoints[1].Y = y2; backgroundPoints[1].Z = 0;
            backgroundPoints[2].X = x2; backgroundPoints[2].Y = y2; backgroundPoints[2].Z = 0;
            backgroundPoints[3].X = x2; backgroundPoints[3].Y = y1; backgroundPoints[3].Z = 0;
            backgroundPoints[4].X = x1; backgroundPoints[4].Y = y1; backgroundPoints[0].Z = 0;
            ShapeElement g = App.CreateShapeElement1(null, backgroundPoints, MsdFillMode.Filled);

            if (contour == "Lden")
            {
                g.FillColor = 2;
                g.let_Level(createAndSetLevel(App, "Kote0dB"));
                App.ActiveModelReference.AddElement(g);
                for (int i = 0; i < contours.Length; i++)               // contour
                {
                    Level lev = createAndSetLevel(App, "Kote" + contours[i].level + "dB");
                    for (int j = 0; j < contours[i].Points.Count; j++)      // line
                    {
                        Point3d[] contourPoints = new Point3d[contours[i].Points[j].GetLength(0)];
                        for (int k = 0; k < contours[i].Points[j].GetLength(0); k++)      // point
                        {
                            contourPoints[k].X = contours[i].Points[j][k, 0];
                            contourPoints[k].Y = contours[i].Points[j][k, 1];
                            if (contours[i].level == 55)
                                contourPoints[k].Z = 1; // Yellow
                            else if (contours[i].level == 65)
                                contourPoints[k].Z = 2; // Red

                        }
                        ShapeElement s = App.CreateShapeElement1(null, contourPoints, MsdFillMode.Filled);
                        if (contours[i].level == 55)
                            s.FillColor = 4; // Yellow
                        else if (contours[i].level == 65)
                            s.FillColor = 3; // Red

                        s.let_Level(lev);
                        App.ActiveModelReference.AddElement(s);
                    }
                }
            }
            else if (contour == "LdenDetail")
            {
                //Color c = Color.FromArgb(112, 25, 25);
                g.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(112, 25, 25).ToArgb());
                g.let_Level(createAndSetLevel(App, "Kote0dB"));
                App.ActiveModelReference.AddElement(g);

                for (int i = 0; i < contours.Length; i++) // contour
                {
                    Level lev = createAndSetLevel(App, "Kote" + contours[i].level + "dB");
                    for (int j = 0; j < contours[i].Points.Count; j++)      // line
                    {
                        Point3d[] contourPoints = new Point3d[contours[i].Points[j].GetLength(0)];
                        for (int k = 0; k < contours[i].Points[j].GetLength(0); k++)      // point
                        {
                            contourPoints[k].X = contours[i].Points[j][k, 0];
                            contourPoints[k].Y = contours[i].Points[j][k, 1];

                            AddingZValueToPoint(contours, i, contourPoints, k);
                        }
                        ShapeElement s = App.CreateShapeElement1(null, contourPoints, MsdFillMode.Filled);

                        AddFillColorToShape(contours, App, i, s);

                        s.let_Level(lev);
                        App.ActiveModelReference.AddElement(s);
                    }
                }
            }
        }

        private static void AddingZValueToPoint(NC.NoiseContour[] contours, int i, Point3d[] contourPoints, int k)
        {
            switch ((int)contours[i].level)
            {
                case 38:
                    contourPoints[k].Z = 0.25;
                    break;
                case 41:
                    contourPoints[k].Z = .5;
                    break;
                case 44:
                    contourPoints[k].Z = 1;
                    break;
                case 47:
                    contourPoints[k].Z = 1.5;
                    break;
                case 50:
                    contourPoints[k].Z = 2;
                    break;
                case 53:
                    contourPoints[k].Z = 2.5;
                    break;
                case 56:
                    contourPoints[k].Z = 3;
                    break;
                case 59:
                    contourPoints[k].Z = 3.5;
                    break;
                case 62:
                    contourPoints[k].Z = 4;
                    break;
                case 65:
                    contourPoints[k].Z = 4.5;
                    break;
                case 68:
                    contourPoints[k].Z = 5;
                    break;
                case 71:
                    contourPoints[k].Z = 5.5;
                    break;
                case 74:
                    contourPoints[k].Z = 6;
                    break;
                case 77:
                    contourPoints[k].Z = 6.5;
                    break;
            }
        }

        private static void AddFillColorToShape(NC.NoiseContour[] contours, Bentley.Interop.MicroStationDGN.Application App, int i, ShapeElement s)
        {
            switch ((int)contours[i].level)
            {
                case 35:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(255, 0, 0).ToArgb());
                    break;
                case 38:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(237, 149, 100).ToArgb());
                    break;
                case 41:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(255, 191, 0).ToArgb());
                    break;
                case 44:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(208, 224, 64).ToArgb());
                    break;
                case 47:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(212, 255, 127).ToArgb());
                    break;
                case 50:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(144, 238, 144).ToArgb());
                    break;
                case 53:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(0, 252, 124).ToArgb());
                    break;
                case 56:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(50, 205, 154).ToArgb());
                    break;
                case 59:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(0, 255, 255).ToArgb());
                    break;
                case 62:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(0, 215, 255).ToArgb());
                    break;
                case 65:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(0, 165, 255).ToArgb());
                    break;
                case 68:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(0, 140, 255).ToArgb());
                    break;
                case 71:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(0, 69, 255).ToArgb());
                    break;
                case 74:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(0, 0, 255).ToArgb());
                    break;
                case 77:
                    s.FillColor = App.ActiveModelReference.InternalColorFromRGBColor(Color.FromArgb(34, 34, 178).ToArgb());
                    break;
            }
        }

        #region Methods for backgroundworker

        private static void bw_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            pBarForm.Dispose();
            // This event handler is called when the background thread finishes. 
            // This method runs on the main thread. 
            if (NC.ComputationMgr.successfullyRun)
                MessageBox.Show("Computation successfully finished in " + NC.ComputationMgr.calcTime.Hours
                + " hour(s), " + NC.ComputationMgr.calcTime.Minutes + " minute(s) and "
                + NC.ComputationMgr.calcTime.Seconds + " second(s). Summary and other values have been written to file.", "Finished Computation", MessageBoxButtons.OKCancel);
            else
                MessageBox.Show("Computation finished with some error.");

            if (e.Error != null)
                MessageBox.Show("Error: " + e.Error.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            else if (e.Cancelled)
                MessageBox.Show("Computation cancelled.");
        }

        private static void bw_DoWork(object sender, DoWorkEventArgs e)
        {
            // This event handler is where the actual work is done. 
            // This method runs on the background thread. 
            BackgroundWorker worker = sender as BackgroundWorker;
            NC.ComputationMgr.Run();
        }

        #endregion


        #region ILocateCommandEvents Members

        public void Accept(Element Element, ref Point3d Point, Bentley.Interop.MicroStationDGN.View View)
        {

        }

        public void Cleanup()
        {
            form.DetachFromMicroStation();
        }

        public void Dynamics(ref Point3d Point, Bentley.Interop.MicroStationDGN.View View, MsdDrawingMode DrawMode)
        {

        }

        public void LocateFailed()
        {

        }

        public void LocateFilter(Element Element, ref Point3d Point, ref bool Accepted)
        {

        }

        public void LocateReset()
        {

        }

        public void Start()
        {

            form.AttachToToolSettings(mainAddin.Instance);

        }
        #endregion


       
    }
}
