﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Collections;

using CommonUtils;
using GeoUtils;
using SoundKernel;
using MapKernel;
using Innemodul;
using System.Windows.Forms;
using System.ComponentModel;

namespace NoiseComputation
{

    public static class ComputationMgr
    {
        
        public static skSubTask currentSubTask;

        // Grid variables:
        public static double x1 = 0;
        public static double y1 = 0;
        public static double x2 = 0;
        public static double y2 = 0;
        public static float[,] zgrid;

        static skTopGrid Grid;

        // Calculation variables
        //public static int numUnits = 0;
        public static double[,] calcPoints;

        // Road variables
        public static NoiseRoad[] noiseRoads;
        public static int roadCount = 0;

        // Source variables
        public static NoiseSource[] noiseSources;

        // Mound variables
        public static NoiseMound[] noiseMounds;

        // Building variables
        public static NoiseBuilding[] noiseBuildings;

        /// <summary>Kotegenereringsobjekt</summary>
        static mkTriMesh MK = new mkTriMesh();

        /// <summary>Rektangel som omslutter ønsket utsnitt av kartet</summary>
        static typBBox MapBox;

        //static float TriNoData = -9999f;
        static int numCalcPoints = 0;

        //ID variables
        //static private int roadCount = 0;
        static private int sourceCount = 0;
        static private int buildingCount = 0;
        static private int moundCount = 0;

        static typmkLayer[][] layers;

        public static NoiseContour[] contours;

        public static bool[] computationUnits;
        private static int numUnits;
        public static bool successfullyRun = false;
        public static TimeSpan calcTime;

        private static string summary = string.Empty;

        public static void Run()
        {
            DateTime startTime = DateTime.Now;

            //AppDomain.CurrentDomain.AppendPrivatePath("C:\\isycad_v8i\\mdlapps\\addin\\bin");
            //AppDomain.CurrentDomain.AppendPrivatePath("C:\isycad_v8i\mdlapps\addin\bin");
            //AppDomain currentDomain = AppDomain.CurrentDomain;
            //currentDomain.AssemblyResolve += assemblyResolver.ResolveEventHandler;

            //Assembly assembly = Assembly.LoadFrom(assemblyPath);
            //return assembly;

            Queue<typError> Errors = new Queue<typError>();
            Queue<typError> Warnings = new Queue<typError>();

            double xorigo = x1, yorigo = y1; 

            // Oppretter en (tom) topografimodell
            skTopography Topo = new skTopography();

            CreateGrid();
            Topo.TopGrid.Add(Grid);

            numUnits = findNumUnits();
            AddCalculationPoints(numUnits);

            currentSubTask.Sources = new AListWrap(enFC.Source);


            // Trenger å lese denne filen for at fasadepunkt skal håndteres korrekt.
            if (!clsImport.LesKonfigurasjon("C:\\Dev\\Microstation_v8i\\DLL\\NoiseComputation\\bin\\Debug\\konfig.dat", out currentSubTask.config))
                throw new Exception("Feil ved lesing av konfig.dat");

            if (noiseRoads != null)
                AddRoads(noiseRoads, Topo);

            if (noiseSources != null)
                AddNoiseSources(noiseSources);

            if (noiseMounds != null)
                AddMounds(noiseMounds, Topo);

            if (noiseBuildings != null)
                AddBuildings(noiseBuildings, Topo);

            // App.MessageCenter.StatusMessage = "Generating triangular model...";
            // Genererer trekantmodell basert på innleste objekter 
            Top2Tri T2T = new Top2Tri(Topo, false);
            SetLimitsAndCreateTriangles(xorigo, yorigo, T2T);

            List<conline> ConLines = new List<conline>();
            List<conpoly> ConPolys = new List<conpoly>();
            PopulateConLinesAndPolys(T2T, ref ConLines, ref ConPolys);

            // Bestemmer de unike punktene i trekantsamlingen
            FindPointsAndRefreshTopo(xorigo, yorigo, Topo);

            // Definerer fem beregningsenheter: LAnatt, Leq, Lden, LAmax og LA5F
            DefiningCalculations(currentSubTask, numUnits);

            // Spesifiserer værforhold 
            CreateAndAddWeather();

            //SK.MainPath = "C:\\Dev\\Microstation_v8i\\App\\ISY.CAD.Noise\\bin\\Debug\\";
            // Dumper 3D-modell til fil (for debug innsyn)
            try //se if it is possible with activityType timecheck for this one
            {
                TriGrid2File.WriteParts(Topo, 1.0, true, false, -1);
            }
            catch
            {
                MessageBox.Show("Some parts are not defined correctly or not on the terrain", "Info", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }


            // Definerer beregningsgrid, 20 meter gridstørrelse
            griddef GrdDef = new griddef();
            SpecifyingCalculationGrid(GrdDef);

            // Bygger trekantmodell for beregningspunkt.
            // Merk at hvis ConLines og ConPolys IKKE er tomme (men inneholder skjermer og bygningsomriss)
            // så vil beregningsgridet IKKE være regulært; Det vil bli etablert beregningspunkt BÅDE for 
            // punktene definert i GrdDef OG for alle punktene i ConLines og ConPolys.
            Grid2Tri G2T = new Grid2Tri(currentSubTask);
            if (!G2T.CreateCalcTri(GrdDef, ConLines, ConPolys, xorigo, yorigo, null))
            {
                // Noe galt skjedde i forbindelse med triangulering
                throw new Exception("Feil ved triangulering av beregningsgrid");
            }

            currentSubTask.SubTopoHash = Topo.Hash;
            currentSubTask.Method = enMethod.N2000R;
            currentSubTask.ih = enHeightBase.terr;
            currentSubTask.Sources.Refresh();
            currentSubTask.Refresh();

            // Utfører beregning for beregningspunktene
            SK.TOPO = Topo;

            // Triks for å sikre at resultatene lagres korrekt
            SK.Task = skSubTask.Load(currentSubTask.Save());

            SK.Task.bldDetailedFP = new Dictionary<long, int>();
            for (int i = 0; i < buildingCount; i++)
                SK.Task.bldDetailedFP[i] = 3;

            int moundCount = (noiseMounds != null) ? noiseMounds.Length : 0;
            int roadCount = (noiseRoads != null) ? noiseRoads.Length : 0;
            int sourceCount = currentSubTask.Sources != null ? currentSubTask.Sources.Count : 0;
            int bCount = noiseBuildings != null ? noiseBuildings.Length : 0;

            //resetting and enabling the computation to be run several times
            SK.Stopped = false;
            SK.Stopping = false;
            SK.Errors.Clear();
            SK.PercentFinished = 0;
            successfullyRun = false;
            contours = null;

            SK.MainStatus = enMStat.active;
            SK.threadcontrol = new System.Threading.EventWaitHandle(false, System.Threading.EventResetMode.ManualReset);
            SK.threadcontrol.Set();

            SK.RunSubTask();

            FinishedComputation(Errors);

            DateTime endTime = DateTime.Now;
            calcTime = endTime.TimeOfDay - startTime.TimeOfDay;

            summary += "\r\n ************************************************************************* \r\n" +
                " \r\n Summary: \r\n Grid found: \t Number of points: " + Grid.zGrid.Length +
                ", Number of lines: " + Grid.zGrid.GetLength(0) + " * " + Grid.zGrid.GetLength(1) +
                "\r\n Start Points: X: \t" + x1 + "\t Y: \t" + y1 +
                "\r\n End Points: X: \t" + x2 + "\t Y: \t" + y2 +
                "\r\n ************************************************************************* \r\n" +
                "\r\n Buildings found: \t\t" + bCount +
                "\r\n Calculation points found: \t" + numCalcPoints +
                "\r\n Roads found: \t\t" + roadCount +
                "\r\n Noise sources found: \t" + sourceCount +
                "\r\n Mounds found: \t\t" + moundCount;
            System.IO.File.WriteAllText(@"C:\TEMP\Summary.txt", summary);
           
        }

        private static void CreateGrid()
        {
            int nx = zgrid.GetLength(0);
            int ny = zgrid.GetLength(1);
            /// <summary>Vinkel for skråplan mot underliggende grid (0=hor., 90=vert.)</summary>
            int a = 90;
            /// <param name="m_">Default marktype</param>
            /// unknown, hard, soft, A, B, C, D, E, F, G, H, Index6
            enGroundType groundType = enGroundType.soft;
            /// <param name="s_">Default skoghøyde</param>
            int s = 0;
            /// <param name="r_">Default ruhet</param>
            int r = 0;

            Grid = new skTopGrid(x1, y1, x2, y2, a, nx, ny, a, groundType, s, r);
            Grid.zGrid = zgrid;
            Grid.Refresh();
        }

        private static int findNumUnits()
        {
            int numUnits = 0;
            foreach (bool b in computationUnits)
                if (b)
                    numUnits++;
            if (numUnits == 0)
            {
                numUnits = 1;
                computationUnits[0] = true; // calculate Lden as default if no others are selected
            }
            return numUnits;
        }

        private static void AddCalculationPoints(int numUnits)
        {
            if (calcPoints == null)
            {
                currentSubTask = new skSubTask(0, 0, 0, numUnits); 
                return;
            }

            numCalcPoints = calcPoints.GetLength(0);

            // Oppretter beregning for i np enkeltpunkt og numUnits støyenheter
            // (resten av objektet blir fylt ut senere)
            currentSubTask = new skSubTask(0, 0, numCalcPoints, numUnits); // Need to find activityType way to create ID for the tasks?

            if (numCalcPoints > 0)
            {
                for (int i = 0; i < calcPoints.GetLength(0); i++)
                {
                    currentSubTask.points[i, 0] = calcPoints[i, 0];
                    currentSubTask.points[i, 1] = calcPoints[i, 1];
                    currentSubTask.hpoints[i] = (float)calcPoints[i,2];
                }
            }
        }

        private static void AddRoads(NoiseRoad[] roads, skTopography Topo)
        {
            foreach (NoiseRoad r in roads)
            {
                skTopRoad newRoad = new skTopRoad("Road" + roadCount, r.width, enHeightBase.sea, enGroundType.hard, "AB16", 1, r.Points.GetLength(0));
                summary += "\r\n ************************************************************************* \r\n" + 
                    "\r\n Noise Source: " + roadCount + "\t of type: road. " +
                    "\r\n AADT: "+r.ydt+", Distribution type: "+r.trafficDistributionType +
                    ", Activity type: "+r.activityType + ",\r\n Velocity: "+r.velocity+", Width: "+r.width;
                  
                        
                for (int pGonIndex = 0; pGonIndex < r.Points.GetLength(0); ++pGonIndex)
                {
                    summary += "\r\n Point "+pGonIndex+":  X: \t" +
                            r.Points[pGonIndex, 0] + "\t Y: \t" + r.Points[pGonIndex, 1] + "\t Z: \t" + r.Points[pGonIndex, 2];
                    newRoad.pGon[pGonIndex, 0] = r.Points[pGonIndex, 0];
                    newRoad.pGon[pGonIndex, 1] = r.Points[pGonIndex, 1];
                    newRoad.hs[pGonIndex] = (float)r.Points[pGonIndex, 2];
                }

                if (r.isTunnel)
                {
                    newRoad.tb = -4f;
                    newRoad.tunsh = enTunShape.semicircular;
                    newRoad.tunsurf = enTunSurf.smooth; //as defined in SoundKernel
                    newRoad.tunopen = enLineEnd.both; //as defined in SoundKernel
                }
                if (r.isBridge)
                {
                    newRoad.tb = -1f;
                    newRoad.brdg_end = enLineEnd.both; //as defined in SoundKernel
                }

                newRoad.Refresh();
                Topo.TopRoads.Add(newRoad);
                AddNoiseSources(newRoad, r);
                roadCount++;
            }
        }

        private static void AddNoiseSources(skTopRoad r, NoiseRoad n)
        {
            float amount = n.ydt;
            int numClasses = 3;

            skSource newSource;
            if (n.isTunnel)
                newSource = new skSource(enActType.tunnel, enPlaceType.line, enQuantityUnit.ydt, amount, "Road" + roadCount, 0, numClasses);
            else
                newSource = new skSource(enActType.road, enPlaceType.line, enQuantityUnit.ydt, amount, "Road" + roadCount, 0, numClasses);

            float[] sourcePeriodTraffic = populateSourcePeriodTraffic(n.trafficDistributionType, amount, n.vehicleDistribution);

            if (n.trafficDistributionType == "UserDefined")
                populateTFacUserDef(newSource, n.vehicleDistribution, n.activityType);
            else
                populateTFac(n.activityType, n.trafficDistributionType, newSource, sourcePeriodTraffic);


            if (n.frequencySpectra != null)
            {
                newSource.Params["RoadSrcSpec"] = populateRoadSourceSpec(n.frequencySpectra);
                summary += "\r\n Frequency spectra: ";
                for (int j = 0; j < n.frequencySpectra.Length; j++)
                    summary += "\r\n " + j + ": " + n.frequencySpectra[j];
            }
            
            newSource.RefID = "0 0.000 0.999"; // Dette vil vanligvis være "ReferanselenkeID FromMeasure Tomeasure"

            if (n.isBridge)
                newSource.Params["TunnelBridgeHeight"] = r.tb;      // Tunnelhøyde / Brotykkelse
            if (n.isTunnel)
            {
                newSource.Params["TunnelBridgeHeight"] = r.tb;      // Tunnelhøyde / Brotykkelse
                newSource.Params["TunnelFactor"] = r.tunfact; // Tunnelfaktor [Er regnet ut v/Road.Refresh() over]
                newSource.Params["TunnelShape"] = r.tunsh;   // Tunnelform
                newSource.Params["TunnelOpening"] = r.tunopen; // Tunnelåpninger
                newSource.Params["TunnelSurf"] = r.tunsurf; // Tunnelvegg-egenskap
            }
            newSource.Params["RoadSurf"] = r.rs;      // Vegdekke 
            newSource.Params["RoadWidth"] = r.b;      // Vegbredde 
            newSource.Params["RoadSpeedLim"] = n.velocity;         // 40 km/t

            newSource.Vectors = r.Vectors;
            newSource.h = r.hs;
            newSource.Refresh();
            currentSubTask.Sources.Add(newSource);
        }

        private static float[] populateSourcePeriodTraffic(string trafficDistributionType, float amount, double[] distribution)
        {
            float[] sourcePeriodTraffic = new float[3];

            if (trafficDistributionType == "General")
            {
                sourcePeriodTraffic[0] = amount * 0.75f; // Distribution Day
                sourcePeriodTraffic[1] = amount * 0.15f; // Distribution Evening
                sourcePeriodTraffic[2] = amount * 0.1f;  // Distribution Night
            }
            else if (trafficDistributionType == "Work")
            {
                sourcePeriodTraffic[0] = amount;         // Distribution Day
                sourcePeriodTraffic[1] = 0;              // Distribution Evening
                sourcePeriodTraffic[2] = 0;              // Distribution Night
            }
            // change this section to somehow input the values from the road noise source
            else if (trafficDistributionType == "UserDefined")
            {
                sourcePeriodTraffic[0] = (float)distribution[0];          // Distribution Day
                sourcePeriodTraffic[1] = (float)distribution[1];          // Distribution Evening
                sourcePeriodTraffic[2] = (float)distribution[2];          // Distribution Night
            }
            return sourcePeriodTraffic;
        }

        private static void populateTFacUserDef(skSource newSource, double[] distribution, string actType)
        {
            string tfaclines = "";
            double weektotal = 0;
            for (int i = 0; i < 9; i++)
            {
                if (i == 0 || i == 1 || i == 2) //mon-fri
                    weektotal += (5 * distribution[i]);
                else
                    weektotal += distribution[i];
            }

            double[] vehicleClasses = new double[3];
            vehicleClasses[2] = 1;          // only heavy (vehicle) class used

            if (actType == "Normal")
            {
                vehicleClasses[0] = 0.85;    // light (vehicle) class
                vehicleClasses[1] = 0.075;   // medium (vehicle) class
                vehicleClasses[2] = 0.075;   // heavy (vehicle) class
            }

            float checkTFac = 0;
         
            for (int period = 0; period < 3; ++period) //Døgnsegment: Dag, Kveld, Natt 
            {
                for (int vehicleType = 0; vehicleType < 3; ++vehicleType) //Kategorier: Lett, Middels og Tung 
                {
                    for (int day = 0; day < 7; ++day) //Ukedag: Mandag til Søndag 
                        for (int month = 0; month < 12; ++month) //Måned: 1-12 
                        {
                            double dayDistribution;
                            if (day == 5)   // Saturday
                                dayDistribution = distribution[period + 3];
                            else if (day == 6)   // Sunday
                                dayDistribution = distribution[period + 6];
                            else dayDistribution = distribution[period];  // Mon-Fri
                            float thisContribution = (float)(dayDistribution / (weektotal * 12)*vehicleClasses[vehicleType]);
                            tfaclines += "\t" + thisContribution;
                            newSource.TFac[period, day, month][vehicleType] = thisContribution;
                            checkTFac += thisContribution;
                        }
                    tfaclines += "\r\n";
                }
            }
            if (Math.Round(checkTFac, 4) != 1)
                MessageBox.Show("Error! The sum of the TFac values are not 1, actual value: "+checkTFac, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            System.IO.File.WriteAllText(@"C:\TEMP\TFacLines.txt", tfaclines);
        }

        private static void populateTFac(string actType, string trafficDistribution, skSource newSource, float[] sourcePeriodTraffic)
        {
            string tfaclines = "";
            float lightVehicles, mediumVehicles, heavyVehicles;
            for (int period = 0; period < 3; ++period) //Døgnsegment: Dag, Kveld, Natt 
            {
                float periodTraffic = sourcePeriodTraffic[period] / newSource.Q;
                if (actType == "Normal")
                {
                    lightVehicles = periodTraffic * 0.85f;      // 85% of vehicles are light
                    mediumVehicles = periodTraffic * 0.075f;    // 7.5% of vehicles are medium
                    heavyVehicles = periodTraffic * 0.075f;     // 7.5% of vehicles are heavy
                }
                else 
                {
                    heavyVehicles = periodTraffic;              // all vehicles are heavy
                    lightVehicles = 0;
                    mediumVehicles = 0;
                }

                for (int vehicleType = 0; vehicleType < 3; ++vehicleType) //Kategorier: Lett, Middels og Tung 
                {
                    float categoryTraffic = (vehicleType == 0 ? lightVehicles : ((vehicleType == 1) ? mediumVehicles : heavyVehicles));
                    for (int day = 0; day < 7; ++day) //Ukedag: Mandag til Søndag 
                        for (int month = 0; month < 12; ++month) //Måned: 1-12 
                        {
                            int numberOfWeekdays = 7;
                            if (trafficDistribution == "General")
                                numberOfWeekdays = 7;
                            else if (trafficDistribution == "Work")
                                numberOfWeekdays = 5;
                            float thisContribution = categoryTraffic / (numberOfWeekdays * 12);
                            if (trafficDistribution == "Work" && (day == 5 || day == 6))
                                thisContribution = 0;
                            newSource.TFac[period, day, month][vehicleType] = thisContribution;
                            tfaclines += "\t" + thisContribution;
                        }
                    tfaclines += "\r\n";
                }
            }
            System.IO.File.WriteAllText(@"C:\TEMP\TFacLines.txt", tfaclines);
        }

        private static double[, , ,] populateRoadSourceSpec(double[] frequencySpectra)
        {
            // 3 kjøretøykategorier, 3 kildehøyder [1 cm, 30 cm, 75 cm], 27 frekvenser, 3 enheter [Lw, L5, Lmax]
            double[, , ,] sourceSpec = new double[3, 3, 27, 3];
            // Verdier trenger bare fylles ut for kjøretøykategorier og kildehøyder som skal brukes.
            // NB: Kategori 0 ("lette") bruker kildehøyde 0 (1 cm) og 1 (30 cm). 
            // Kategori 1 ("middels") og 2 ("tunge") bruker kildehøyde 0 (1 cm) og 2 (75 cm)! 
            // Pass på at ulike kategorier har utfylt data for sin(e) kildehøyder.
            // Enhet 0 må fylles ut for å få beregnet Lden og Leq. 
            // Enhet 2 må fylles ut for å få beregnet Lmax (hvis det er aktuelt). 
            // Enhet 1 (L5) kan utelates.

            for (int vehicle_i = 0; vehicle_i < 3; vehicle_i++)
            {
                if (vehicle_i == 0 || vehicle_i == 1)
                    continue;
                for (int sourceheight_i = 0; sourceheight_i < 3; sourceheight_i++)
                {
                    if (sourceheight_i == 1)
                        continue;
                    for (int freq_i = 0; freq_i < 27; freq_i++)
                    {
                        for (int unit_i = 0; unit_i < 3; unit_i++)
                        {
                            if (unit_i == 1 || unit_i == 2)
                                sourceSpec[vehicle_i, sourceheight_i, freq_i, unit_i] = 0;
                            //else fylle ut for unit = 1, og frekvenser
                            else
                                sourceSpec[vehicle_i, sourceheight_i, freq_i, unit_i] = frequencySpectra[freq_i];

                        }
                    }
                }
            }
            return sourceSpec;
        }
   
        private static void AddNoiseSources(NoiseSource[] noiseSources)
        {
            sourceCount = roadCount + 1;
            foreach (NoiseSource s in noiseSources)
            {
                float amount = s.amount; // mengde
                enActType actType = assignActType(s.sourceType);
                enPlaceType placeType = assignPlaceType(s.geometry);
                enQuantityUnit quantityUnit = assignQuantityUnit(s.amountUnit);
                if (placeType == enPlaceType.point && s.Points.GetLength(0) > 1)
                    MessageBox.Show("An uncorrect geometry has been specified", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);

                skSource newSource;
                if (s.sourceType == "industry")
                {
                    newSource = new skSource(actType, placeType, quantityUnit, amount, "industry" + sourceCount.ToString(), s.Points.GetLength(0), 3);
                    summary += "\r\n ************************************************************************* \r\n" + 
                        "\r\n Noise Source: " + sourceCount + "\t of type: industry. "+
                    "\r\n Amount: "+s.amount+", Distribution type: "+s.timeDistributionType +
                    ", Activity type: "+s.activityType + "\r\n Frequency spectra:";
                    for (int j = 0; j < s.frequencySpectra.Length; j++)
                        summary += "\r\n " +j+ ": "+s.frequencySpectra[j];
                    for (int i = 0; i < s.Points.GetLength(0); ++i)
                    {
                        summary += "\r\n Points: X: \t" +
                            s.Points[i, 0] + "\t Y: \t" + s.Points[i, 1] + "\t Z: \t" + s.Points[i, 0];
                        newSource.Dots[i, 0] = s.Points[i, 0];
                        newSource.Dots[i, 1] = s.Points[i, 1];
                        newSource.h[i] = (float)s.Points[i, 2];
                    }

                    float[] sourcePeriodTraffic = populateSourcePeriodTraffic(s.timeDistributionType, amount, s.timeDistribution);
                    
                    if (s.timeDistributionType == "UserDefined")
                        populateTFacUserDef(newSource, s.timeDistribution, s.activityType);
                    else 
                        populateTFac(s.activityType, s.timeDistributionType, newSource, sourcePeriodTraffic);

                    newSource.Params["SourceSpec"] = populateIndustrySourceSpec(s.frequencySpectra);
                    newSource.Params["NumClasses"] = 3;

                    newSource.Refresh();
                    currentSubTask.Sources.Add(newSource);
                }
                else MessageBox.Show("Only industry noise source has been implemented");
            }
        }

        private static enActType assignActType(string p)
        {
            switch (p)
            {
                case "industry":
                    return enActType.industry;
                case "road":
                    return enActType.road;
                case "tunnel":
                    return enActType.tunnel;
                default:
                    return enActType.unknown;
            }
        }

        private static enPlaceType assignPlaceType(string p)
        {
            switch (p)
            {
                case "point":
                    return enPlaceType.point;
                case "line":
                    return enPlaceType.line;
                case "area":
                    return enPlaceType.area;
                case "volume":
                    return enPlaceType.volume;
                default:
                    return enPlaceType.point;
            }
        }

        private static enQuantityUnit assignQuantityUnit(string p)
        {
            switch (p)
            {
                case "usehours":
                    return enQuantityUnit.usehours;
                case "events":
                    return enQuantityUnit.events;
                case "ydt":
                    return enQuantityUnit.ydt;
                default:
                    return enQuantityUnit.usehours;
            }
        }

        private static double[, ,] populateIndustrySourceSpec(double[] frequencySpectra)
        {
            // 3 klasser, 27 frekvenser, 3 enheter [Lw, L5, Lmax]
            double[, ,] sourceSpec = new double[3, 27, 3];
            // Verdier trenger bare fylles ut for klasser som skal brukes.
            // Enhet 0 må fylles ut for å få beregnet Lden og Leq. 
            // Enhet 2 må fylles ut for å få beregnet Lmax (hvis det er aktuelt). 
            // Enhet 1 (L5) kan utelates.

            for (int class_i = 0; class_i < 3; class_i++)
            {
                if (class_i == 0 || class_i == 1)
                    continue;
                for (int freq_i = 0; freq_i < 27; freq_i++)
                    for (int unit_i = 0; unit_i < 3; unit_i++)
                    {
                        if (unit_i == 1 || unit_i == 2)
                            sourceSpec[class_i, freq_i, unit_i] = 0;
                        //else fylle ut for unit = 1, og frekvenser
                        else
                            sourceSpec[class_i, freq_i, unit_i] = frequencySpectra[freq_i];
                    }
            }
            return sourceSpec;
        }

        private static void generateYDT(skSource newSource, int period, float heavyTraffic, float lightTraffic)
        {
            for (int vehicleType = 0; vehicleType < 3; ++vehicleType) //Kategorier: Lett, Middels og Tung 
            {
                float categoryTraffic = (vehicleType == 0 ? lightTraffic : heavyTraffic);
                for (int day = 0; day < 7; ++day) //Ukedag: Mandag til Søndag 
                    for (int month = 0; month < 12; ++month) //Måned: 1-12 
                    {
                        float thisContribution = categoryTraffic / (7 * 12);
                        newSource.TFac[period, day, month][vehicleType] = thisContribution;
                    }
            }
        }

        private static void AddMounds(NoiseMound[] noiseMounds, skTopography Topo)
        {
            foreach (NoiseMound m in noiseMounds)
            {
                skTopScreen newMound = new skTopScreen(m.Points.GetLength(0), enHeightBase.sea, enMoundMat.earth, false, false);
                bool isScreen = m.topWidth == 0 ? true : false;
                bool isWall = m.topWidth != 0 && m.bottomWidth == 0 ? true : false;
                bool isMound = isScreen || isWall ? false : true;

                if (isScreen)
                    summary += "\r\n ************************************************************************* \r\n" +
                        "\r\n Screen Noise Mitigation: \r\n Height: "+m.height;
                else if (isWall)
                    summary += "\r\n Wall Noise Mitigation: \r\n Left Height: " + 
                        m.height + "Right Height: "+m.rightHeight + " Width: "+m.topWidth;
                else summary += "\r\n Mound Noise Mitigation: \r\n Left Height: " +
                        m.height + "Right Height: " + m.rightHeight + " Top Width: " + 
                        m.topWidth + " Bottom Width: " + m.bottomWidth;

                for (int i = 0; i < m.Points.GetLength(0); i++)
                {
                    newMound.pGon[i, 0] = m.Points[i, 0];
                    newMound.pGon[i, 1] = m.Points[i, 1];
                    summary += "\r\n Point " + i + ", \t x: \t" + m.Points[i, 0] +
                                      ", \t y: \t" + m.Points[i, 1];
                    if (isScreen)
                        newMound.h0[i] = m.height + (float)m.Points[i, 2];
                    else
                    {
                        // høyre (h1) og venstre (h2)
                        newMound.h1[i] = m.rightHeight + (float)m.Points[i, 2];
                        newMound.h2[i] = m.height + (float)m.Points[i, 2];
                    }
                }
                //Bredde til topp (bt) og bunn (bb) av voll
                if (isWall || isMound)
                    newMound.bt = m.topWidth;
                if (isMound)
                    newMound.bb = m.bottomWidth;
                newMound.Refresh();
                Topo.TopScreens.Add(newMound);



            }
        }

        private static void AddBuildings(NoiseBuilding[] noiseBuildings, skTopography Topo)
        {
            
            foreach (NoiseBuilding n in noiseBuildings)
            {
                //check if building is activityType closed line
                int numberOfPoints = n.outLines.GetLength(0);
                if ((n.outLines[0, 0] == n.outLines[numberOfPoints - 1, 0]) && (n.outLines[0, 1] == n.outLines[numberOfPoints - 1, 1]) && (n.outLines[0, 2] == n.outLines[numberOfPoints - 1, 2]))
                {

                    skTopBuilding newBuilding = new skTopBuilding(enHeightBase.sea, n.outLines.GetLength(0), n.inLines.GetLength(0), 1, 1);
                    newBuilding.ID = buildingCount++;
                    summary += "\r\n ************************************************************************* \r\n" +
                       "\r\n Building points:";
                    for (int i = 0; i < n.outLines.GetLength(0); i++)
                    {
                        newBuilding.OutLines[i].X = Math.Round(n.outLines[i, 0], 1);
                        newBuilding.OutLines[i].Y = Math.Round(n.outLines[i, 1], 1);
                        newBuilding.OutLines[i].Z = (float)Math.Round(n.outLines[i, 2], 1);
                        summary += " \r\n Outlinepoint " + i + ", \t x: \t" + newBuilding.OutLines[i].X +
                                                          ", \t y: \t" + newBuilding.OutLines[i].Y +
                                                          ", \t z: \t" + newBuilding.OutLines[i].Z;
                    }
                    for (int i = 0; i < n.inLines.GetLength(0); i++)
                    {
                        newBuilding.InLines[i].X = Math.Round(n.inLines[i, 0], 1);
                        newBuilding.InLines[i].Y = Math.Round(n.inLines[i, 1], 1);
                        newBuilding.InLines[i].Z = (float)Math.Round(n.inLines[i, 2], 1);
                        summary += "\r\n Inlinepoint " + i + ", \t x: \t" + newBuilding.InLines[i].X +
                                      ", \t y: \t" + newBuilding.InLines[i].Y +
                                      ", \t z: \t" + newBuilding.InLines[i].Z;
                    }
                    if (n.computeFacadePoints)
                        newBuilding.DoFPnts = true;
                    newBuilding.Refresh();
                    Topo.TopBuildings.Add(newBuilding);
                }
            }
            currentSubTask.sitidx.Add(0);
        }

        private static void SetLimitsAndCreateTriangles(double xorigo, double yorigo, Top2Tri T2T)
        {
            T2T.SetLimits(x1, x2, y1, y2, xorigo, yorigo);
            T2T.SetBldIDTrans(currentSubTask.buildings, currentSubTask.bldIDTrans, currentSubTask.ih, 0f, currentSubTask.config);
            if (!T2T.CreateTriangles(true, true, true, true, true, true, true, null))
            {
                //  Noe galt skjedde i forbindelse med triangulering 
                MessageBox.Show("Error in triangulation!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private static void PopulateConLinesAndPolys(Top2Tri T2T, ref List<conline> ConLines, ref List<conpoly> ConPolys)
        {
            // Henter kanter fra triangulering av topografi (er nødvendig for å bygge opp trekantsamling
            // for beregningspunkt, for at kotegenerering ikke skal krysse kanter, som f.eks. skjermer og bygninger.
            // Dersom dette IKKE er et krav, kan dette utelates. ConLines og ConPolys skal da være tomme.

            // Henter alle linjer og alle TOMME polygoner ( = bygninger)
            T2T.GetLinesPolys(true, true, true, out ConLines, out ConPolys);
            // Tar med bare et utvalg av linjer og polygoner; Bare skjermer og bygningsomriss; Ikke taklinjer, etc
            // Må også sikre at evt. endringer i punktene i forbindelse med trekantsamling for beregning
            // IKKE påvirker punktene i topografien. Tar derfor KOPIER av linjene og polygonene
            List<conline> TmpLines = new List<conline>(ConLines.Count);
            List<conpoly> TmpPolys = new List<conpoly>(ConPolys.Count);
            for (int i = 0; i < ConLines.Count; i++)
            {
                if ((string)ConLines[i].par["Otyp"] != "TopoScreen") continue;
                TmpLines.Add(ConLines[i].Copy(false));
            }
            for (int i = 0; i < ConPolys.Count; i++)
            {
                if ((string)ConPolys[i].par["Otyp"] != "TopoBuilding") continue;
                TmpPolys.Add(ConPolys[i].Copy(false));
            }
            ConLines = TmpLines; ConLines.TrimExcess(); TmpLines = null;
            ConPolys = TmpPolys; ConPolys.TrimExcess(); TmpPolys = null;
        }

        private static void FindPointsAndRefreshTopo(double xorigo, double yorigo, skTopography Topo)
        {
            Topo.TopTriang.FindUniquePoints();
            // Må konvertere alle genererte objekter (trekanter og speil) til globale koordinater 
            // (ConLines og ConPolys forblir i lokale koordinater)
            Topo.TopTriang.ConvertToGlobal(xorigo, yorigo);
            Topo.ConvertMirrorsToGlobal(xorigo, yorigo);
            Topo.TopTriang.RefreshHash();
            Topo.Refresh();
        }

        private static void DefiningCalculations(skSubTask currentSubTask, int numUnits)
        {
            int index = -1; // from 0 to numUnits-1
            for (int unitIndex = 0; unitIndex < computationUnits.Length; unitIndex++)
            {
                if (computationUnits[unitIndex] == true)
                {
                    index++;
                    switch (unitIndex)
                    {
                        case 0:
                            // Lden
                            defineLdenUnit(currentSubTask, index);
                            continue;
                        case 1:
                            // Leq
                            defineLeqUnit(currentSubTask, index);
                            continue;
                        case 2:
                            // LAnatt or Lnight
                            defineLnightUnit(currentSubTask, index);
                            continue;
                        case 3:
                            // LAFmax
                            defineLAFmaxUnit(currentSubTask, index);
                            continue;
                        case 4:
                            // L5AF
                            defineL5AFUnit(currentSubTask, index);
                            continue;
                        case 5:
                            // Levening
                            defineLeveningUnit(currentSubTask, index);
                            continue;
                        default:
                            // Lden
                            defineLdenUnit(currentSubTask, index);
                            continue;
                    }
                }
            }
        }

        private static void defineLeveningUnit(skSubTask currentSubTask, int index)
        {
            currentSubTask.Units[index].Unit = enSoundUnit.Leq;
            currentSubTask.Units[index].TW = enTimeWeight.H1923;
            currentSubTask.Units[index].FW = enFreqWeight.A;
            currentSubTask.Units[index].L1 = 10.0f;
            currentSubTask.Units[index].Toler = 0.2f;
        }

        private static void defineL5AFUnit(skSubTask currentSubTask, int index)
        {
            currentSubTask.Units[index].Unit = enSoundUnit.Lfast;
            currentSubTask.Units[index].TW = enTimeWeight.L5NGT;
            currentSubTask.Units[index].FW = enFreqWeight.A;
            currentSubTask.Units[index].L1 = 10.0f;
            currentSubTask.Units[index].Toler = 0.2f;
        }

        private static void defineLAFmaxUnit(skSubTask currentSubTask, int index)
        {
            currentSubTask.Units[index].Unit = enSoundUnit.Lfast;
            currentSubTask.Units[index].TW = enTimeWeight.none;
            currentSubTask.Units[index].FW = enFreqWeight.A;
            currentSubTask.Units[index].L1 = 10.0f;
            currentSubTask.Units[index].Toler = 0.2f;
        }

        private static void defineLdenUnit(skSubTask currentSubTask, int index)
        {
            currentSubTask.Units[index].Unit = enSoundUnit.Leq;
            currentSubTask.Units[index].TW = enTimeWeight.DEN;
            currentSubTask.Units[index].FW = enFreqWeight.A;
            currentSubTask.Units[index].L1 = 10.0f;
            currentSubTask.Units[index].Toler = 0.2f;
        }

        private static void defineLeqUnit(skSubTask currentSubTask, int index)
        {
            currentSubTask.Units[index].Unit = enSoundUnit.Leq;
            currentSubTask.Units[index].TW = enTimeWeight.none;
            currentSubTask.Units[index].FW = enFreqWeight.A;
            currentSubTask.Units[index].L1 = 10.0f;
            currentSubTask.Units[index].Toler = 0.2f;
        }

        private static void defineLnightUnit(skSubTask currentSubTask, int index)
        {
            currentSubTask.Units[index].Unit = enSoundUnit.Leq;
            currentSubTask.Units[index].TW = enTimeWeight.H2307;
            currentSubTask.Units[index].FW = enFreqWeight.A;
            currentSubTask.Units[index].L1 = 10.0f;
            currentSubTask.Units[index].Toler = 0.2f;
        }

        private static void CreateAndAddWeather()
        {
            skMetClass Met = new skMetClass();
            Met.TGrad = 0.0f;
            Met.Refresh(); // Sluttberegner interne egenskaper for været 
            currentSubTask.MET = new AListWrap(enFC.MET, Met); // Legger været til oppdraget 
            currentSubTask.MET.Refresh();
        }

        private static void SpecifyingCalculationGrid(griddef GrdDef)
        {
            GrdDef.x0 = x1; GrdDef.y0 = y1; 
            GrdDef.xi = 1.0f; GrdDef.yi = 0.0f;
            GrdDef.Lu = x2 - x1; GrdDef.Lv = y2 - y1;
            GrdDef.nu = zgrid.GetLength(0); GrdDef.nv = zgrid.GetLength(1);
            currentSubTask.GrdDef = GrdDef;
        }

        private static void FinishedComputation(Queue<typError> Errors)
        {
            if (SK.MainStatus == enMStat.error)
            {
                //  Noe galt skjedde i forbindelse med beregning 
                MessageBox.Show("Feil oppsto under beregning: \n" + SK.Errors.Peek().Desc
                    + ", \n" + SK.Errors.Peek().ErrMsg + ", \n in module: "
                    + SK.Errors.Peek().Module + ", \n and in function: " + SK.Errors.Peek().Func, "Info", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
            else
            {
                successfullyRun = true;

                // Må konvertere til dB
                ConvertTodB();

                calculateValues(SK.Task, numUnits);

                saveFacadeResults((skSubTask)SK.Task, Errors);

                #region To output SINTEF contours and values:
                //// Alle koter lages for visning i målestokk 1:5000
                //float mscale = 5000f;
                //SK.Contours = new List<skContour>();

                //// Definerer kotesett for Lden
                //CreatingLdenContours(mscale);

                //// Definerer detaljert kotesett for Lden
                //CreatingDetailedLdenContours(mscale);

                //bool stError, SaveOK;
                //SaveResults((skSubTask)SK.Task, Errors, out stError, out SaveOK);
                #endregion
            }
        }

        private static void ConvertTodB()
        {
            for (int i = 0; i < SK.Task.np; i++)
            {
                for (int u = 0; u < SK.Task.numUnits; u++)
                {
                    if (SK.Task.resultpoints[i, u] <= 0.00001f) // Alt under -50 dB regnes som NoData
                        SK.Task.resultpoints[i, u] = -50.0f;
                    else
                        SK.Task.resultpoints[i, u] = 10.0f * (float)Math.Log10(SK.Task.resultpoints[i, u]);
                }
            }
        }

        private static void calculateValues(skTask task, int numUnits)
        {
            float[,] p = task.resultpoints;
            summary += "\r\n ************************************************************************* \r\n" +
                " \r\n Computation Point Results:"; 
            for (int i = 0; i < numCalcPoints; i++)
            {
                int ind = 0;
                summary += "\r\n Point " + i + ":\t x: \t" + 
                    task.points[i,0] + "\t y: \t" + task.points[i,1] +
                    "\t z: \t" + task.hpoints[i] +"\t";
                if (computationUnits[0])
                    summary += "Lden: \t" + p[i, ind++] + "\t";
                if (computationUnits[1])
                    summary += "Lequ: \t" + p[i, ind++] + "\t";
                if (computationUnits[2])
                    summary += "Lnight: \t" + p[i, ind++] + "\t";
                if (computationUnits[3])
                    summary += "LAFmax: \t" + p[i, ind++] + "\t";
                if (computationUnits[4])
                    summary += "L5AF: \t" + p[i, ind++] + "\t";
                if (computationUnits[5])
                    summary += "Levening: \t" + p[i, ind++] + "\t";
            }
        }

        private static void saveFacadeResults(skSubTask subtask, Queue<typError> Errors)
        {
            double dbval;
            Bygning B;
            double mmin = double.PositiveInfinity;
            double mmax = double.NegativeInfinity;

            int iBld, ifp, iu = 0, nu = subtask.numUnits;
            string[] units = new string[nu];
        
                int ind = 0;
                if (computationUnits[0])
                    units[ind++] = "Lden";
                if (computationUnits[1])
                    units[ind++] = "Lequ";
                if (computationUnits[2])
                    units[ind++] = "Lnight";
                if (computationUnits[3])
                    units[ind++] = "LAFmax";
                if (computationUnits[4])
                    units[ind++] = "L5AF";
                if (computationUnits[5])
                    units[ind++] = "Levening";
           

            for (iBld = 0; iBld < subtask.buildings.Count; iBld++)
            {
                B = subtask.buildings[iBld];

                if (B.fasadepunkt.Count == 0) continue; // Skriver ikke "tomme" bygninger
                summary += "\r\n ************************************************************************* \r\n" +
                       "\r\n Facade Point Results:"+"\r\n Building number: " + iBld + ", " + B.bygningsnummer;

                

                for (ifp = 0; ifp < B.fasadepunkt.Count; ifp++)
                    //recnum++;
                    for (iu = 0; iu < nu; iu++)
                    {
                        summary += "\r\n Facade point: \t" + ifp + "\tPosition East: \t" + B.fasadepunkt[ifp].sted.posisjonE +
                            "\tPosition North: \t" + B.fasadepunkt[ifp].sted.posisjonN + "\tHeight: \t" + B.fasadepunkt[ifp].sted.h +
                            "\tDecibel value: \t" + B.fasadepunkt[ifp].utesituasjoner[0].utestøy.nivå[iu].verdi +
                            "\tUnit: \t" + units[iu].ToString();
                        
                        dbval = B.fasadepunkt[ifp].utesituasjoner[0].utestøy.nivå[iu].verdi;
                        mmin = Math.Min(dbval, mmin);
                        mmax = Math.Max(dbval, mmax);
                    }
                summary += "\r\nMin Value: \t" + mmin + "\tMax Value: \t" + mmax;
                mmin = double.PositiveInfinity; mmax = double.NegativeInfinity;
            }
        }

        public static void FindAndAddContours(string contour)
        {
            // Hvis koter finnes fra før:
            if (contours != null)
            {
                if (contour == "Lden")
                {
                    if (contours[0].level == 55)
                        return;
                    else
                    {
                        CreateAndAddContours(contour);
                        return;
                    }
                }
                else if (contour == "LdenDetail")
                {
                    if (contours[0].level == 35)
                        return;
                    else
                    {
                        CreateAndAddContours(contour);
                        return;
                    }
                }
            }

            // Alle koter lages for visning i målestokk 1:5000
            float mscale = 5000f;
            SK.Contours = new List<skContour>();

            // Definerer kotesett for Lden
            CreatingLdenContours(mscale);

            // Definerer detaljert kotesett for Lden
            CreatingDetailedLdenContours(mscale);

            // saves both contours into layers
            FindContours();

            // check if both contours were created properly:
            if (layers != null)
            {
                if (layers.Length == 0)
                    MessageBox.Show("No contours were found, this may be activityType result of errors or too low noise levels.");
                else if (layers.Length == 1)
                {
                    if (layers[0][0].Level == 55f)
                        MessageBox.Show("Only contours for Lden were found.");
                    else if (layers[0][0].Level == 35f)
                        MessageBox.Show("Only contours for detailed Lden were found.");
                }
            }
            else
                MessageBox.Show("No contours were found, this may be activityType result of errors or too low noise levels.");
            // gets the specific contour
            CreateAndAddContours(contour);
        }

        private static void FindContours()
        {
            float[] val;
            typBBox bbox;

            SK.Task.tris.Refresh(); // For å bygge topografisk indeks som brukes av MapKernel
            SK.Task.tris.CheckGridPoints(SK.Task.GrdDef, 0, 0); // For at IsGrdP-listen skal ha rett lengde

            bbox = SK.Task.tris.BBox;
            for (int i = 0; i < SK.Task.tris.nPnts; i++)
            {
                val = SK.Task.tris.Values[i];
                for (int u = 0; u < SK.Task.numUnits; u++)
                {
                    if (float.IsNaN(val[u]))
                        throw new Exception("Value is NaN for unit " + u.ToString() + " at (" + SK.Task.tris.Points[i].x.ToString() + " " + SK.Task.tris.Points[i].y.ToString() + ")");
                    if (val[u] <= 0.00001f) // Alt under -50 dB regnes som NoData
                        val[u] = -9999f;
                    else
                        val[u] = 10.0f * (float)Math.Log10(val[u]);
                }
            }
            SK.Task.triValsAredB = true;

            layers = new typmkLayer[2][];
            if (SK.Task.tris.Count > 0)
            {
                // Linker trekantsamling til MapKernel
                MK.TriAngs = SK.Task.tris;
                // Setter omsluttende rektangel for MapKernel
                bbox.Snap(1); // Avrunder trekantsamling-boks utover til nærmeste meter
                MapBox = bbox;
                // Lagrer til filer for hver kontur-definisjon for trekantmodell
                for (int i = 0; i < SK.Contours.Count; i++)
                {
                    skContour Con;
                    try
                    {
                        Con = SK.Contours[i];
                        Con.Layers = MK.ToContourMap(Con.UnitIdx, Con.Levels, Con.SmoothRadius, Con.CellSize, MapBox);
                        layers[i] = Con.Layers.ToArray();
                    }
                    catch (Exception ee)
                    {
                        MessageBox.Show("Error generating contours for contour no. " + i.ToString()
                            + ", \n Error message: " + ee.Message + ", \n InnerException: " + ee.InnerException
                            + ", \n Data: " + ee.Data, "Info", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                        //CU.SetError(Errors, ee, "Error generating contours for contour no. " + i.ToString(), SK.enMod2Str(enMod.Results));
                    }
                }
            }
            GC.Collect();
        }

        private static void CreateAndAddContours(string contour)
        {
            if (layers != null)
            {
                int index;
                if (contour == "Lden")
                    index = 0;
                else if (contour == "LdenDetail")
                    index = 1;
                else return;

                contours = new NoiseContour[layers[index].Length];

                for (int i = 0; i < layers[index].Length; i++) // layers
                {
                    float level;
                    List<double[,]> p = new List<double[,]>();
                    //double[,,] points= new double[k,layers[i].PolyLines[k].Length,2];
                    level = layers[index][i].Level;
                    for (int k = 0; k < layers[index][i].PolyLines.Count; k++) // lines in layer
                    {
                        double[,] points = new double[layers[index][i].PolyLines[k].Length, 2];
                        for (int j = 0; j < layers[index][i].PolyLines[k].Length; j++) // points in lines
                        {
                            try
                            {
                                points[j, 0] = layers[index][i].PolyLines[k][j].x;
                                points[j, 1] = layers[index][i].PolyLines[k][j].y;
                            }
                            catch { }
                        }
                        p.Add(points);
                    }

                    contours[i] = new NoiseContour(p, level);
                }
            }
        }

        private static void CreatingLdenContours(float mscale)
        {
            skContour Cont = new skContour();
            Cont.UnitIdx.Add(0); // Indeks til støyenhet støykotene skal lages for
            Cont.SmoothRadius = 10f; // Glatteradius 10 m
            Cont.CellSize = mscale / 1000f;
            //test 
            Cont.Levels.Add(new float[2]); // Lager koter for Lden 55 og 65 dB
            Cont.Levels[0][0] = 55f;
            Cont.Levels[0][1] = 65f;
            SK.Contours.Add(Cont);
        }

        private static void CreatingDetailedLdenContours(float mscale)
        {
            skContour Cont = new skContour();
            Cont.UnitIdx.Add(0); // Indeks til støyenhet støykotene skal lages for
            Cont.SmoothRadius = 10f; // Glatteradius 10 m
            Cont.CellSize = mscale / 1000f;
            Cont.Levels.Add(new float[15]); // Lager koter for Lden 35, 38, 41, 44, 47, 50, 53, 56, 59, 62, 65, 68, 71, 74 og 77dB
            for (int i = 0; i < 15; i++)
                Cont.Levels[0][i] = 35 + i * 3;
            SK.Contours.Add(Cont);
        }


        #region Methods for saving results to file

        // ************************************************************
        // DISSE RUTINENE ER NØDVENDIG FOR Å LAGRE RESULTATER TIL FILER
        // DET SKAL IKKE VÆRE NØDVENDIG Å GJØRE ENDRINGER HER
        // ************************************************************

        /// <summary>
        /// Lagrer alle resultater til fil(er) (uavhengig av status for beregningen).
        /// </summary>
        /// <param name="AutoSave">Skal være false ved manuelle anrop</param>
        /// <param name="stError">Er false bare dersom alle deloppdrag er ferdige, og resultatene fra disse kunne leses</param>
        /// <param name="SaveOK">Er true bare dersom resultatfiler ble korrekt skrevet</param>
        /// <param name="STnum">Hvis innhold, skriv ut resultat bare for oppgitt deloppdrag</param>
        public static void SaveResults(skSubTask subtask, Queue<typError> Errors, out bool stError, out bool SaveOK)
        {
            string MetN = "SaveResults";
            stError = false;
            SaveOK = false;

            int STno = 0, i, u;
            float[] val;
            bool thisOK;
            typBBox bbox;

            string[] BFilename = null; int Brecnum = 0, Bcnt = 0; long BcntPos = 0;
            double Bxmin = 0.0, Bxmax = 0.0;
            double Bymin = 0.0, Bymax = 0.0;
            double Bzmin = 0.0, Bzmax = 0.0;
            double[] Bmmin = null; double[] Bmmax = null;
            FileStream[] Bfs = null; BinaryWriter[] Bbw = null;
            FileStream[] Bfs2 = null; BinaryWriter[] Bbw2 = null;
            FileStream BfsB = null; BinaryWriter BbwB = null;
            List<string> BNames = null; List<string> BTypes = null; List<string> BFormats = null;
            List<ArrayList> BValues = null;

            try
            {
                if (!InitSaveFpnts(STno, out BFilename, out Brecnum, out BcntPos, out Bcnt,
                                   out Bxmin, out Bxmax,
                                   out Bymin, out Bymax,
                                   out Bzmin, out Bzmax,
                                   out Bmmin, out Bmmax,
                                   out Bfs, out Bbw,
                                   out Bfs2, out Bbw2,
                                   out BfsB, out BbwB,
                                   out BNames, out BTypes, out BFormats,
                                   out BValues, Errors))
                    throw new Exception("Could not initialize files for buildings and facade points");

                try
                {
                    // Skriver resultater for bygninger
                    if (subtask.buildings.Count > 0)
                        if (!SaveBldFpnts(subtask, BFilename, ref Brecnum, ref Bcnt, ref Bxmin, ref Bxmax, ref Bymin, ref Bymax, ref Bzmin, ref Bzmax, Bmmin, Bmmax,
                                          Bfs, Bbw, Bfs2, Bbw2, BfsB, BbwB, BValues, Errors))
                            throw new Exception("Could not write to files for buildings or facade points");
                }
                catch (Exception ee)
                {
                    stError = true; // Flagger feil, men fortsetter løkke
                }

                if (stError) return;

                if (!FinishSaveFpnts(BFilename, ref BcntPos, ref Bcnt, ref Bxmin, ref Bxmax, ref Bymin, ref Bymax, ref Bzmin, ref Bzmax, ref Bmmin, ref Bmmax,
                                     Bfs, Bbw, Bfs2, Bbw2, BfsB, BbwB, BNames, BTypes, BFormats, BValues, Errors))
                    throw new Exception("Could not finalize files for buildings or facade points");
                 //Sletter evt. tomme resf_*.shp og resb.dat-filer
                if (Bcnt == 0)
                {
                    for (i = 0; i < SK.Task.numUnits; i++)
                    {
                        CU.DeleteAnyFile(BFilename[i] + ".shp", Errors, SK.enMod2Str(enMod.Results));
                        CU.DeleteAnyFile(BFilename[i] + ".shx", Errors, SK.enMod2Str(enMod.Results));
                        CU.DeleteAnyFile(BFilename[i] + ".dbf", Errors, SK.enMod2Str(enMod.Results));
                    }
                    CU.DeleteAnyFile(SK.MainPath + "resb.dat", Errors, SK.enMod2Str(enMod.Results));
                }

                // Disse nullstilles uansett
                SK.Task.nB = 0;
                SK.Task.nfp = 0;

                GC.Collect();

                SK.Task.tris.Refresh(); // For å bygge topografisk indeks som brukes av MapKernel
                SK.Task.tris.CheckGridPoints(SK.Task.GrdDef, 0, 0); // For at IsGrdP-listen skal ha rett lengde

                // Må konvertere til dB
                //for (i = 0; i < SK.Task.np; i++)
                //{
                //    for (u = 0; u < SK.Task.numUnits; u++)
                //    {
                //        if (SK.Task.resultpoints[i, u] <= 0.00001f) // Alt under -50 dB regnes som NoData
                //            SK.Task.resultpoints[i, u] = -50.0f;
                //        else
                //            SK.Task.resultpoints[i, u] = 10.0f * (float)Math.Log10(SK.Task.resultpoints[i, u]);
                //    }
                //}

                bbox = SK.Task.tris.BBox;
                for (i = 0; i < SK.Task.tris.nPnts; i++)
                {
                    val = SK.Task.tris.Values[i];
                    for (u = 0; u < SK.Task.numUnits; u++)
                    {
                        if (float.IsNaN(val[u]))
                            throw new Exception("Value is NaN for unit " + u.ToString() + " at (" + SK.Task.tris.Points[i].x.ToString() + " " + SK.Task.tris.Points[i].y.ToString() + ")");
                        if (val[u] <= 0.00001f) // Alt under -50 dB regnes som NoData
                            val[u] = -9999f;
                        else
                            val[u] = 10.0f * (float)Math.Log10(val[u]);
                    }
                }
                SK.Task.triValsAredB = true;

                // Lagrer til filer for hver enhet i punktsamling separat
                SaveOK = true;
                for (i = 0; i < SK.Task.numUnits; i++)
                {
                    thisOK = SaveUnit(i, STno, stError, Errors);
                    SaveOK = SaveOK && thisOK;
                }

                if (SK.Task.tris.Count > 0)
                {
                    // Linker trekantsamling til MapKernel
                    MK.TriAngs = SK.Task.tris;
                    // Setter omsluttende rektangel for MapKernel
                    bbox.Snap(1); // Avrunder trekantsamling-boks utover til nærmeste meter
                    MapBox = bbox;
                    // Lagrer til filer for hver kontur-definisjon for trekantmodell
                    for (i = 0; i < SK.Contours.Count; i++)
                    {
                        thisOK = SaveContour(i, stError, Errors);
                        SaveOK = SaveOK && thisOK;
                    }
                }

                GC.Collect();

            }
            catch (Exception ee)
            {
                SaveOK = false;
            }
            finally
            {
            }
        }

        /// <summary>
        /// Lagrer alle aktuelle resultatfiler for en gitt enhet
        /// </summary>
        /// <param name="UnitNum">Enhet nummer</param>
        /// <param name="STno">Hvis 0 eller større, legges dette tallet til filnavnene</param>
        /// <param name="stError">Er false bare dersom alle deloppdrag er ferdige, og resultatene fra disse kunne leses</param>
        /// <returns>False dersom det oppsto en feil, true hvis alt går feilfritt</returns>
        static bool SaveUnit(int UnitNum, int STno, bool stError, Queue<typError> Errors)
        {
            if (SK.Task.np == 0) return true;
            string Filename = "";
            bool NoDet = (SK.Task.numDetail == 0);
            int DetNo = SK.Task.numDetail - 1;
            List<string> Names = new List<string>(), Types = new List<string>(), Formats = new List<string>();
            List<ArrayList> Values = new List<ArrayList>();

            while (NoDet || (DetNo >= 0))
            {
                if (NoDet)
                {
                    Filename = SK.MainPath + "resp_" + UnitNum.ToString();
                }
                else
                {
                    Filename = SK.MainPath + "resp_" + UnitNum.ToString() + "_" + DetNo.ToString();
                }
                if (STno >= 0) Filename += "_" + STno.ToString();

                Names.Clear(); Types.Clear(); Formats.Clear();
                Names.Add("ID"); Types.Add("N"); Formats.Add("10");
                Names.Add("Facade"); Types.Add("L"); Formats.Add("");
                Values.Clear();

                FileStream fs = null;
                FileStream fs2 = null;
                BinaryWriter bw = null;
                BinaryWriter bw2 = null;
                try
                {
                    fs = File.Open(Filename + ".shp", FileMode.Create);
                    fs2 = File.Open(Filename + ".shx", FileMode.Create);
                    bw = new BinaryWriter(fs);
                    bw2 = new BinaryWriter(fs2);
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error opening Shape file '" + Filename + ".*' for writing", SK.enMod2Str(enMod.Results)); return false;
                }
                byte[] Header;
                double xmin = 1e38, ymin = 1e38, zmin = 1e38, mmin = 1e38, xmax = -xmin, ymax = -ymin, zmax = -zmin, mmax = -mmin;
                double dummy = 0; // Z-verdi er ikke i bruk
                Hashtable[] Params = new Hashtable[SK.Task.np];
                Hashtable[] ParTyp = new Hashtable[SK.Task.np]; // B = Bool, I = Integer, D = Double, S(N) = string(N)
                try
                {
                    int i = -1, j;
                    // Skriver dummy headere
                    bw.Write(new byte[100]); bw2.Write(new byte[100]);
                    for (i = 0; i < SK.Task.np; i++)
                    {
                        // Skriver til index (posisjon og lengde)
                        bw2.Write(CU.B2L((int)(fs.Position / 2)));
                        bw2.Write(CU.B2L(36 / 2));
                        // Skriver header til hovedfil (1-basert teller og lengde)
                        bw.Write(CU.B2L(i + 1));
                        bw.Write(CU.B2L(36 / 2));
                        // Skriver record til hovedfil (type, X, Y, Z, M)
                        bw.Write((int)11);
                        bw.Write(SK.Task.points[i, 0]);
                        bw.Write(SK.Task.points[i, 1]);
                        xmin = Math.Min(SK.Task.points[i, 0], xmin); ymin = Math.Min(SK.Task.points[i, 1], ymin);
                        xmax = Math.Max(SK.Task.points[i, 0], xmax); ymax = Math.Max(SK.Task.points[i, 1], ymax);
                        bw.Write(dummy);
                        zmin = 0.0; zmax = 0.0;
                        bw.Write((double)(SK.Task.resultpoints[i, UnitNum]));
                        mmin = Math.Min(SK.Task.resultpoints[i, UnitNum], mmin);
                        mmax = Math.Max(SK.Task.resultpoints[i, UnitNum], mmax);
                        // Skriver attributter til Hashtable for senere skriving til .DBF-fil
                        Values.Add(new ArrayList()); j = Values.Count - 1;
                        Values[j].Add(SK.Task.pntID[i]);
                        Values[j].Add(SK.Task.facade[i]);
                        // LATER: Legg til details her...?
                    }
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error writing to Shape file '" + Filename + ".*'", SK.enMod2Str(enMod.Results)); return false;
                }
                // Spoler tilbake og skriver hovedheadere
                try
                {
                    Header = GetHeader((int)fs.Position, 11, xmin, xmax, ymin, ymax, zmin, zmax, mmin, mmax);
                    bw.Seek(0, SeekOrigin.Begin);
                    bw.Write(Header);
                    Header = GetHeader((int)fs2.Position, 11, xmin, xmax, ymin, ymax, zmin, zmax, mmin, mmax);
                    bw2.Seek(0, SeekOrigin.Begin);
                    bw2.Write(Header);
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error writing main header to Shape file '" + Filename + ".*'", SK.enMod2Str(enMod.Results)); return false;
                }
                if (bw2 != null) bw2.Close();
                if (fs2 != null) fs2.Close();
                if (bw != null) bw.Close();
                if (fs != null) fs.Close();
                // Skriver til .DBF-fil
                if (!Geo.WriteDBF(Filename + ".dbf", Names, Types, Formats, Values, Errors, SK.enMod2Str(enMod.Results))) return false;
                DetNo--; NoDet = false;
            }
            return true;
        }


        /// <summary>Gjør klar til lagring av bygninger med fasadepunkter</summary>
        /// <param name="STno">Hvis 0 eller større, legges dette tallet til filnavnene</param>
        /// <returns>False dersom det oppsto en feil, true hvis alt går feilfritt</returns>
        static bool InitSaveFpnts(int STno,
                           out string[] Filename, out int recnum, out long BcntPos, out int Bcnt,
                           out double xmin, out double xmax,
                           out double ymin, out double ymax,
                           out double zmin, out double zmax,
                           out double[] mmin, out double[] mmax,
                           out FileStream[] fs, out BinaryWriter[] bw,
                           out FileStream[] fs2, out BinaryWriter[] bw2,
                           out FileStream fsB, out BinaryWriter bwB,
                           out List<string> Names, out List<string> Types, out List<string> Formats,
                           out List<ArrayList> Values,
                           Queue<typError> Errors)
        {

            int iu = 0, nu = SK.Task.numUnits;
            Filename = new string[nu];
            recnum = 0;
            BcntPos = 0;
            Bcnt = 0;
            xmin = 1e38; ymin = 1e38; zmin = 1e38; mmin = new double[nu];
            xmax = -xmin; ymax = -ymin; zmax = -zmin; mmax = new double[nu];
            fs = new FileStream[nu];
            fs2 = new FileStream[nu];
            bw = new BinaryWriter[nu];
            bw2 = new BinaryWriter[nu];
            fsB = null;
            bwB = null;
            Names = new List<string>();
            Types = new List<string>();
            Formats = new List<string>();
            Values = new List<ArrayList>();

            Names.Clear(); Types.Clear(); Formats.Clear();
            Names.Add("ID"); Types.Add("N"); Formats.Add("10");
            Names.Add("BuildID"); Types.Add("N"); Formats.Add("10");
            Names.Add("Facade"); Types.Add("L"); Formats.Add("");
            Names.Add("Ref"); Types.Add("C"); Formats.Add("12");

            for (iu = 0; iu < nu; iu++)
            {
                mmin[iu] = 1e38; mmax[iu] = -mmin[iu];
                Filename[iu] = SK.MainPath + "resf_" + iu.ToString();
                if (STno >= 0) Filename[iu] += "_" + STno.ToString();

                try
                {
                    fs[iu] = File.Open(Filename[iu] + ".shp", FileMode.Create);
                    fs2[iu] = File.Open(Filename[iu] + ".shx", FileMode.Create);
                    bw[iu] = new BinaryWriter(fs[iu]);
                    bw2[iu] = new BinaryWriter(fs2[iu]);
                    // Skriver dummy headere
                    bw[iu].Write(new byte[100]); bw2[iu].Write(new byte[100]);
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error opening/initializing Shape file '" + Filename[iu] + ".*'", SK.enMod2Str(enMod.Results));
                    for (iu = 0; iu < nu; iu++)
                    {
                        if (bw[iu] != null) bw[iu].Close();
                        if (bw2[iu] != null) bw2[iu].Close();
                        if (fs[iu] != null) fs[iu].Close();
                        if (fs2[iu] != null) fs2[iu].Close();
                    }
                    if (bwB != null) bwB.Close();
                    if (fsB != null) fsB.Close();
                    return false;
                }
            }

            return clsImport.BygningerTilFilÅpne(SK.MainPath + "resb.dat", SK.Task.config, Errors, out fsB, out bwB, out BcntPos);
        }


        /// <summary>
        /// Lagrer filer for en gitt kontur-definisjon
        /// </summary>
        /// <param name="ConNum">Konturnummer</param>
        /// <param name="stError">Er false bare dersom alle deloppdrag er ferdige, og resultatene fra disse kunne leses</param>
        /// <returns>False dersom det oppsto en feil, true hvis alt går feilfritt</returns>
        static bool SaveContour(int ConNum, bool stError, Queue<typError> Errors)
        {
            skContour Con;
            string Filename = "";
            List<string> Names = new List<string>(), Types = new List<string>(), Formats = new List<string>();
            List<ArrayList> Values = new List<ArrayList>();
            layers = new typmkLayer[2][];
            try
            {
                Con = SK.Contours[ConNum];
                Con.Layers = MK.ToContourMap(Con.UnitIdx, Con.Levels, Con.SmoothRadius, Con.CellSize, MapBox);
                layers[ConNum] = Con.Layers.ToArray();
            }
            catch (Exception ee)
            {
                CU.SetError(Errors, ee, "Error generating contours for contour no. " + ConNum.ToString(), SK.enMod2Str(enMod.Results));
                return false;
            }
            try
            {
                Names.Clear(); Types.Clear(); Formats.Clear();
                Names.Add("Unit"); Types.Add("C"); Formats.Add("50");
                Names.Add("Level"); Types.Add("N"); Formats.Add("8:2");
                Values.Clear();

                Filename = SK.MainPath + "resg_" + ConNum.ToString();
                FileStream fs = null;
                FileStream fs2 = null;
                BinaryWriter bw = null;
                BinaryWriter bw2 = null;
                try
                {
                    fs = File.Open(Filename + ".shp", FileMode.Create);
                    fs2 = File.Open(Filename + ".shx", FileMode.Create);
                    bw = new BinaryWriter(fs);
                    bw2 = new BinaryWriter(fs2);
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error opening Shape file '" + Filename + ".*' for writing", SK.enMod2Str(enMod.Results)); return false;
                }
                byte[] Header;
                Hashtable[] Params = new Hashtable[Con.Layers.Count];
                Hashtable[] ParTyp = new Hashtable[Con.Layers.Count]; // I = Integer, D = Double, S(N) = string(N)
                typBBox polybox, totalbox = new typBBox(1e38, -1e38, 1e38, -1e38);
                int[] parts;
                int i, j, k, npnts, nparts, polysize;
                StringBuilder unitstr;
                try
                {
                    // Skriver dummy headere
                    bw.Write(new byte[100]); bw2.Write(new byte[100]);
                    // Skriver ett og ett polygon
                    for (i = 0; i < Con.Layers.Count; i++)
                    {
                        
                        // Trenger å vite hvor mange punkt det er i polygonet, samt utstrekning
                        nparts = Con.Layers[i].PolyLines.Count;
                        npnts = 0;
                        polybox = new typBBox(1e38, -1e38, 1e38, -1e38);
                        parts = new int[nparts];
                        for (j = 0; j < nparts; j++)
                        {
                            parts[j] = npnts;
                            npnts += Con.Layers[i].PolyLines[j].Length;
                            for (k = 0; k < Con.Layers[i].PolyLines[j].Length; k++)
                                polybox.Extend(Con.Layers[i].PolyLines[j][k].x, Con.Layers[i].PolyLines[j][k].y);
                        }
                        totalbox.Extend(polybox);
                        polysize = 44 + 4 * nparts + 16 * npnts;
                        // Skriver til index (posisjon og lengde)
                        bw2.Write(CU.L2B((int)(fs.Position / 2)));
                        bw2.Write(CU.L2B(polysize / 2));
                        // Skriver header til hovedfil (1-basert teller og lengde)
                        bw.Write(CU.L2B(i + 1));
                        bw.Write(CU.L2B(polysize / 2));
                        // Skriver record til hovedfil (type, box, parts, points)
                        bw.Write((int)5);
                        bw.Write(polybox.Xlow);
                        bw.Write(polybox.Ylow);
                        bw.Write(polybox.Xhigh);
                        bw.Write(polybox.Yhigh);
                        bw.Write(nparts);
                        bw.Write(npnts);
                        for (j = 0; j < nparts; j++)
                            bw.Write(parts[j]);
                        for (j = 0; j < nparts; j++)
                            for (k = 0; k < Con.Layers[i].PolyLines[j].Length; k++)
                            {
                                bw.Write(Con.Layers[i].PolyLines[j][k].x);
                                bw.Write(Con.Layers[i].PolyLines[j][k].y);
                            }
                        // Skriver attributter til Hashtable for senere skriving til .DBF-fil
                        Values.Add(new ArrayList()); j = Values.Count - 1;
                        unitstr = new StringBuilder();
                        for (k = 0; k < Con.UnitIdx.Count; k++)
                        {
                            if (k > 0) unitstr.Append("/");
                            unitstr.Append(SK.enUnit2Str(SK.Task.Units[Con.UnitIdx[k]].Unit, SK.Task.Units[Con.UnitIdx[k]].FW, SK.Task.Units[Con.UnitIdx[k]].TW));
                        }
                        Values[j].Add(unitstr.ToString());
                        Values[j].Add((double)Con.Layers[i].Level);
                    }
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error writing to Shape file '" + Filename + ".*'", SK.enMod2Str(enMod.Results)); return false;
                }
                // Spoler tilbake og skriver hovedheadere
                try
                {
                    Header = GetHeader((int)fs.Position, 5, totalbox.Xlow, totalbox.Xhigh, totalbox.Ylow, totalbox.Yhigh, 0, 0, 0, 0);
                    bw.Seek(0, SeekOrigin.Begin);
                    bw.Write(Header);
                    Header = GetHeader((int)fs2.Position, 5, totalbox.Xlow, totalbox.Xhigh, totalbox.Ylow, totalbox.Yhigh, 0, 0, 0, 0);
                    bw2.Seek(0, SeekOrigin.Begin);
                    bw2.Write(Header);
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error writing main header to Shape file '" + Filename + ".*'", SK.enMod2Str(enMod.Results)); return false;
                }
                if (bw2 != null) bw2.Close();
                if (fs2 != null) fs2.Close();
                if (bw != null) bw.Close();
                if (fs != null) fs.Close();
                // Skriver til .DBF-fil
                if (!Geo.WriteDBF(Filename + ".dbf", Names, Types, Formats, Values, Errors, SK.enMod2Str(enMod.Results))) return false;
            }
            catch (Exception ee)
            {
                CU.SetError(Errors, ee, "Error writing to contour file '" + Filename + ".*'", SK.enMod2Str(enMod.Results));
                return false;
            }
            return true;
        }


        /// <summary>Lagrer alle bygninger og tilhørende fasadepunkt i et deloppdrag</summary>
        /// <returns>False dersom det oppsto en feil, true hvis alt går feilfritt</returns>
        static bool SaveBldFpnts(skSubTask subtask,
                          string[] Filename, ref int recnum, ref int Bcnt,
                          ref double xmin, ref double xmax,
                          ref double ymin, ref double ymax,
                          ref double zmin, ref double zmax,
                          double[] mmin, double[] mmax,
                          FileStream[] fs, BinaryWriter[] bw,
                          FileStream[] fs2, BinaryWriter[] bw2,
                          FileStream fsB, BinaryWriter bwB,
                          List<ArrayList> Values,
                          Queue<typError> Errors)
        {

            double dbval;
            Bygning B;
            int iBld, ifp, j, iu = 0, nu = subtask.numUnits;
            try
            {
                for (iBld = 0; iBld < subtask.buildings.Count; iBld++)
                {
                    B = subtask.buildings[iBld];
                    if (B.fasadepunkt.Count == 0) continue; // Skriver ikke "tomme" bygninger
                    for (ifp = 0; ifp < B.fasadepunkt.Count; ifp++)
                    {
                        recnum++;
                        for (iu = 0; iu < nu; iu++)
                        {
                            // Skriver til index (posisjon og lengde)
                            bw2[iu].Write(CU.B2L((int)(fs[iu].Position / 2)));
                            bw2[iu].Write(CU.B2L(36 / 2));
                            // Skriver header til hovedfil (1-basert teller og lengde)
                            bw[iu].Write(CU.B2L(recnum));
                            bw[iu].Write(CU.B2L(36 / 2));
                            // Skriver record til hovedfil (type, X, Y, Z, M)
                            bw[iu].Write((int)11);
                            bw[iu].Write(B.fasadepunkt[ifp].sted.posisjonE);
                            bw[iu].Write(B.fasadepunkt[ifp].sted.posisjonN);
                            xmin = Math.Min(B.fasadepunkt[ifp].sted.posisjonE, xmin); ymin = Math.Min(B.fasadepunkt[ifp].sted.posisjonN, ymin);
                            xmax = Math.Max(B.fasadepunkt[ifp].sted.posisjonE, xmax); ymax = Math.Max(B.fasadepunkt[ifp].sted.posisjonN, ymax);
                            bw[iu].Write(B.fasadepunkt[ifp].sted.h);
                            zmin = Math.Min(B.fasadepunkt[ifp].sted.h, zmin);
                            zmax = Math.Max(B.fasadepunkt[ifp].sted.h, zmax);
                            // Henter dB-verdier fra FØRSTE Situasjon
                            // NB: Forutsetter at nivå[] inneholder de samme enheter
                            //     som definert i task.xml, og i samme rekkefølge!
                            dbval = B.fasadepunkt[ifp].utesituasjoner[0].utestøy.nivå[iu].verdi;
                            bw[iu].Write(dbval);
                            mmin[iu] = Math.Min(dbval, mmin[iu]);
                            mmax[iu] = Math.Max(dbval, mmax[iu]);
                        }
                        // Skriver attributter til Hashtable for senere skriving til .DBF-fil
                        // (Denne er felles for alle enheter)
                        Values.Add(new ArrayList()); j = Values.Count - 1;
                        Values[j].Add(Convert.ToInt64(B.beskrivelse));
                        Values[j].Add(B.bygningsnummer);
                        Values[j].Add(true);
                        Values[j].Add(B.fasadepunkt[ifp].sted.skissereferanse);
                    }
                    B.beskrivelse = "Autogenerert fra NorCalc";
                    Bcnt++;
                    if (!clsImport.BygningTilFilSkriv(fsB, bwB, B, Errors))
                    {
                        for (iu = 0; iu < nu; iu++)
                        {
                            if (bw[iu] != null) bw[iu].Close();
                            if (bw2[iu] != null) bw2[iu].Close();
                            if (fs[iu] != null) fs[iu].Close();
                            if (fs2[iu] != null) fs2[iu].Close();
                        }
                        if (bwB != null) bwB.Close();
                        if (fsB != null) fsB.Close();
                        return false;
                    }
                }
            }
            catch (Exception ee)
            {
                CU.SetError(Errors, ee, "Error writing to Shape file '" + Filename[iu] + ".*'", SK.enMod2Str(enMod.Results));
                for (iu = 0; iu < nu; iu++)
                {
                    if (bw[iu] != null) bw[iu].Close();
                    if (bw2[iu] != null) bw2[iu].Close();
                    if (fs[iu] != null) fs[iu].Close();
                    if (fs2[iu] != null) fs2[iu].Close();
                }
                if (bwB != null) bwB.Close();
                if (fsB != null) fsB.Close();
                return false;
            }
            return true;
        }

        /// <summary>Lagrer alle aktuelle resultatfiler for en gitt enhet</summary>
        /// <returns>False dersom det oppsto en feil, true hvis alt går feilfritt</returns>
        static bool FinishSaveFpnts(string[] Filename, ref long BcntPos, ref int Bcnt,
                             ref double xmin, ref double xmax,
                             ref double ymin, ref double ymax,
                             ref double zmin, ref double zmax,
                             ref double[] mmin, ref double[] mmax,
                             FileStream[] fs, BinaryWriter[] bw,
                             FileStream[] fs2, BinaryWriter[] bw2,
                             FileStream fsB, BinaryWriter bwB,
                             List<string> Names, List<string> Types, List<string> Formats,
                             List<ArrayList> Values,
                             Queue<typError> Errors)
        {

            byte[] Header;
            int iu = 0, nu = SK.Task.numUnits;
            // Spoler tilbake og skriver hovedheadere
            try
            {
                for (iu = 0; iu < nu; iu++)
                {
                    Header = GetHeader((int)fs[iu].Position, 11, xmin, xmax, ymin, ymax, zmin, zmax, mmin[iu], mmax[iu]);
                    bw[iu].Seek(0, SeekOrigin.Begin);
                    bw[iu].Write(Header);
                    Header = GetHeader((int)fs2[iu].Position, 11, xmin, xmax, ymin, ymax, zmin, zmax, mmin[iu], mmax[iu]);
                    bw2[iu].Seek(0, SeekOrigin.Begin);
                    bw2[iu].Write(Header);
                    // Skriver til .DBF-fil
                    if (!Geo.WriteDBF(Filename[iu] + ".dbf", Names, Types, Formats, Values, Errors, SK.enMod2Str(enMod.Results))) return false;
                }

                return clsImport.BygningerTilFilLukk(fsB, bwB, BcntPos, Bcnt, Errors);

            }
            catch (Exception ee)
            {
                CU.SetError(Errors, ee, "Error writing main header to Shape file '" + Filename[iu] + ".*'", SK.enMod2Str(enMod.Results));
                return false;
            }
            finally
            {
                for (iu = 0; iu < nu; iu++)
                {
                    if (bw[iu] != null) bw[iu].Close();
                    if (bw2[iu] != null) bw2[iu].Close();
                    if (fs[iu] != null) fs[iu].Close();
                    if (fs2[iu] != null) fs2[iu].Close();
                }
                if (bwB != null) bwB.Close();
                if (fsB != null) fsB.Close();
            }
        }

        /// <summary>
        /// Returnerer en Shape Main File Header (100 bytes)
        /// </summary>
        /// <param name="FileSize">Størrelsen på Shape-filen</param>
        /// <param name="ShapeType">Shape Type som filen inneholder</param>
        /// <param name="xmin">Minste x-verdi</param>
        /// <param name="xmax">Største x-verdi</param>
        /// <param name="ymin">Minste y-verdi</param>
        /// <param name="ymax">Største y-verdi</param>
        /// <param name="zmin">Minste z-verdi</param>
        /// <param name="zmax">Største z-verdi</param>
        /// <param name="mmin">Minste measure-verdi</param>
        /// <param name="mmax">Største measure-verdi</param>
        /// <returns>Main header (100 byte array)</returns>
        static byte[] GetHeader(int FileSize, int ShapeType,
                         double xmin, double xmax,
                         double ymin, double ymax,
                         double zmin, double zmax,
                         double mmin, double mmax)
        {
            byte[] result = new byte[100];
            BitConverter.GetBytes(CU.L2B((int)9994)).CopyTo(result, 0);
            BitConverter.GetBytes(CU.L2B(FileSize / 2)).CopyTo(result, 24);
            BitConverter.GetBytes((int)1000).CopyTo(result, 28);
            BitConverter.GetBytes(ShapeType).CopyTo(result, 32);
            BitConverter.GetBytes(xmin).CopyTo(result, 36);
            BitConverter.GetBytes(ymin).CopyTo(result, 44);
            BitConverter.GetBytes(xmax).CopyTo(result, 52);
            BitConverter.GetBytes(ymax).CopyTo(result, 60);
            BitConverter.GetBytes(zmin).CopyTo(result, 68);
            BitConverter.GetBytes(zmax).CopyTo(result, 76);
            BitConverter.GetBytes(mmin).CopyTo(result, 84);
            BitConverter.GetBytes(mmax).CopyTo(result, 92);
            return result;
        }

        #endregion


        
    }
}
