﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace NoiseComputation
{
    public class NoiseRoad
    {
        public double[,] Points;
        public float width;
        public string activityType; // Heavy, Normal, Dumper or User Defined
        public float velocity;
        public float ydt;
        public bool isTunnel;
        public bool isBridge;
        public string trafficDistributionType;
        public double[] frequencySpectra;
        public double[] vehicleDistribution;

        //public NoiseRoad(double[,] points, float width_, string activityType_, float velocity_, float ydt_, bool isTunnel_, bool isBridge_, string trafficDistribution_)
        //{
        //    Points = points;
        //    width = width_;
        //    activityType = activityType_;
        //    velocity = velocity_;
        //    ydt = ydt_;
        //    isTunnel = isTunnel_;
        //    isBridge = isBridge_;
        //    trafficDistributionType = trafficDistribution_;
        //}

        public NoiseRoad(double[,] points, float width_, string activityType_, float velocity_, float ydt_, bool isTunnel_, bool isBridge_, string trafficDistribution_, double[] frequencySpectra_, double[] vehicleDistribution_)
        {
            Points = points;
            width = width_;
            activityType = activityType_;
            velocity = velocity_;
            ydt = ydt_;
            isTunnel = isTunnel_;
            isBridge = isBridge_;
            trafficDistributionType = trafficDistribution_;
            frequencySpectra = frequencySpectra_;
            vehicleDistribution = vehicleDistribution_;

            //check frequency spectra
            if (activityType == "UserDefined")
            {
            if (frequencySpectra_.Length == 27)
                frequencySpectra = frequencySpectra_;
            else if (frequencySpectra_.Length == 9)
                frequencySpectra = convertToThirdOctave(frequencySpectra_);
            else MessageBox.Show("Error in frequency spectra for noise source! Needs to be either 9 octaves or 27 1/3-octaves.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }

            // add checks to check for valid distribution input
            if (trafficDistributionType == "UserDefined" && vehicleDistribution == null)
            {
                MessageBox.Show("Error! Vehicle distribution is null", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }
            if (trafficDistributionType == "UserDefined") 
            {
                if (vehicleDistribution[9] != ydt )
                {
                    MessageBox.Show("Vehicle distribution total is different from AAD.!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                    return;
                }
                double ydtMonFri = vehicleDistribution[9];
                double ydtSat = vehicleDistribution[3] + vehicleDistribution[4] + vehicleDistribution[5];
                double ydtSun = vehicleDistribution[6] + vehicleDistribution[7] + vehicleDistribution[8];
                if (ydtMonFri > ydt || ydtSat > ydt || ydtSun > ydt)
                {
                    MessageBox.Show("Vehicle distribution numbers are larger than AAD.!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                    return;
                }
               
            }
        }

        private double[] convertToThirdOctave(double[] frequencySpectra_)
        {
            // converting from whole octaves to 1/3 of an octave
            double third = (double)1 / 3;
            double reduceBy = 10 * Math.Log10(third);
            double[] freqSpectra = new double[27];

            for (int i = 0; i < 9; i++)
                for (int j = 0; j < 3; j++)
                    freqSpectra[(3 * i) + j] = Math.Pow(10, ((frequencySpectra_[i] + reduceBy) / 10));

            return freqSpectra;
        }
    }

    public class NoiseSource
    {
        public double[,] Points;
        public string sourceType; // unknown, road, industry, tunnel
        public string activityType; // Heavy, Normal or User Defined, Normal will probably not be used, add Crusher and Drilling
        public string timeDistributionType; // true for general use throughout activityType week, false for weekday, daytime use only. Gives TFac
        public string geometry; // point, line, area or volume
        /// <param name="amount_">Aktivitetstype</param>
        public float amount; // hours if usehours, aadt if ydt, number of events? if events
        public string amountUnit; // events, usehours or ydt
        public double[] frequencySpectra;
        public double[] timeDistribution;

        // For noise sources that are not roads: does not need speed information
        public NoiseSource(double[,] Points_, string activityType_, string timeDistributionType_, string geometry_, float amount_, double[] frequencySpectra_)
        {
            Points = Points_;
            sourceType = "industry";
            activityType = activityType_;
            timeDistributionType = timeDistributionType_;
            geometry = geometry_;
            amount = amount_;
            amountUnit = "usehours";
            if (frequencySpectra_.Length == 27)
                frequencySpectra = frequencySpectra_;
            else if (frequencySpectra_.Length == 9)
                frequencySpectra = convertToThirdOctave(frequencySpectra_);
            else MessageBox.Show("Error in frequency spectra for noise source! Needs to be either 9 octaves or 27 1/3-octaves.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
        }

        // For noise sources that are not roads: does not need speed information
        public NoiseSource(double[,] Points_, string activityType_, string timeDistributionType_, string geometry_, float amount_, double[] frequencySpectra_, double[] timeDistribution_)
        {
            Points = Points_;
            sourceType = "industry";
            activityType = activityType_;
            timeDistributionType = timeDistributionType_;
            geometry = geometry_;
            amount = amount_;
            amountUnit = "usehours";
            timeDistribution = timeDistribution_;

            if (frequencySpectra_.Length == 27)
                frequencySpectra = frequencySpectra_;
            else if (frequencySpectra_.Length == 9)
                frequencySpectra = convertToThirdOctave(frequencySpectra_);
            else MessageBox.Show("Error in frequency spectra for noise source! Needs to be either 9 octaves or 27 1/3-octaves.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);


            if (timeDistributionType == "UserDefined")
            {
                if (timeDistribution == null)
                {
                    MessageBox.Show("Error! Time distribution is null", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                    return;
                }
                if (timeDistribution[9] != amount)
                {
                    MessageBox.Show("Time distribution total is different from amount.!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                    return;
                }
                double ydtMonFri = timeDistribution[9];
                double ydtSat = timeDistribution[3] + timeDistribution[4] + timeDistribution[5];
                double ydtSun = timeDistribution[6] + timeDistribution[7] + timeDistribution[8];
                if (ydtMonFri > amount || ydtSat > amount || ydtSun > amount)
                {
                    MessageBox.Show("Time distribution numbers are larger than amount.!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                    return;
                }

                if (ydtMonFri > 24 || amount > 24 || ydtSat > 24 || ydtSun > 24)
                {
                    MessageBox.Show("Time distribution numbers are larger than amount.!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                    return;
                }
            }
        }



        private double[] convertToThirdOctave(double[] frequencySpectra_)
        {
            // converting from whole octaves to 1/3 of an octave
            double third = (double)1 / 3;
            double reduceBy = 10 * Math.Log10(third);
            double[] freqSpectra = new double[27];

            for (int i = 0; i < 9; i++)
                for (int j = 0; j < 3; j++)
                    freqSpectra[(3 * i) + j] = Math.Pow(10, ((frequencySpectra_[i] + reduceBy) / 10));

            return freqSpectra;
        }
    }

    public class NoiseMound
    {
        public double[,] Points;
        public float height;        // same as leftHeight for activityType mound or wall
        public float rightHeight;
        public float topWidth;
        public float bottomWidth;

        public NoiseMound(double[,] points, float height_, float topWidth_, float bottomWidth_, float rightHeight_)
        {
            Points = points;
            height = height_;
            topWidth = topWidth_;
            bottomWidth = bottomWidth_;
            rightHeight = rightHeight_;
        }
    }

    public class NoiseContour
    {
        public List<double[,]> Points;
        public float level;

        public NoiseContour(List<double[,]> points, float level_)
        {
            Points = points;
            level = level_;
        }
    }

    public class NoiseBuilding
    {
        public double[,] outLines;
        public double[,] inLines;
        public bool computeFacadePoints;

        public NoiseBuilding(double[,] outLines_, double[,] inLines_)
        {
            outLines = outLines_;
            inLines = inLines_;
        }

        public NoiseBuilding(double[,] outLines_, double[,] inLines_, bool computeFacadePoints_)
        {
            outLines = outLines_;
            inLines = inLines_;
            computeFacadePoints = computeFacadePoints_;
        }

    }
}
