﻿using System;
using System.IO;
using System.Text;
using System.Collections;
using System.Collections.Generic;

using CommonUtils;
using GeoUtils;
using SoundKernel;
using MapKernel;
using Innemodul;
using System.Windows.Forms;

namespace NoiseComputation
{
    public static class Program
    {
        /// <summary>Kotegenereringsobjekt</summary>
        static mkTriMesh MK = new mkTriMesh();

        /// <summary>Rektangel som omslutter ønsket utsnitt av kartet</summary>
        static typBBox MapBox;

        static int numCalcPoints;


        static typmkLayer[][] layers;

        public static List<NoiseContour[]> contours = new List<NoiseContour[]>();

        public static void Main(string[] args)
        {

            Queue<typError> Errors = new Queue<typError>();
            Queue<typError> Warnings = new Queue<typError>();

            double xorigo = 418208.5783;
            double yorigo = 6981471.4255;

            // Oppretter en (tom) topografimodell
            skTopography Topo = new skTopography();

            #region Creating Grid
            int nx = 68;
            int ny = 56;
            double x1 = 418208.5783;
            double y1 = 6981471.4255;
            double x2 = 418543.5783;
            double y2 = 6981746.4255;
            /// <summary>Vinkel for skråplan mot underliggende grid (0=hor., 90=vert.)</summary>
            int a = 90;
            /// <param name="m_">Default marktype</param>
            /// unknown, hard, soft, A, B, C, D, E, F, G, H, Index6
            enGroundType groundType = enGroundType.soft;
            /// <param name="s_">Default skoghøyde</param>
            int s = 0;
            /// <param name="r_">Default ruhet</param>
            int r = 0;

            skTopGrid Grid = new skTopGrid(x1, y1, x2, y2, a, nx, ny, a, groundType, s, r);

            Grid.zGrid = new float[nx, ny];
            Grid.Refresh();
            Topo.TopGrid.Add(Grid);
            #endregion

            #region Adding Calculation Points
            int numUnits = 2;
            numCalcPoints = 24;
            skSubTask currentSubTask = new skSubTask(0, 0, numCalcPoints, numUnits);

            currentSubTask.points[0, 0] = 418439.97; currentSubTask.points[0, 1] = 6981677.93; currentSubTask.hpoints[0] = 0;
            currentSubTask.points[1, 0] = 418437.375; currentSubTask.points[1, 1] = 6981674.085; currentSubTask.hpoints[1] = 0;
            currentSubTask.points[2, 0] = 418434.78; currentSubTask.points[2, 1] = 6981670.24; currentSubTask.hpoints[2] = 0;
            currentSubTask.points[3, 0] = 418432.53; currentSubTask.points[3, 1] = 6981666.01; currentSubTask.hpoints[3] = 0;
            currentSubTask.points[4, 0] = 418429.845; currentSubTask.points[4, 1] = 6981660.245; currentSubTask.hpoints[4] = 0;
            currentSubTask.points[5, 0] = 418427.16; currentSubTask.points[5, 1] = 6981654.48; currentSubTask.hpoints[5] = 0;

            // Same line, 4m higher up:
            currentSubTask.points[6, 0] = 418439.97; currentSubTask.points[6, 1] = 6981677.93; currentSubTask.hpoints[6] = 4;
            currentSubTask.points[7, 0] = 418437.375; currentSubTask.points[7, 1] = 6981674.085; currentSubTask.hpoints[7] = 4;
            currentSubTask.points[8, 0] = 418434.78; currentSubTask.points[8, 1] = 6981670.24; currentSubTask.hpoints[8] = 4;
            currentSubTask.points[9, 0] = 418432.53; currentSubTask.points[9, 1] = 6981666.01; currentSubTask.hpoints[9] = 4;
            currentSubTask.points[10, 0] = 418429.845; currentSubTask.points[10, 1] = 6981660.245; currentSubTask.hpoints[10] = 4;
            currentSubTask.points[11, 0] = 418427.16; currentSubTask.points[11, 1] = 6981654.48; currentSubTask.hpoints[11] = 4;

            // A different line, moving away from the noise source
            currentSubTask.points[12, 0] = 418495.94; currentSubTask.points[12, 1] = 6981606.73; currentSubTask.hpoints[12] = 0;
            currentSubTask.points[13, 0] = 418494.93; currentSubTask.points[13, 1] = 6981610.87; currentSubTask.hpoints[13] = 0;
            currentSubTask.points[14, 0] = 418493.86; currentSubTask.points[14, 1] = 6981612.06; currentSubTask.hpoints[14] = 0;
            currentSubTask.points[15, 0] = 418492.71; currentSubTask.points[15, 1] = 6981613.4; currentSubTask.hpoints[15] = 0;
            currentSubTask.points[16, 0] = 418491.58; currentSubTask.points[16, 1] = 6981614.45; currentSubTask.hpoints[16] = 0;
            currentSubTask.points[17, 0] = 418490.33; currentSubTask.points[17, 1] = 6981615.42; currentSubTask.hpoints[17] = 0;
            currentSubTask.points[18, 0] = 418486.43; currentSubTask.points[18, 1] = 6981617.28; currentSubTask.hpoints[18] = 0;
            currentSubTask.points[19, 0] = 418471.8; currentSubTask.points[19, 1] = 6981622.78; currentSubTask.hpoints[19] = 0;
            currentSubTask.points[20, 0] = 418446.08; currentSubTask.points[20, 1] = 6981633.97; currentSubTask.hpoints[20] = 0;
            currentSubTask.points[21, 0] = 418426.3; currentSubTask.points[21, 1] = 6981642.67; currentSubTask.hpoints[21] = 0;
            currentSubTask.points[22, 0] = 418417.26; currentSubTask.points[22, 1] = 6981648.1; currentSubTask.hpoints[22] = 0;
            currentSubTask.points[23, 0] = 418412; currentSubTask.points[23, 1] = 6981651.54; currentSubTask.hpoints[23] = 0;

            #endregion

            currentSubTask.Sources = new AListWrap(enFC.Source);


            // Trenger å lese denne filen for at fasadepunkt skal håndteres korrekt.
            if (!clsImport.LesKonfigurasjon("C:\\Dev\\Microstation_v8i\\DLL\\NoiseComputation\\bin\\Debug\\konfig.dat", out currentSubTask.config))
                throw new Exception("Feil ved lesing av konfig.dat");

            #region Adding Road Noise Source

            skTopRoad newRoad = new skTopRoad("Road0", 10, enHeightBase.sea, enGroundType.hard, "AB16", 1, 30);
 
            newRoad.pGon[0, 0] = 418548.641875835; newRoad.pGon[0, 1] = 6981739.26413355;
            newRoad.pGon[1, 0] = 418541.31; newRoad.pGon[1, 1] = 6981739.42;
            newRoad.pGon[2, 0] = 418518.34; newRoad.pGon[2, 1] = 6981738.37;
            newRoad.pGon[3, 0] = 418498.99; newRoad.pGon[3, 1] = 6981736.21;
            newRoad.pGon[4, 0] = 418485.44; newRoad.pGon[4, 1] = 6981734.39;
            newRoad.pGon[5, 0] = 418476.6; newRoad.pGon[5, 1] = 6981731.48;

            newRoad.pGon[6, 0] = 418468.61; newRoad.pGon[6, 1] = 6981727.44;
            newRoad.pGon[7, 0] = 418462.49; newRoad.pGon[7, 1] = 6981723.12;
            newRoad.pGon[8, 0] = 418456.12; newRoad.pGon[8, 1] = 6981717.62;
            newRoad.pGon[9, 0] = 418448.61; newRoad.pGon[9, 1] = 6981709.9;
            newRoad.pGon[10, 0] = 418442.67; newRoad.pGon[10, 1] = 6981703.08;

            newRoad.pGon[11, 0] = 418437.45; newRoad.pGon[11, 1] = 6981695.18;
            newRoad.pGon[12, 0] = 418431.36; newRoad.pGon[12, 1] = 6981685.99;
            newRoad.pGon[13, 0] = 418420.82; newRoad.pGon[13, 1] = 6981667.12;
            newRoad.pGon[14, 0] = 418412; newRoad.pGon[14, 1] = 6981651.54;
            newRoad.pGon[15, 0] = 418406.9; newRoad.pGon[15, 1] = 6981642.59;

            newRoad.pGon[16, 0] = 418399.49; newRoad.pGon[16, 1] = 6981631.51;
            newRoad.pGon[17, 0] = 418392.47; newRoad.pGon[17, 1] = 6981620.68;
            newRoad.pGon[18, 0] = 418381.75; newRoad.pGon[18, 1] = 6981604.59;
            newRoad.pGon[19, 0] = 418370.59; newRoad.pGon[19, 1] = 6981589.89;
            newRoad.pGon[20, 0] = 418361.81; newRoad.pGon[20, 1] = 6981579.08;

            newRoad.pGon[21, 0] = 418354.52; newRoad.pGon[21, 1] = 6981571.99;
            newRoad.pGon[22, 0] = 418347.62; newRoad.pGon[22, 1] = 6981566.21;
            newRoad.pGon[23, 0] = 418339.67; newRoad.pGon[23, 1] = 6981559.76;
            newRoad.pGon[24, 0] = 418329.42; newRoad.pGon[24, 1] = 6981552.74;
            newRoad.pGon[25, 0] = 418315.84; newRoad.pGon[25, 1] = 6981545.57;

            newRoad.pGon[26, 0] = 418295.84; newRoad.pGon[26, 1] = 6981533.39;
            newRoad.pGon[27, 0] = 418281.46; newRoad.pGon[27, 1] = 6981524.83;
            newRoad.pGon[28, 0] = 418270.18; newRoad.pGon[28, 1] = 6981517.09;
            newRoad.pGon[29, 0] = 418255.2; newRoad.pGon[29, 1] = 6981501.19;

            newRoad.hs = new float[33];

            newRoad.Refresh();
            Topo.TopRoads.Add(newRoad);

            // adding the noise source
            float amount = 200;
            float velocity = (float)(50 / 3.6); // 50km/t
            int numClasses = 3;
            skSource newSource = new skSource(enActType.road, enPlaceType.line, enQuantityUnit.ydt, amount, "Road0", 0, numClasses);

            float[] sourcePeriodTraffic = new float[3];
            sourcePeriodTraffic[0] = amount;         // Distribution Day
            sourcePeriodTraffic[1] = 0;              // Distribution Evening
            sourcePeriodTraffic[2] = 0;              // Distribution Night

            float lightVehicles, mediumVehicles, heavyVehicles;
            for (int period = 0; period < 3; ++period) //Døgnsegment: Dag, Kveld, Natt 
            {
                float periodTraffic = sourcePeriodTraffic[period] / newSource.Q;

                heavyVehicles = periodTraffic;              // all vehicles are heavy
                lightVehicles = 0;
                mediumVehicles = 0;

                for (int vehicleType = 0; vehicleType < 3; ++vehicleType) //Kategorier: Lett, Middels og Tung 
                {
                    float categoryTraffic = (vehicleType == 0 ? lightVehicles : ((vehicleType == 1) ? mediumVehicles : heavyVehicles));
                    for (int day = 0; day < 7; ++day) //Ukedag: Mandag til Søndag 
                        for (int month = 0; month < 12; ++month) //Måned: 1-12 
                        {
                            int numberOfWeekdays = 5;
                            float thisContribution = categoryTraffic / (numberOfWeekdays * 12);
                            if (day == 5 || day == 6)
                                thisContribution = 0;
                            newSource.TFac[period, day, month][vehicleType] = thisContribution;
                        }
                }
            }

            double[] frequencySpectra = new double[27];
            frequencySpectra[0] = 3.33333333333333E-21; frequencySpectra[1] = 3.33333333333333E-21;
            frequencySpectra[2] = 3.33333333333333E-21; frequencySpectra[3] = 12102597.5734628;
            frequencySpectra[4] = 12102597.5734628;     frequencySpectra[5] = 12102597.5734628;
            frequencySpectra[6] = 635153572.654415;     frequencySpectra[7] = 635153572.654415;
            frequencySpectra[8] = 635153572.654415;     frequencySpectra[9] = 2903211966.52028;
            frequencySpectra[10] = 2903211966.52028;    frequencySpectra[11] = 2903211966.52028;
            frequencySpectra[12] = 4007548115.39138;    frequencySpectra[13] = 4007548115.39138;
            frequencySpectra[14] = 4007548115.39138;    frequencySpectra[15] = 4818132569.15309;
            frequencySpectra[16] = 4818132569.15309;    frequencySpectra[17] = 4818132569.15309;
            frequencySpectra[18] = 3571731017.45868;    frequencySpectra[19] = 3571731017.45868;
            frequencySpectra[20] = 3571731017.45868;    frequencySpectra[21] = 856798594.256288;
            frequencySpectra[22] = 856798594.256288;    frequencySpectra[23] = 856798594.256288;
            frequencySpectra[24] = 71265402.9834077;    frequencySpectra[25] = 71265402.9834077;
            frequencySpectra[26] = 71265402.9834077;

            newSource.Params["RoadSrcSpec"] = populateRoadSourceSpec(frequencySpectra);

            newSource.RefID = "0 0.000 0.999"; // Dette vil vanligvis være "ReferanselenkeID FromMeasure Tomeasure"
            newSource.Params["RoadSurf"] = newRoad.rs;      // Vegdekke 
            newSource.Params["RoadWidth"] = newRoad.b;      // Vegbredde 
            newSource.Params["RoadSpeedLim"] = velocity;         // 40 km/t

            newSource.Vectors = newRoad.Vectors;
            newSource.h = newRoad.hs;
            newSource.Refresh();
            currentSubTask.Sources.Add(newSource);

            #endregion

            #region Defining a Building
            skTopBuilding newBuilding = new skTopBuilding(enHeightBase.sea, 7, 4, 1, 1);
            newBuilding.ID = 0;

            newBuilding.OutLines[0].X = 418468.3; newBuilding.OutLines[0].Y = 6981655.5; newBuilding.OutLines[0].Z = 3.5f;
            newBuilding.OutLines[1].X = 418469.9; newBuilding.OutLines[1].Y = 6981660.7; newBuilding.OutLines[1].Z = 3.5f;
            newBuilding.OutLines[2].X = 418477.9; newBuilding.OutLines[2].Y = 6981658.3; newBuilding.OutLines[2].Z = 3.5f;
            newBuilding.OutLines[3].X = 418473.9; newBuilding.OutLines[3].Y = 6981644.9; newBuilding.OutLines[3].Z = 3.5f;
            newBuilding.OutLines[4].X = 418459.5; newBuilding.OutLines[4].Y = 6981649.3; newBuilding.OutLines[4].Z = 3.5f;
            newBuilding.OutLines[5].X = 418462; newBuilding.OutLines[5].Y = 6981657.4; newBuilding.OutLines[5].Z = 3.5f;
            newBuilding.OutLines[6].X = 418468.3; newBuilding.OutLines[6].Y = 6981655.5; newBuilding.OutLines[6].Z = 3.5f;

            newBuilding.InLines[0].X = 418475.1; newBuilding.InLines[0].Y = 6981649; newBuilding.InLines[0].Z = 5.1f;
            newBuilding.InLines[1].X = 418460.7; newBuilding.InLines[1].Y = 6981653.3; newBuilding.InLines[1].Z = 5.1f;
            newBuilding.InLines[2].X = 418471.2; newBuilding.InLines[2].Y = 6981650.6; newBuilding.InLines[2].Z = 5.1f;
            newBuilding.InLines[3].X = 418473.9; newBuilding.InLines[3].Y = 6981659.5; newBuilding.InLines[3].Z = 5.1f;

            newBuilding.DoFPnts = true;
            newBuilding.Refresh();
            Topo.TopBuildings.Add(newBuilding);

            currentSubTask.sitidx.Add(0);
            #endregion

            #region All the other stuff

            // Genererer trekantmodell basert på innleste objekter 
            Top2Tri T2T = new Top2Tri(Topo, false);
            T2T.SetLimits(x1, x2, y1, y2, xorigo, yorigo);
            T2T.SetBldIDTrans(currentSubTask.buildings, currentSubTask.bldIDTrans, currentSubTask.ih, 0f, currentSubTask.config);
            if (!T2T.CreateTriangles(true, true, true, true, true, true, true, null))
            {
                //  Noe galt skjedde i forbindelse med triangulering 
                MessageBox.Show("Error in triangulation!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }

            List<conline> ConLines = new List<conline>();
            List<conpoly> ConPolys = new List<conpoly>();
            // Henter kanter fra triangulering av topografi (er nødvendig for å bygge opp trekantsamling
            // for beregningspunkt, for at kotegenerering ikke skal krysse kanter, som f.eks. skjermer og bygninger.
            // Dersom dette IKKE er et krav, kan dette utelates. ConLines og ConPolys skal da være tomme.

            // Henter alle linjer og alle TOMME polygoner ( = bygninger)
            T2T.GetLinesPolys(true, true, true, out ConLines, out ConPolys);
            // Tar med bare et utvalg av linjer og polygoner; Bare skjermer og bygningsomriss; Ikke taklinjer, etc
            // Må også sikre at evt. endringer i punktene i forbindelse med trekantsamling for beregning
            // IKKE påvirker punktene i topografien. Tar derfor KOPIER av linjene og polygonene
            List<conline> TmpLines = new List<conline>(ConLines.Count);
            List<conpoly> TmpPolys = new List<conpoly>(ConPolys.Count);
            for (int i = 0; i < ConLines.Count; i++)
            {
                if ((string)ConLines[i].par["Otyp"] != "TopoScreen") continue;
                TmpLines.Add(ConLines[i].Copy(false));
            }
            for (int i = 0; i < ConPolys.Count; i++)
            {
                if ((string)ConPolys[i].par["Otyp"] != "TopoBuilding") continue;
                TmpPolys.Add(ConPolys[i].Copy(false));
            }
            ConLines = TmpLines; ConLines.TrimExcess(); TmpLines = null;
            ConPolys = TmpPolys; ConPolys.TrimExcess(); TmpPolys = null;

            // Bestemmer de unike punktene i trekantsamlingen
            Topo.TopTriang.FindUniquePoints();
            // Må konvertere alle genererte objekter (trekanter og speil) til globale koordinater 
            // (ConLines og ConPolys forblir i lokale koordinater)
            Topo.TopTriang.ConvertToGlobal(xorigo, yorigo);
            Topo.ConvertMirrorsToGlobal(xorigo, yorigo);
            Topo.TopTriang.RefreshHash();
            Topo.Refresh();

            // Definerer beregningsenheter: Lden
            currentSubTask.Units[0].Unit = enSoundUnit.Leq;
            currentSubTask.Units[0].TW = enTimeWeight.DEN;
            currentSubTask.Units[0].FW = enFreqWeight.A;
            currentSubTask.Units[0].L1 = 10.0f;
            currentSubTask.Units[0].Toler = 0.2f;

            // Definerer Leq
            currentSubTask.Units[1].Unit = enSoundUnit.Leq;
            currentSubTask.Units[1].TW = enTimeWeight.none;
            currentSubTask.Units[1].FW = enFreqWeight.A;
            currentSubTask.Units[1].L1 = 10.0f;
            currentSubTask.Units[1].Toler = 0.2f;

            // Spesifiserer værforhold 
            skMetClass Met = new skMetClass();
            Met.TGrad = 0.0f;
            Met.Refresh(); // Sluttberegner interne egenskaper for været 
            currentSubTask.MET = new AListWrap(enFC.MET, Met); // Legger været til oppdraget 
            currentSubTask.MET.Refresh();

            // Dumper 3D-modell til fil (for debug innsyn)
            try //se if it is possible with activityType timecheck for this one
            {
                TriGrid2File.WriteParts(Topo, 1.0, true, false, -1);
            }
            catch
            {
                MessageBox.Show("Some parts are not defined correctly or not on the terrain", "Info", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }

            // Definerer beregningsgrid, 20 meter gridstørrelse
            griddef GrdDef = new griddef();
            GrdDef.x0 = x1; GrdDef.y0 = y1;
            GrdDef.xi = 1.0f; GrdDef.yi = 0.0f;
            GrdDef.Lu = x2 - x1; GrdDef.Lv = y2 - y1;
            GrdDef.nu = nx; GrdDef.nv = ny;
            currentSubTask.GrdDef = GrdDef;

            // Bygger trekantmodell for beregningspunkt.
            // Merk at hvis ConLines og ConPolys IKKE er tomme (men inneholder skjermer og bygningsomriss)
            // så vil beregningsgridet IKKE være regulært; Det vil bli etablert beregningspunkt BÅDE for 
            // punktene definert i GrdDef OG for alle punktene i ConLines og ConPolys.
            Grid2Tri G2T = new Grid2Tri(currentSubTask);
            if (!G2T.CreateCalcTri(GrdDef, ConLines, ConPolys, xorigo, yorigo, null))
                // Noe galt skjedde i forbindelse med triangulering
                throw new Exception("Feil ved triangulering av beregningsgrid");
            
            #endregion
            
            #region Computation
            currentSubTask.SubTopoHash = Topo.Hash;
            currentSubTask.Method = enMethod.N2000R;
            currentSubTask.ih = enHeightBase.terr;
            currentSubTask.Sources.Refresh();
            currentSubTask.Refresh();

            // Utfører beregning for beregningspunktene
            SK.TOPO = Topo;

            // Triks for å sikre at resultatene lagres korrekt
            SK.Task = skSubTask.Load(currentSubTask.Save());

            SK.Task.bldDetailedFP = new Dictionary<long, int>();
            SK.Task.bldDetailedFP[0] = 3;

            //resetting and enabling the computation to be run several times
            SK.Stopped = false;
            SK.Stopping = false;
            SK.Errors.Clear();
            SK.PercentFinished = 0;
            
            SK.MainStatus = enMStat.active;
            SK.threadcontrol = new System.Threading.EventWaitHandle(false, System.Threading.EventResetMode.ManualReset);
            SK.threadcontrol.Set();

            SK.RunSubTask();

            if (SK.MainStatus == enMStat.error)
            {
                //  Noe galt skjedde i forbindelse med beregning 
                MessageBox.Show("Feil oppsto under beregning: \n" + SK.Errors.Peek().Desc
                    + ", \n" + SK.Errors.Peek().ErrMsg + ", \n in module: "
                    + SK.Errors.Peek().Module + ", \n and in function: " + SK.Errors.Peek().Func, "Info", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
            else
            {
                float mscale = 5000f;
                SK.Contours = new List<skContour>();

                // Definerer kotesett for Lden
                skContour Cont = new skContour();
                Cont.UnitIdx.Add(0); // Indeks til støyenhet støykotene skal lages for
                Cont.SmoothRadius = 10f; // Glatteradius 10 m
                Cont.CellSize = mscale / 1000f;
                Cont.Levels.Add(new float[2]); // Lager koter for Lden 55 og 65 dB
                Cont.Levels[0][0] = 55f;
                Cont.Levels[0][1] = 65f;
                SK.Contours.Add(Cont);

                // Definerer detaljert kotesett for Lden
                skContour Cont2 = new skContour();
                Cont2.UnitIdx.Add(0); // Indeks til støyenhet støykotene skal lages for
                Cont2.SmoothRadius = 10f; // Glatteradius 10 m
                Cont2.CellSize = mscale / 1000f;
                Cont2.Levels.Add(new float[15]); // Lager koter for Lden 35, 38, 41, 44, 47, 50, 53, 56, 59, 62, 65, 68, 71, 74 og 77dB
                for (int i = 0; i < 15; i++)
                    Cont2.Levels[0][i] = 35 + i * 3;
                SK.Contours.Add(Cont2);

                layers = new typmkLayer[2][];

                bool stError, SaveOK;
                SaveResults((skSubTask)SK.Task, Errors, out stError, out SaveOK);

                calculateValues(SK.Task, numUnits);
                saveFacadeResults((skSubTask)SK.Task, Errors);

                for (int index = 0; index < 2; index++)
                {
                    if (layers[index] == null)
                        continue;
                    NoiseContour[] newContour = new NoiseContour[layers[index].Length];

                    for (int i = 0; i < layers[index].Length; i++) // layers
                    {
                        float level;
                        List<double[,]> p = new List<double[,]>();
                        //double[,,] points= new double[k,layers[i].PolyLines[k].Length,2];
                        level = layers[index][i].Level;
                        for (int k = 0; k < layers[index][i].PolyLines.Count; k++) // lines in layer
                        {
                            double[,] points = new double[layers[index][i].PolyLines[k].Length, 2];
                            for (int j = 0; j < layers[index][i].PolyLines[k].Length; j++) // points in lines
                            {
                                try
                                {
                                    points[j, 0] = layers[index][i].PolyLines[k][j].x;
                                    points[j, 1] = layers[index][i].PolyLines[k][j].y;
                                }
                                catch { }
                            }
                            p.Add(points);
                        }
                        newContour[i] = new NoiseContour(p, level);
                    }
                    contours.Add(newContour);
                }
            }
    

            #endregion

        }

        private static double[, , ,] populateRoadSourceSpec(double[] frequencySpectra)
        {
            // 3 kjøretøykategorier, 3 kildehøyder [1 cm, 30 cm, 75 cm], 27 frekvenser, 3 enheter [Lw, L5, Lmax]
            double[, , ,] sourceSpec = new double[3, 3, 27, 3];
            // Verdier trenger bare fylles ut for kjøretøykategorier og kildehøyder som skal brukes.
            // NB: Kategori 0 ("lette") bruker kildehøyde 0 (1 cm) og 1 (30 cm). 
            // Kategori 1 ("middels") og 2 ("tunge") bruker kildehøyde 0 (1 cm) og 2 (75 cm)! 
            // Pass på at ulike kategorier har utfylt data for sin(e) kildehøyder.
            // Enhet 0 må fylles ut for å få beregnet Lden og Leq. 
            // Enhet 2 må fylles ut for å få beregnet Lmax (hvis det er aktuelt). 
            // Enhet 1 (L5) kan utelates.

            for (int vehicle_i = 0; vehicle_i < 3; vehicle_i++)
            {
                if (vehicle_i == 0 || vehicle_i == 1)
                    continue;
                for (int sourceheight_i = 0; sourceheight_i < 3; sourceheight_i++)
                {
                    if (sourceheight_i == 1)
                        continue;
                    for (int freq_i = 0; freq_i < 27; freq_i++)
                    {
                        for (int unit_i = 0; unit_i < 3; unit_i++)
                        {
                            if (unit_i == 1 || unit_i == 2)
                                sourceSpec[vehicle_i, sourceheight_i, freq_i, unit_i] = 0;
                            //else fylle ut for unit = 1, og frekvenser
                            else
                                sourceSpec[vehicle_i, sourceheight_i, freq_i, unit_i] = frequencySpectra[freq_i];

                        }
                    }
                }
            }
            return sourceSpec;
        }

        private static void calculateValues(skTask task, int numUnits)
        {
            string lines = "";
            float[,] p = task.resultpoints;
            for (int i = 0; i < numCalcPoints; i++)
            {

                lines += "\r\n Point " + i + ":\t x: \t" +
                    task.points[i, 0] + "\t y: \t" + task.points[i, 1] +
                    "\t z: \t" + task.hpoints[i] + "\t Lden: \t" + p[i, 0] + "\t" + 
                    "\t Leq: \t" + p[i, 1] + "\t";
            }
           
            System.IO.File.WriteAllText(@"C:\TEMP\WriteLines.txt", lines);
        }

        private static void saveFacadeResults(skSubTask subtask, Queue<typError> Errors)
        {
            double dbval;
            Bygning B;
            double mmin = double.PositiveInfinity;
            double mmax = double.NegativeInfinity;

            int iBld, ifp, iu = 0, nu = subtask.numUnits;
            string lines = string.Empty;

            for (iBld = 0; iBld < subtask.buildings.Count; iBld++)
            {
                B = subtask.buildings[iBld];

                if (B.fasadepunkt.Count == 0) continue; // Skriver ikke "tomme" bygninger
                lines += "\r\n Building number: " + iBld + ", " + B.bygningsnummer;

                for (ifp = 0; ifp < B.fasadepunkt.Count; ifp++)
                    //recnum++;
                    for (iu = 0; iu < nu; iu++)
                    {
                        lines += "\r\n Facade point: \t" + ifp + "\tPosition East: \t" + B.fasadepunkt[ifp].sted.posisjonE +
                            "\tPosition North: \t" + B.fasadepunkt[ifp].sted.posisjonN + "\tHeight: \t" + B.fasadepunkt[ifp].sted.h +
                            "\tDecibel value: \t" + B.fasadepunkt[ifp].utesituasjoner[0].utestøy.nivå[iu].verdi +
                            "\tUnit: \t" + subtask.Units[iu].Unit.ToString();

                        dbval = B.fasadepunkt[ifp].utesituasjoner[0].utestøy.nivå[iu].verdi;
                        mmin = Math.Min(dbval, mmin);
                        mmax = Math.Max(dbval, mmax);
                    }
                lines += "\r\nMin Value: \t" + mmin + "\tMax Value: \t" + mmax;
                mmin = double.PositiveInfinity; mmax = double.NegativeInfinity;
            }
            System.IO.File.WriteAllText(@"C:\TEMP\WriteFacadePoints.txt", lines);
        }

        #region SaveFunctions
        // ************************************************************
        // DISSE RUTINENE ER NØDVENDIG FOR Å LAGRE RESULTATER TIL FILER
        // DET SKAL IKKE VÆRE NØDVENDIG Å GJØRE ENDRINGER HER
        // ************************************************************

        /// <summary>
        /// Lagrer alle resultater til fil(er) (uavhengig av status for beregningen).
        /// </summary>
        /// <param name="AutoSave">Skal være false ved manuelle anrop</param>
        /// <param name="stError">Er false bare dersom alle deloppdrag er ferdige, og resultatene fra disse kunne leses</param>
        /// <param name="SaveOK">Er true bare dersom resultatfiler ble korrekt skrevet</param>
        /// <param name="STnum">Hvis innhold, skriv ut resultat bare for oppgitt deloppdrag</param>
        public static void SaveResults(skSubTask subtask, Queue<typError> Errors, out bool stError, out bool SaveOK)
        {
            string MetN = "SaveResults";
            stError = false;
            SaveOK = false;

            int STno = 0, i, u;
            float[] val;
            bool thisOK;
            typBBox bbox;

            string[] BFilename = null; int Brecnum = 0, Bcnt = 0; long BcntPos = 0;
            double Bxmin = 0.0, Bxmax = 0.0;
            double Bymin = 0.0, Bymax = 0.0;
            double Bzmin = 0.0, Bzmax = 0.0;
            double[] Bmmin = null; double[] Bmmax = null;
            FileStream[] Bfs = null; BinaryWriter[] Bbw = null;
            FileStream[] Bfs2 = null; BinaryWriter[] Bbw2 = null;
            FileStream BfsB = null; BinaryWriter BbwB = null;
            List<string> BNames = null; List<string> BTypes = null; List<string> BFormats = null;
            List<ArrayList> BValues = null;

            try
            {
                if (!InitSaveFpnts(STno, out BFilename, out Brecnum, out BcntPos, out Bcnt,
                                   out Bxmin, out Bxmax,
                                   out Bymin, out Bymax,
                                   out Bzmin, out Bzmax,
                                   out Bmmin, out Bmmax,
                                   out Bfs, out Bbw,
                                   out Bfs2, out Bbw2,
                                   out BfsB, out BbwB,
                                   out BNames, out BTypes, out BFormats,
                                   out BValues, Errors))
                    throw new Exception("Could not initialize files for buildings and facade points");

                try
                {
                    // Skriver resultater for bygninger
                    if (subtask.buildings.Count > 0)
                        if (!SaveBldFpnts(subtask, BFilename, ref Brecnum, ref Bcnt, ref Bxmin, ref Bxmax, ref Bymin, ref Bymax, ref Bzmin, ref Bzmax, Bmmin, Bmmax,
                                          Bfs, Bbw, Bfs2, Bbw2, BfsB, BbwB, BValues, Errors))
                            throw new Exception("Could not write to files for buildings or facade points");
                }
                catch (Exception ee)
                {
                    stError = true; // Flagger feil, men fortsetter løkke
                }

                if (stError) return;

                if (!FinishSaveFpnts(BFilename, ref BcntPos, ref Bcnt, ref Bxmin, ref Bxmax, ref Bymin, ref Bymax, ref Bzmin, ref Bzmax, ref Bmmin, ref Bmmax,
                                     Bfs, Bbw, Bfs2, Bbw2, BfsB, BbwB, BNames, BTypes, BFormats, BValues, Errors))
                    throw new Exception("Could not finalize files for buildings or facade points");
                // Sletter evt. tomme resf_*.shp og resb.dat-filer
                if (Bcnt == 0)
                {
                    for (i = 0; i < SK.Task.numUnits; i++)
                    {
                        CU.DeleteAnyFile(BFilename[i] + ".shp", Errors, SK.enMod2Str(enMod.Results));
                        CU.DeleteAnyFile(BFilename[i] + ".shx", Errors, SK.enMod2Str(enMod.Results));
                        CU.DeleteAnyFile(BFilename[i] + ".dbf", Errors, SK.enMod2Str(enMod.Results));
                    }
                    CU.DeleteAnyFile(SK.MainPath + "resb.dat", Errors, SK.enMod2Str(enMod.Results));
                }

                // Disse nullstilles uansett
                SK.Task.nB = 0;
                SK.Task.nfp = 0;

                GC.Collect();

                SK.Task.tris.Refresh(); // For å bygge topografisk indeks som brukes av MapKernel
                SK.Task.tris.CheckGridPoints(SK.Task.GrdDef, 0, 0); // For at IsGrdP-listen skal ha rett lengde

                // Må konvertere til dB
                for (i = 0; i < SK.Task.np; i++)
                {
                    for (u = 0; u < SK.Task.numUnits; u++)
                    {
                        if (SK.Task.resultpoints[i, u] <= 0.00001f) // Alt under -50 dB regnes som NoData
                            SK.Task.resultpoints[i, u] = -50.0f;
                        else
                            SK.Task.resultpoints[i, u] = 10.0f * (float)Math.Log10(SK.Task.resultpoints[i, u]);
                    }
                }

                bbox = SK.Task.tris.BBox;
                for (i = 0; i < SK.Task.tris.nPnts; i++)
                {
                    val = SK.Task.tris.Values[i];
                    for (u = 0; u < SK.Task.numUnits; u++)
                    {
                        if (float.IsNaN(val[u]))
                            throw new Exception("Value is NaN for unit " + u.ToString() + " at (" + SK.Task.tris.Points[i].x.ToString() + " " + SK.Task.tris.Points[i].y.ToString() + ")");
                        if (val[u] <= 0.00001f) // Alt under -50 dB regnes som NoData
                            val[u] = -9999f;
                        else
                            val[u] = 10.0f * (float)Math.Log10(val[u]);
                    }
                }
                SK.Task.triValsAredB = true;

                // Lagrer til filer for hver enhet i punktsamling separat
                SaveOK = true;
                for (i = 0; i < SK.Task.numUnits; i++)
                {
                    thisOK = SaveUnit(i, STno, stError, Errors);
                    SaveOK = SaveOK && thisOK;
                }

                if (SK.Task.tris.Count > 0)
                {
                    // Linker trekantsamling til MapKernel
                    MK.TriAngs = SK.Task.tris;
                    // Setter omsluttende rektangel for MapKernel
                    bbox.Snap(1); // Avrunder trekantsamling-boks utover til nærmeste meter
                    MapBox = bbox;
                    // Lagrer til filer for hver kontur-definisjon for trekantmodell
                    for (i = 0; i < SK.Contours.Count; i++)
                    {
                        thisOK = SaveContour(i, stError, Errors);
                        SaveOK = SaveOK && thisOK;
                    }
                }

                GC.Collect();

            }
            catch (Exception ee)
            {
                SaveOK = false;
            }
            finally
            {
            }
        }

        /// <summary>
        /// Lagrer alle aktuelle resultatfiler for en gitt enhet
        /// </summary>
        /// <param name="UnitNum">Enhet nummer</param>
        /// <param name="STno">Hvis 0 eller større, legges dette tallet til filnavnene</param>
        /// <param name="stError">Er false bare dersom alle deloppdrag er ferdige, og resultatene fra disse kunne leses</param>
        /// <returns>False dersom det oppsto en feil, true hvis alt går feilfritt</returns>
        static bool SaveUnit(int UnitNum, int STno, bool stError, Queue<typError> Errors)
        {
            if (SK.Task.np == 0) return true;
            string Filename = "";
            bool NoDet = (SK.Task.numDetail == 0);
            int DetNo = SK.Task.numDetail - 1;
            List<string> Names = new List<string>(), Types = new List<string>(), Formats = new List<string>();
            List<ArrayList> Values = new List<ArrayList>();

            while (NoDet || (DetNo >= 0))
            {
                if (NoDet)
                {
                    Filename = SK.MainPath + "resp_" + UnitNum.ToString();
                }
                else
                {
                    Filename = SK.MainPath + "resp_" + UnitNum.ToString() + "_" + DetNo.ToString();
                }
                if (STno >= 0) Filename += "_" + STno.ToString();

                Names.Clear(); Types.Clear(); Formats.Clear();
                Names.Add("ID"); Types.Add("N"); Formats.Add("10");
                Names.Add("Facade"); Types.Add("L"); Formats.Add("");
                Values.Clear();

                FileStream fs = null;
                FileStream fs2 = null;
                BinaryWriter bw = null;
                BinaryWriter bw2 = null;
                try
                {
                    fs = File.Open(Filename + ".shp", FileMode.Create);
                    fs2 = File.Open(Filename + ".shx", FileMode.Create);
                    bw = new BinaryWriter(fs);
                    bw2 = new BinaryWriter(fs2);
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error opening Shape file '" + Filename + ".*' for writing", SK.enMod2Str(enMod.Results)); return false;
                }
                byte[] Header;
                double xmin = 1e38, ymin = 1e38, zmin = 1e38, mmin = 1e38, xmax = -xmin, ymax = -ymin, zmax = -zmin, mmax = -mmin;
                double dummy = 0; // Z-verdi er ikke i bruk
                Hashtable[] Params = new Hashtable[SK.Task.np];
                Hashtable[] ParTyp = new Hashtable[SK.Task.np]; // B = Bool, I = Integer, D = Double, S(N) = string(N)
                try
                {
                    int i = -1, j;
                    // Skriver dummy headere
                    bw.Write(new byte[100]); bw2.Write(new byte[100]);
                    for (i = 0; i < SK.Task.np; i++)
                    {
                        // Skriver til index (posisjon og lengde)
                        bw2.Write(CU.B2L((int)(fs.Position / 2)));
                        bw2.Write(CU.B2L(36 / 2));
                        // Skriver header til hovedfil (1-basert teller og lengde)
                        bw.Write(CU.B2L(i + 1));
                        bw.Write(CU.B2L(36 / 2));
                        // Skriver record til hovedfil (type, X, Y, Z, M)
                        bw.Write((int)11);
                        bw.Write(SK.Task.points[i, 0]);
                        bw.Write(SK.Task.points[i, 1]);
                        xmin = Math.Min(SK.Task.points[i, 0], xmin); ymin = Math.Min(SK.Task.points[i, 1], ymin);
                        xmax = Math.Max(SK.Task.points[i, 0], xmax); ymax = Math.Max(SK.Task.points[i, 1], ymax);
                        bw.Write(dummy);
                        zmin = 0.0; zmax = 0.0;
                        bw.Write((double)(SK.Task.resultpoints[i, UnitNum]));
                        mmin = Math.Min(SK.Task.resultpoints[i, UnitNum], mmin);
                        mmax = Math.Max(SK.Task.resultpoints[i, UnitNum], mmax);
                        // Skriver attributter til Hashtable for senere skriving til .DBF-fil
                        Values.Add(new ArrayList()); j = Values.Count - 1;
                        Values[j].Add(SK.Task.pntID[i]);
                        Values[j].Add(SK.Task.facade[i]);
                        // LATER: Legg til details her...?
                    }
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error writing to Shape file '" + Filename + ".*'", SK.enMod2Str(enMod.Results)); return false;
                }
                // Spoler tilbake og skriver hovedheadere
                try
                {
                    Header = GetHeader((int)fs.Position, 11, xmin, xmax, ymin, ymax, zmin, zmax, mmin, mmax);
                    bw.Seek(0, SeekOrigin.Begin);
                    bw.Write(Header);
                    Header = GetHeader((int)fs2.Position, 11, xmin, xmax, ymin, ymax, zmin, zmax, mmin, mmax);
                    bw2.Seek(0, SeekOrigin.Begin);
                    bw2.Write(Header);
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error writing main header to Shape file '" + Filename + ".*'", SK.enMod2Str(enMod.Results)); return false;
                }
                if (bw2 != null) bw2.Close();
                if (fs2 != null) fs2.Close();
                if (bw != null) bw.Close();
                if (fs != null) fs.Close();
                // Skriver til .DBF-fil
                if (!Geo.WriteDBF(Filename + ".dbf", Names, Types, Formats, Values, Errors, SK.enMod2Str(enMod.Results))) return false;
                DetNo--; NoDet = false;
            }
            return true;
        }

        /// <summary>
        /// Lagrer filer for en gitt kontur-definisjon
        /// </summary>
        /// <param name="ConNum">Konturnummer</param>
        /// <param name="stError">Er false bare dersom alle deloppdrag er ferdige, og resultatene fra disse kunne leses</param>
        /// <returns>False dersom det oppsto en feil, true hvis alt går feilfritt</returns>
        static bool SaveContour(int ConNum, bool stError, Queue<typError> Errors)
        {
            skContour Con;
            string Filename = "";
            List<string> Names = new List<string>(), Types = new List<string>(), Formats = new List<string>();
            List<ArrayList> Values = new List<ArrayList>();
            
            try
            {
                Con = SK.Contours[ConNum];
                Con.Layers = MK.ToContourMap(Con.UnitIdx, Con.Levels, Con.SmoothRadius, Con.CellSize, MapBox);
                layers[ConNum] = Con.Layers.ToArray();
            }
            catch (Exception ee)
            {
                CU.SetError(Errors, ee, "Error generating contours for contour no. " + ConNum.ToString(), SK.enMod2Str(enMod.Results));
                return false;
            }
            try
            {
                Names.Clear(); Types.Clear(); Formats.Clear();
                Names.Add("Unit"); Types.Add("C"); Formats.Add("50");
                Names.Add("Level"); Types.Add("N"); Formats.Add("8:2");
                Values.Clear();

                Filename = SK.MainPath + "resg_" + ConNum.ToString();
                FileStream fs = null;
                FileStream fs2 = null;
                BinaryWriter bw = null;
                BinaryWriter bw2 = null;
                try
                {
                    fs = File.Open(Filename + ".shp", FileMode.Create);
                    fs2 = File.Open(Filename + ".shx", FileMode.Create);
                    bw = new BinaryWriter(fs);
                    bw2 = new BinaryWriter(fs2);
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error opening Shape file '" + Filename + ".*' for writing", SK.enMod2Str(enMod.Results)); return false;
                }
                byte[] Header;
                Hashtable[] Params = new Hashtable[Con.Layers.Count];
                Hashtable[] ParTyp = new Hashtable[Con.Layers.Count]; // I = Integer, D = Double, S(N) = string(N)
                typBBox polybox, totalbox = new typBBox(1e38, -1e38, 1e38, -1e38);
                int[] parts;
                int i, j, k, npnts, nparts, polysize;
                StringBuilder unitstr;
                try
                {
                    // Skriver dummy headere
                    bw.Write(new byte[100]); bw2.Write(new byte[100]);
                    // Skriver ett og ett polygon
                    for (i = 0; i < Con.Layers.Count; i++)
                    {
                        // Trenger å vite hvor mange punkt det er i polygonet, samt utstrekning
                        nparts = Con.Layers[i].PolyLines.Count;
                        npnts = 0;
                        polybox = new typBBox(1e38, -1e38, 1e38, -1e38);
                        parts = new int[nparts];
                        for (j = 0; j < nparts; j++)
                        {
                            parts[j] = npnts;
                            npnts += Con.Layers[i].PolyLines[j].Length;
                            for (k = 0; k < Con.Layers[i].PolyLines[j].Length; k++)
                                polybox.Extend(Con.Layers[i].PolyLines[j][k].x, Con.Layers[i].PolyLines[j][k].y);
                        }
                        totalbox.Extend(polybox);
                        polysize = 44 + 4 * nparts + 16 * npnts;
                        // Skriver til index (posisjon og lengde)
                        bw2.Write(CU.L2B((int)(fs.Position / 2)));
                        bw2.Write(CU.L2B(polysize / 2));
                        // Skriver header til hovedfil (1-basert teller og lengde)
                        bw.Write(CU.L2B(i + 1));
                        bw.Write(CU.L2B(polysize / 2));
                        // Skriver record til hovedfil (type, box, parts, points)
                        bw.Write((int)5);
                        bw.Write(polybox.Xlow);
                        bw.Write(polybox.Ylow);
                        bw.Write(polybox.Xhigh);
                        bw.Write(polybox.Yhigh);
                        bw.Write(nparts);
                        bw.Write(npnts);
                        for (j = 0; j < nparts; j++)
                            bw.Write(parts[j]);
                        for (j = 0; j < nparts; j++)
                            for (k = 0; k < Con.Layers[i].PolyLines[j].Length; k++)
                            {
                                bw.Write(Con.Layers[i].PolyLines[j][k].x);
                                bw.Write(Con.Layers[i].PolyLines[j][k].y);
                            }
                        // Skriver attributter til Hashtable for senere skriving til .DBF-fil
                        Values.Add(new ArrayList()); j = Values.Count - 1;
                        unitstr = new StringBuilder();
                        for (k = 0; k < Con.UnitIdx.Count; k++)
                        {
                            if (k > 0) unitstr.Append("/");
                            unitstr.Append(SK.enUnit2Str(SK.Task.Units[Con.UnitIdx[k]].Unit, SK.Task.Units[Con.UnitIdx[k]].FW, SK.Task.Units[Con.UnitIdx[k]].TW));
                        }
                        Values[j].Add(unitstr.ToString());
                        Values[j].Add((double)Con.Layers[i].Level);
                    }
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error writing to Shape file '" + Filename + ".*'", SK.enMod2Str(enMod.Results)); return false;
                }
                // Spoler tilbake og skriver hovedheadere
                try
                {
                    Header = GetHeader((int)fs.Position, 5, totalbox.Xlow, totalbox.Xhigh, totalbox.Ylow, totalbox.Yhigh, 0, 0, 0, 0);
                    bw.Seek(0, SeekOrigin.Begin);
                    bw.Write(Header);
                    Header = GetHeader((int)fs2.Position, 5, totalbox.Xlow, totalbox.Xhigh, totalbox.Ylow, totalbox.Yhigh, 0, 0, 0, 0);
                    bw2.Seek(0, SeekOrigin.Begin);
                    bw2.Write(Header);
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error writing main header to Shape file '" + Filename + ".*'", SK.enMod2Str(enMod.Results)); return false;
                }
                if (bw2 != null) bw2.Close();
                if (fs2 != null) fs2.Close();
                if (bw != null) bw.Close();
                if (fs != null) fs.Close();
                // Skriver til .DBF-fil
                if (!Geo.WriteDBF(Filename + ".dbf", Names, Types, Formats, Values, Errors, SK.enMod2Str(enMod.Results))) return false;
            }
            catch (Exception ee)
            {
                CU.SetError(Errors, ee, "Error writing to contour file '" + Filename + ".*'", SK.enMod2Str(enMod.Results));
                return false;
            }
            return true;
        }

        /// <summary>Gjør klar til lagring av bygninger med fasadepunkter</summary>
        /// <param name="STno">Hvis 0 eller større, legges dette tallet til filnavnene</param>
        /// <returns>False dersom det oppsto en feil, true hvis alt går feilfritt</returns>
        static bool InitSaveFpnts(int STno,
                           out string[] Filename, out int recnum, out long BcntPos, out int Bcnt,
                           out double xmin, out double xmax,
                           out double ymin, out double ymax,
                           out double zmin, out double zmax,
                           out double[] mmin, out double[] mmax,
                           out FileStream[] fs, out BinaryWriter[] bw,
                           out FileStream[] fs2, out BinaryWriter[] bw2,
                           out FileStream fsB, out BinaryWriter bwB,
                           out List<string> Names, out List<string> Types, out List<string> Formats,
                           out List<ArrayList> Values,
                           Queue<typError> Errors)
        {

            int iu = 0, nu = SK.Task.numUnits;
            Filename = new string[nu];
            recnum = 0;
            BcntPos = 0;
            Bcnt = 0;
            xmin = 1e38; ymin = 1e38; zmin = 1e38; mmin = new double[nu];
            xmax = -xmin; ymax = -ymin; zmax = -zmin; mmax = new double[nu];
            fs = new FileStream[nu];
            fs2 = new FileStream[nu];
            bw = new BinaryWriter[nu];
            bw2 = new BinaryWriter[nu];
            fsB = null;
            bwB = null;
            Names = new List<string>();
            Types = new List<string>();
            Formats = new List<string>();
            Values = new List<ArrayList>();

            Names.Clear(); Types.Clear(); Formats.Clear();
            Names.Add("ID"); Types.Add("N"); Formats.Add("10");
            Names.Add("BuildID"); Types.Add("N"); Formats.Add("10");
            Names.Add("Facade"); Types.Add("L"); Formats.Add("");
            Names.Add("Ref"); Types.Add("C"); Formats.Add("12");

            for (iu = 0; iu < nu; iu++)
            {
                mmin[iu] = 1e38; mmax[iu] = -mmin[iu];
                Filename[iu] = SK.MainPath + "resf_" + iu.ToString();
                if (STno >= 0) Filename[iu] += "_" + STno.ToString();

                try
                {
                    fs[iu] = File.Open(Filename[iu] + ".shp", FileMode.Create);
                    fs2[iu] = File.Open(Filename[iu] + ".shx", FileMode.Create);
                    bw[iu] = new BinaryWriter(fs[iu]);
                    bw2[iu] = new BinaryWriter(fs2[iu]);
                    // Skriver dummy headere
                    bw[iu].Write(new byte[100]); bw2[iu].Write(new byte[100]);
                }
                catch (Exception ee)
                {
                    CU.SetError(Errors, ee, "Error opening/initializing Shape file '" + Filename[iu] + ".*'", SK.enMod2Str(enMod.Results));
                    for (iu = 0; iu < nu; iu++)
                    {
                        if (bw[iu] != null) bw[iu].Close();
                        if (bw2[iu] != null) bw2[iu].Close();
                        if (fs[iu] != null) fs[iu].Close();
                        if (fs2[iu] != null) fs2[iu].Close();
                    }
                    if (bwB != null) bwB.Close();
                    if (fsB != null) fsB.Close();
                    return false;
                }
            }

            return clsImport.BygningerTilFilÅpne(SK.MainPath + "resb.dat", SK.Task.config, Errors, out fsB, out bwB, out BcntPos);
        }


        /// <summary>Lagrer alle bygninger og tilhørende fasadepunkt i et deloppdrag</summary>
        /// <returns>False dersom det oppsto en feil, true hvis alt går feilfritt</returns>
        static bool SaveBldFpnts(skSubTask subtask,
                          string[] Filename, ref int recnum, ref int Bcnt,
                          ref double xmin, ref double xmax,
                          ref double ymin, ref double ymax,
                          ref double zmin, ref double zmax,
                          double[] mmin, double[] mmax,
                          FileStream[] fs, BinaryWriter[] bw,
                          FileStream[] fs2, BinaryWriter[] bw2,
                          FileStream fsB, BinaryWriter bwB,
                          List<ArrayList> Values,
                          Queue<typError> Errors)
        {

            double dbval;
            Bygning B;
            int iBld, ifp, j, iu = 0, nu = subtask.numUnits;
            try
            {
                for (iBld = 0; iBld < subtask.buildings.Count; iBld++)
                {
                    B = subtask.buildings[iBld];
                    if (B.fasadepunkt.Count == 0) continue; // Skriver ikke "tomme" bygninger
                    for (ifp = 0; ifp < B.fasadepunkt.Count; ifp++)
                    {
                        recnum++;
                        for (iu = 0; iu < nu; iu++)
                        {
                            // Skriver til index (posisjon og lengde)
                            bw2[iu].Write(CU.B2L((int)(fs[iu].Position / 2)));
                            bw2[iu].Write(CU.B2L(36 / 2));
                            // Skriver header til hovedfil (1-basert teller og lengde)
                            bw[iu].Write(CU.B2L(recnum));
                            bw[iu].Write(CU.B2L(36 / 2));
                            // Skriver record til hovedfil (type, X, Y, Z, M)
                            bw[iu].Write((int)11);
                            bw[iu].Write(B.fasadepunkt[ifp].sted.posisjonE);
                            bw[iu].Write(B.fasadepunkt[ifp].sted.posisjonN);
                            xmin = Math.Min(B.fasadepunkt[ifp].sted.posisjonE, xmin); ymin = Math.Min(B.fasadepunkt[ifp].sted.posisjonN, ymin);
                            xmax = Math.Max(B.fasadepunkt[ifp].sted.posisjonE, xmax); ymax = Math.Max(B.fasadepunkt[ifp].sted.posisjonN, ymax);
                            bw[iu].Write(B.fasadepunkt[ifp].sted.h);
                            zmin = Math.Min(B.fasadepunkt[ifp].sted.h, zmin);
                            zmax = Math.Max(B.fasadepunkt[ifp].sted.h, zmax);
                            // Henter dB-verdier fra FØRSTE Situasjon
                            // NB: Forutsetter at nivå[] inneholder de samme enheter
                            //     som definert i task.xml, og i samme rekkefølge!
                            dbval = B.fasadepunkt[ifp].utesituasjoner[0].utestøy.nivå[iu].verdi;
                            bw[iu].Write(dbval);
                            mmin[iu] = Math.Min(dbval, mmin[iu]);
                            mmax[iu] = Math.Max(dbval, mmax[iu]);
                        }
                        // Skriver attributter til Hashtable for senere skriving til .DBF-fil
                        // (Denne er felles for alle enheter)
                        Values.Add(new ArrayList()); j = Values.Count - 1;
                        Values[j].Add(Convert.ToInt64(B.beskrivelse));
                        Values[j].Add(B.bygningsnummer);
                        Values[j].Add(true);
                        Values[j].Add(B.fasadepunkt[ifp].sted.skissereferanse);
                    }
                    B.beskrivelse = "Autogenerert fra NorCalc";
                    Bcnt++;
                    if (!clsImport.BygningTilFilSkriv(fsB, bwB, B, Errors))
                    {
                        for (iu = 0; iu < nu; iu++)
                        {
                            if (bw[iu] != null) bw[iu].Close();
                            if (bw2[iu] != null) bw2[iu].Close();
                            if (fs[iu] != null) fs[iu].Close();
                            if (fs2[iu] != null) fs2[iu].Close();
                        }
                        if (bwB != null) bwB.Close();
                        if (fsB != null) fsB.Close();
                        return false;
                    }
                }
            }
            catch (Exception ee)
            {
                CU.SetError(Errors, ee, "Error writing to Shape file '" + Filename[iu] + ".*'", SK.enMod2Str(enMod.Results));
                for (iu = 0; iu < nu; iu++)
                {
                    if (bw[iu] != null) bw[iu].Close();
                    if (bw2[iu] != null) bw2[iu].Close();
                    if (fs[iu] != null) fs[iu].Close();
                    if (fs2[iu] != null) fs2[iu].Close();
                }
                if (bwB != null) bwB.Close();
                if (fsB != null) fsB.Close();
                return false;
            }
            return true;
        }

        /// <summary>Lagrer alle aktuelle resultatfiler for en gitt enhet</summary>
        /// <returns>False dersom det oppsto en feil, true hvis alt går feilfritt</returns>
        static bool FinishSaveFpnts(string[] Filename, ref long BcntPos, ref int Bcnt,
                             ref double xmin, ref double xmax,
                             ref double ymin, ref double ymax,
                             ref double zmin, ref double zmax,
                             ref double[] mmin, ref double[] mmax,
                             FileStream[] fs, BinaryWriter[] bw,
                             FileStream[] fs2, BinaryWriter[] bw2,
                             FileStream fsB, BinaryWriter bwB,
                             List<string> Names, List<string> Types, List<string> Formats,
                             List<ArrayList> Values,
                             Queue<typError> Errors)
        {

            byte[] Header;
            int iu = 0, nu = SK.Task.numUnits;
            // Spoler tilbake og skriver hovedheadere
            try
            {
                for (iu = 0; iu < nu; iu++)
                {
                    Header = GetHeader((int)fs[iu].Position, 11, xmin, xmax, ymin, ymax, zmin, zmax, mmin[iu], mmax[iu]);
                    bw[iu].Seek(0, SeekOrigin.Begin);
                    bw[iu].Write(Header);
                    Header = GetHeader((int)fs2[iu].Position, 11, xmin, xmax, ymin, ymax, zmin, zmax, mmin[iu], mmax[iu]);
                    bw2[iu].Seek(0, SeekOrigin.Begin);
                    bw2[iu].Write(Header);
                    // Skriver til .DBF-fil
                    if (!Geo.WriteDBF(Filename[iu] + ".dbf", Names, Types, Formats, Values, Errors, SK.enMod2Str(enMod.Results))) return false;
                }

                return clsImport.BygningerTilFilLukk(fsB, bwB, BcntPos, Bcnt, Errors);

            }
            catch (Exception ee)
            {
                CU.SetError(Errors, ee, "Error writing main header to Shape file '" + Filename[iu] + ".*'", SK.enMod2Str(enMod.Results));
                return false;
            }
            finally
            {
                for (iu = 0; iu < nu; iu++)
                {
                    if (bw[iu] != null) bw[iu].Close();
                    if (bw2[iu] != null) bw2[iu].Close();
                    if (fs[iu] != null) fs[iu].Close();
                    if (fs2[iu] != null) fs2[iu].Close();
                }
                if (bwB != null) bwB.Close();
                if (fsB != null) fsB.Close();
            }
        }


        /// <summary>
        /// Returnerer en Shape Main File Header (100 bytes)
        /// </summary>
        /// <param name="FileSize">Størrelsen på Shape-filen</param>
        /// <param name="ShapeType">Shape Type som filen inneholder</param>
        /// <param name="xmin">Minste x-verdi</param>
        /// <param name="xmax">Største x-verdi</param>
        /// <param name="ymin">Minste y-verdi</param>
        /// <param name="ymax">Største y-verdi</param>
        /// <param name="zmin">Minste z-verdi</param>
        /// <param name="zmax">Største z-verdi</param>
        /// <param name="mmin">Minste measure-verdi</param>
        /// <param name="mmax">Største measure-verdi</param>
        /// <returns>Main header (100 byte array)</returns>
        static byte[] GetHeader(int FileSize, int ShapeType,
                         double xmin, double xmax,
                         double ymin, double ymax,
                         double zmin, double zmax,
                         double mmin, double mmax)
        {
            byte[] result = new byte[100];
            BitConverter.GetBytes(CU.L2B((int)9994)).CopyTo(result, 0);
            BitConverter.GetBytes(CU.L2B(FileSize / 2)).CopyTo(result, 24);
            BitConverter.GetBytes((int)1000).CopyTo(result, 28);
            BitConverter.GetBytes(ShapeType).CopyTo(result, 32);
            BitConverter.GetBytes(xmin).CopyTo(result, 36);
            BitConverter.GetBytes(ymin).CopyTo(result, 44);
            BitConverter.GetBytes(xmax).CopyTo(result, 52);
            BitConverter.GetBytes(ymax).CopyTo(result, 60);
            BitConverter.GetBytes(zmin).CopyTo(result, 68);
            BitConverter.GetBytes(zmax).CopyTo(result, 76);
            BitConverter.GetBytes(mmin).CopyTo(result, 84);
            BitConverter.GetBytes(mmax).CopyTo(result, 92);
            return result;
        }

    }
    #endregion
}
